/******************************************************************************
 * $Id: qctdataset.cpp,v 1.134 2021/02/02 14:50:02 werdna Exp werdna $
 *
 * Project:  GDAL
 * Purpose:  QCT (Memory Map Quick Chart) Driver
 * Authors:  Andrew Brooks <arb@sat.dundee.ac.uk>
 *           Andrew C Aitchison
 ******************************************************************************
 * Copyright (c) 2010-4, Andrew Brooks <arb@sat.dundee.ac.uk>
 * Copyright (c) 2014-20, Andrew C Aitchison <qct@aitchison.me.uk>
 ****************************************************************************/
 
/* -*- tab-width: 4 ; indent-tabs-mode: nil ; c-basic-offset 'tab-width -*- */

#ifdef FRMT_qct

#if __cplusplus < 201103 // ie before clang++-6.0
#define override
#define nullptr 0
#endif

#include "time.h"   // to print nOrigFileTime

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Winconsistent-missing-destructor-override"
#pragma clang diagnostic ignored "-Wdocumentation-unknown-command"
#pragma clang diagnostic ignored "-Wweak-vtables"
#pragma clang diagnostic ignored "-Wunused-template"
#pragma clang diagnostic ignored "-Wunknown-warning-option"
#pragma clang diagnostic ignored "-Wimplicit-int-float-conversion"
#include "gdal_pam.h"
#pragma clang diagnostic pop
#include "ogr_spatialref.h"

#include "qct.h"

// Copied from gcore/gdal_priv.h
#define DIV_ROUND_UP(a, b) ( ((a) % (b)) == 0 ? ((a) / (b)) : (((a) / (b)) + 1) )


// Set *one* of these definitions (to 1)
// QC3_PIXEL_IS_PIXEL is to be assumed if none are set.
// #define QC3_PIXEL_IS_PIXEL 1
// #define QC3_PIXEL_IS_TILE 1

#if !defined CODE_ANALYSIS
// cppcheck quits at this line
CPL_CVSID("$Id: qctdataset.cpp,v 1.134 2021/02/02 14:50:02 werdna Exp werdna $")
#endif

CPL_C_START

#if GDAL_VERSION_MAJOR >= 2
    // Should work with GDAL 1.11.5 if I knew how to write an ogr plugin for it.
    // defined in qctvector.cpp
    extern void   GDALRegister_qctvector(void);
#endif
#ifdef FRMT_qed
    // defined in qed.cpp
    extern void   GDALRegister_qed(void);
#endif

void   GDALRegister_qct(void);
CPL_C_END

#if GDAL_VERSION_NUM < 2010000
#define GDAL_IDENTIFY_UNKNOWN -1
#define GDAL_IDENTIFY_FALSE 0
#define GDAL_IDENTIFY_TRUE 1
#else
// These are defined in gdal/gdal_priv.h
#endif // GDAL_VERSION_NUM < 2010000

#ifdef CODE_ANALYSIS

// Printing variables with CPLDebug can hide
// the fact that they are not otherwise used ...
#define CPLDebug(...)

#endif // CODE_ANALYSIS

void dumpPGM(unsigned int width,
             unsigned int height,
             unsigned char* const data,
             unsigned int rowlength,
             CPLString osBaseLabel,
             unsigned int nMaxPGM
             );

static const unsigned int nQCTNoData = 255; // warped tile nodata is 0 but that often appears on the maps.
static const unsigned int nQCTonlineData = 129; // indicate tiles that may be available online
static const unsigned int nQCTencryptedData = 130; // indicate qct tiles that use some sort of encryption
static const unsigned int nQCTYesData = 254; // when we want to draw something

static const size_t nQCT_TileIndexStart=0x45A0;
static const size_t nQC3_TileIndexStart=0x28;
static const size_t MINUSONE_as_SIZET = static_cast<unsigned long long>(-1LL);

#if 0 // not currently used
static short QCTGetShort( short *pszField )
{
    /* this is fine for little-endian, but will need changing for bigendian */
    return *pszField;
}
#endif

static void QCTreport(char const* label, char const* value) {
    if (value) {
        CPLDebug("QCT", "%s: %s", label, value);
    } else {
        CPLDebug("QCT", "%s: empty", label);
    }
    (void)label; // Appease static analyzers
}


/************************************************************************/
/* ==================================================================== */
/*                         QCTDataset                                   */
/* ==================================================================== */
/************************************************************************/

class QCTRasterBand;

class QCTDataset : public GDALPamDataset
{
    friend class QCTRasterBand;
    
    VSILFILE        *fp;
    GDALColorTable  *poColorTable;
    GByte            abyHeader[0x5a0];
    unsigned int     anExtended[8];
    
    QCTRasterBand*   papoOverviewBands[10];

    unsigned int magic, version;
    char *pszLongTitle, *pszName, *pszIdentifier,
        *pszEdition, *pszRevision, *pszKeywords,
        *pszCopyright, *pszScale, *pszDatum,
        *pszDepths, *pszHeights, *pszProjection,
        *pszOrigFileName;
    int nQCTFlags, nOrigFileSize, nReserved1;
    unsigned int nOutlinePts;
    time_t  nOrigFileTime;
    double  adfQCTdatumShift[2], *adfOutlinePoints ;
    double QCTpolynomial[40];
    latlon TL, TR, BL, BR;
    size_t nQCTfilesize;
    double dScalePixels;
#ifdef QCT_USE_GCP
    int         nGCPCount;
    GDAL_GCP    *pasGCPList;
    int QCTsetGCPMetaData_from_coefficients(int count, const char* szMDItem,
                                         double values[], const char* szCollection);
#endif
    QC3Header *poQC3_header;
    long long *anTileIndex;
    
    long long QCTGetTileIndex(unsigned int x, unsigned int y);
    long long* QCTBuildTileIndex( );
    int quickQC3tileColor(unsigned int i, unsigned int j);
    void QCTprintExtendedData( );
    int nOverviewCount;
    unsigned char acInterpolationMatrix[128][128];

public:
    QCTDataset();
    ~QCTDataset() override;
    
    static GDALDataset *Open( GDALOpenInfo * );
    static int Identify( GDALOpenInfo * );

    inline bool bNeedQC3data() {
        if (((version & qct_version_for_qc3_mask) == qct_version_for_qc3)
            || magic == qcthree_magic
            ) {
            return TRUE;
        }
        return FALSE;
    }
    inline bool bHaveQC3data() {
        if (poQC3_header && poQC3_header->fpQC3) return TRUE;
        return FALSE;
    }

#ifdef QCT_USE_GCP
    virtual int     GetGCPCount() override;
    virtual const GDAL_GCP* GetGCPs() override;
#if GDAL_VERSION_MAJOR >= 3
    virtual const char *GetGCPProjection();
#else
    virtual const char *GetGCPProjection() override;
#endif

#else // ! QCT_USE_GCP
    CPLErr GetGeoTransform( double * padfTransform ) override;
#endif // QCT_USE_GCP


#if GDAL_VERSION_MAJOR >= 3
    const OGRSpatialReference* GetSpatialRef() const override {
        return GetSpatialRefFromOldGetProjectionRef();
    }
    const char *_GetProjectionRef() override;
#else
    const char *GetProjectionRef() override;
#endif

    char       **GetFileList() override;

    char *QCTGetString(size_t byteaddr);
    char *QCTGetStringIndirect(size_t byteaddr);
    void xy_to_latlon(double pixel_x, double pixel_y, latlon *latlon);
}; // class QCTDataset


/************************************************************************/
/* ==================================================================== */
/*                            QCTRasterBand                             */
/* ==================================================================== */
/************************************************************************/

class QCTRasterBand : public GDALPamRasterBand
{
    friend class QCTDataset;
    
    int          iOverview;
    // int          nOverviewCount; // now in QCTDataset
    // QCTRasterBand**  papoOverviewBands;
public:

    QCTRasterBand( QCTDataset *poDSIn,
                   int nBandIn,
                   int nThisOverviewIn
                   // , int nOverviewCountIn
                   // , QCTRasterBand**papoOverviewBandsIn
                   );

    ~QCTRasterBand() override;
    
    virtual GDALColorInterp GetColorInterpretation() override;
    virtual GDALColorTable  *GetColorTable() override;

    // virtual double GetNoDataValue( int *pbSuccess = nullptr ) override;
    virtual double GetNoDataValue( int *) override;
    // virtual CPLErr SetNoDataValue( double ) override;
    
    virtual CPLErr IReadBlock( int, int, void * ) override;
    virtual void read_QCT_Tile( VSILFILE *fp,
                                int tile_xx, int tile_yy,
                                void *pImage);
    virtual void read_QCthree_Tile( VSILFILE *fp,
                                int tile_xx, int tile_yy,
                                void *pImage);

    virtual int    GetOverviewCount() override;
    virtual GDALRasterBand *GetOverview( int ) override;
    // virtual QCTRasterBand *GetOverview( int ) override;

#if GDAL_VERSION_NUM >= 2020000
    virtual int IGetDataCoverageStatus( int nXOff, int nYOff,
                                        int nXSize, int nYSize,
                                        int nMaskFlagStop,
                                        double* pdfDataPct) override;
#endif

}; // class QCTRasterBand


/************************************************************************/
/*                           QCTRasterBand()                            */
/************************************************************************/

QCTRasterBand::QCTRasterBand(
                             QCTDataset *poDSIn,
                             int nBandIn,
                             int iOverviewIn
                             // , int nOverviewCountIn
                             //, QCTRasterBand**papoOverviewBandsIn
                             )
    :
    iOverview(iOverviewIn)
    // ,papoOverviewBands(papoOverviewBandsIn)
{
    CPLAssert (iOverview >= -1 );
    poDS = poDSIn;
    nBand = nBandIn;

    eDataType = GDT_Byte;
    
    if ( poDSIn->magic != qct_info_magic
         && poDSIn->magic != qct_magic
         && poDSIn->magic != qcthree_magic) {
        CPLError(CE_Failure, CPLE_NotSupported,
                 "unexpected file magic 0x%8x", poDSIn->magic );
        return;
    }

    if (poDSIn->bNeedQC3data()) { 
        if (poDSIn->bHaveQC3data()) {
            // Image data in .qc3 file
            // Work in progress
#if defined(QC3_PIXEL_IS_TILE)
            CPLDebug("QCT", "qc3_pixel_is_tile/dummy %d %d",
                     poDSIn->nRasterXSize, poDSIn->nRasterYSize);
            
            // QC3_PIXEL_IS_TILE
            nBlockXSize = poDSIn->nRasterXSize;
            nBlockYSize = poDSIn->nRasterYSize;

            // nRecordSize = (GUIntBig)nBlockXSize * nBlockYSize * sizeof(GUInt32);
#else // QC3_PIXEL_IS_PIXEL
            CPLDebug("QCT", "qc3_pixel_is_pixel");
            
            // QC3_PIXEL_IS_PIXEL
            // this will be the default        
            if (iOverview > 10) {
                CPLDebug( "QCTRasterBand", 
                          "Overview %d too high - qc3 files have 10 levels.",
                          iOverview );
                iOverview = 10;
            }
            nBlockXSize = nBlockYSize = QC3_TILE_SIZE >> (1+iOverview);
            nRasterXSize=poDSIn->nRasterXSize >> (1+iOverview);
            nRasterYSize=poDSIn->nRasterYSize >> (1+iOverview);
#endif // QC3_PIXEL_IS_PIXEL
            CPLDebug("QCT", "QCTRasterBand(%p, %d, %d) with qc3 file",
                     static_cast<void*>(poDS), nBand, iOverview ); 
        } else {
            // No data in qct file, cannot access qc3 file.
            poDSIn->nRasterXSize = poDSIn->nRasterYSize = 1;
            nBlockXSize = nBlockYSize = 1;
            CPLDebug("QCT",
                     "QCTRasterBand(%p, %d, %d) with missing qc3 file\n",
                     // static_cast<void*>(
                     poDS //)
                     , nBand, iOverview );
            if (poDSIn->nOverviewCount>0) {
                CPLDebug("QCT",
                         "QCTRasterBand() has %d overviews but .qc3 file is missing",
                         poDSIn->nOverviewCount);
            }

        }
    } else {
        // Image data in .qct file
        if (iOverview > 6) {
            CPLDebug( "QCTRasterBand", 
                      "Overview %d too high - qct files have 6 levels.\n",
                      iOverview );
            iOverview = 6;
        }

        nBlockXSize = nBlockYSize = QCT_TILE_SIZE >> (1+iOverview);
        nRasterXSize=poDSIn->nRasterXSize >> (1+iOverview);
        nRasterYSize=poDSIn->nRasterYSize >> (1+iOverview);
        CPLDebug("QCT", "QCTRasterBand(%p, %d, %d) with qct file - dataset has %d overviews\n",
                 poDS, nBand, iOverview, poDSIn->nOverviewCount); 
    }
    CPLDebug("QCT",
             "nBlockXSize %d nBlockYSize %d",
             nBlockXSize, nBlockYSize);

    if (nBlockXSize==0||nBlockYSize==0) {
        if (nBlockXSize==0) {
            CPLDebug("QCT", "nBlockXSize is zero");
        }
        if (nBlockYSize==0) {
            CPLDebug("QCT", "nBlockYSize is zero");
        }
        exit (0);
    }
    
    nBlocksPerRow =
        DIV_ROUND_UP(nRasterXSize, nBlockXSize);
    nBlocksPerColumn =
        DIV_ROUND_UP(nRasterYSize, nBlockYSize);
    CPLDebug("QCT",
             "nBlocksPerRow %d nBlocksPerColumn %d",
             nBlocksPerRow, nBlocksPerColumn);


    
} // QCTRasterBand::QCTRasterBand

/************************************************************************/
/*                          ~QCTRasterBand()                            */
/************************************************************************/

QCTRasterBand::~QCTRasterBand()
{
}

/************************************************************************/
/*                          quickQC3tileColor()                            */
/************************************************************************/

int QCTDataset::quickQC3tileColor(unsigned int i, unsigned int j)
{
    // Look for the first colour in the huff table

    long long addTile = QCTGetTileIndex(i, j);
    if (0>addTile ||
        static_cast<unsigned long long>(addTile) >= poQC3_header->nQC3filesize) {
        CPLDebug("QCT",
                 "Data at 0x%llx for tile(%d, %d) is out of range [0, 0x%llx)",
                 addTile, i, j, poQC3_header->nQC3filesize);
        return nQCTNoData;
    } else if (0==addTile) {
        // Data may be available online
        return nQCTonlineData;
    }

    int nHuffSize = 4*QCTreadInt(poQC3_header->fpQC3, addTile+0x4C);
    CPLDebug("QCT",
             "Data at 0x%llx for tile(%d, %d) Huff table 0x%08x bytes",
             addTile, i, j, nHuffSize);
    short nColour=-1;
    while (nColour<0 && nHuffSize>0) {
        nColour = QCTFGetc(poQC3_header->fpQC3);
        // short nRunLen =
        QCTFGetc(poQC3_header->fpQC3);
        nHuffSize-=2;
    };
    // if (nHuffSize<=0) {
    if (nColour<0) {
        CPLDebug("QCT",
                 "PIXEL_IS_TILE(%d, %d) mis-parsed Huff table",
                 i ,j);
        return nQCTNoData;
    } else {
        // return = nHuffSize & 0x7f;
        return static_cast<int>(nColour);
    }
} // QCTDataset::quickQC3tileColor()


/************************************************************************/
/*                             IReadBlock()                             */
/************************************************************************/

CPLErr QCTRasterBand::IReadBlock( int nBlockXOff, int nBlockYOff,
                                  void * pImage )
{
    QCTDataset *poQDS = static_cast<QCTDataset *>(poDS);

    if (nBlockXOff<0 || nBlockYOff<0 || pImage==nullptr) {
        CPLError( CE_Failure, CPLE_IllegalArg,
                  "bad arg in QCTRasterBand::IReadBlock(%d,%d,%p)",
                  nBlockXOff, nBlockYOff, pImage );
            return CE_Failure;
    }

    if (poQDS->magic != qct_magic && poQDS->magic != qct_info_magic) {
        CPLDebug("QCT", "unsuported magic 0x%08x in QCT file", poQDS->magic);
        return CE_Failure;
    }
    // Must now have poQDS->magic = qct_magic or = qct_info_magic
    
    CPLDebug("QCT",
             "QCTRasterBand::IReadBlock(%d %d %p) band %d overview %d",
             nBlockXOff, nBlockYOff, pImage,
             nBand, iOverview
             );
    
    if (poQDS->bHaveQC3data()) {
        // We have a .qct / .qc3 file pair.
#if defined QC3_PIXEL_IS_TILE
        CPLDebug("QCT", "qc3_pixel_is_tile %d x %d",
                 poQDS->poQC3_header->width3, poQDS->poQC3_header->height3);
        unsigned char* pPix = static_cast<unsigned char*>(pImage);
        for(unsigned int j = 0; j < poQDS->poQC3_header->height3 ; j++ ) {
            for(unsigned int i = 0; i < poQDS->poQC3_header->width3 ; i++, pPix++ ) {
                long long addTilePtr = nQC3_TileIndexStart + 8*(poQDS->poQC3_header->width3*j+i);
                signed long long addTile = poQDS->QCTGetTileIndex(i, j);
                switch (addTile) {
                case -1 : // No Data
                    *pPix = nQCTNoData;
                    CPLDebug("QCT",
                             "[0x%llx]=0x%llx Data for tile(%d, %d) does not exist ",
                             addTilePtr, addTile, i, j);
                    break;
                case 0 :  // Data may be available online
                    *pPix = nQCTonlineData;
                    CPLDebug("QCT",
                             "[0x%llx]=0x%llx Data for tile(%d, %d) may be available online",
                             addTilePtr, addTile, i, j);
                    break;
                default:
                    *pPix = static_cast<unsigned char>(poQDS->quickQC3tileColor(i,j));
                    CPLDebug("QCT",
                             "QC3_PIXEL_IS_TILE %p -> %u=0x%x [0x%llx]=0x%llx for tile(%d, %d)",
                             pPix, *pPix, *pPix, addTilePtr, addTile, i, j);
                    break;
                } // switch (addTile)
            } // for i ...
        } // for j ...
        CPLDebug("QCT", "qc3_pixel_is_tile still being implemented");
#else // so !defined QC3_PIXEL_IS_TILE

        // QC3_PIXEL_IS_PIXEL
        // This will be the default
        // CPLDebug("QCT", "qc3_pixel_is_pixel");
        // long long addTilePtr = nQC3_TileIndexStart
        //   + 8*(poQDS->poQC3_header->width3*nBlockYOff+nBlockXOff);
        signed long long addTile = poQDS->QCTGetTileIndex(nBlockXOff, nBlockYOff);
        switch (addTile) {
        case -1 : // No Data
            for(int i = 0; i < nBlockXSize *nBlockYSize ; i++ ) {
                static_cast<unsigned char*>(pImage)[i] = nQCTNoData;
            }
            break;
        case 0 :  // Data may be available online
            for(int i = 0; i < nBlockXSize *nBlockYSize ; i++ ) {
                static_cast<unsigned char*>(pImage)[i] = nQCTonlineData;
            }
#if 0
            CPLDebug("QCT",
                     "[0x%llx]=0x%llx Data for tile(%d, %d) may be available online",
                     addTilePtr, addTile, nBlockXOff, nBlockYOff);
#endif
            break;
        default:
            CPLDebug("QCT", "qc3_pixel_is_pixel (overview %d) still being implemented",
                     iOverview);
            read_QCthree_Tile(poQDS->fp, nBlockXOff, nBlockYOff, pImage);
            break;
        }
#endif //  QC3_PIXEL_IS_TILE
     } else { // so not if (poQDS->bHaveQC3data()) {
        /*
          CPLAssert ( (poQDS->magic == qct_magic
                     && (poQDS->version == 2 || poQDS->version == 4))
                    || (poQDS->magic == qct_info_magic && poQDS->version == 1) );
        */
        if (poQDS->bNeedQC3data()) {
            // This .qct file needs a .qc3 file, but we have no .qc3 data.
            CPLDebug("QCT",
                     "QCTRasterBand::IReadBlock(%d,%d,%p) magic:%x version:%x lost QC3 image data",
                     nBlockXOff, nBlockYOff, pImage,
                     poQDS->magic, poQDS->version
                     );
            for(int i = 0; i < nBlockXSize *nBlockYSize ; i++ ) {
                // We cannot find the data, but it *may* be available,
                // so do not use nQCTNoData 
                static_cast<unsigned char*>(pImage)[i] = nQCTonlineData;
            }
            return CE_None;
        }
        if ( poQDS->magic == qct_magic && (poQDS->version != 2 && poQDS->version != 4)) {
            CPLError(CE_Failure, CPLE_NotSupported,
                     "QCT qct_magic (0x%08x) file support limited to versions 2 and 4. This is version %d=0x%08x",
                     qct_magic, poQDS->version, poQDS->version);
            return CE_Failure;
        }
        if ( poQDS->magic == qct_info_magic && (poQDS->version != 1)) {
            CPLError(CE_Failure, CPLE_NotSupported,
                     "QCT qct_info_magic (0x%08x) file support limited to version 1. This is version %d",
                     qct_info_magic, poQDS->version);
            return CE_Failure;
        }
        CPLDebug("QCT",
                 "QCTRasterBand::IReadBlock(%d,%d,%p) magic:0x%x version:%d",
                 nBlockXOff, nBlockYOff, pImage,
                 poQDS->magic, poQDS->version
                 );

        unsigned long long nTilePos =
            poQDS->QCTGetTileIndex(nBlockXOff,nBlockYOff);
        if (nTilePos==0) {
            for(int i = 0; i < nBlockXSize *nBlockYSize ; i++ ) {
                static_cast<unsigned char*>(pImage)[i]
                    = nQCTonlineData; // may change this value ?
            }
            return CE_None;
        }

        if ((nTilePos < nQC3_TileIndexStart || nTilePos >= poQDS->nQCTfilesize) ) {
            for(int i = 0; i < nBlockXSize *nBlockYSize ; i++ ) {
                static_cast<unsigned char*>(pImage)[i] = nQCTNoData;
            }
            if (nTilePos!=MINUSONE_as_SIZET) {
                CPLDebug("QCT",
                         "Skipping tile %d %d: nTilePos %lld out of range [%zu,%zu)\n",
                         nBlockXOff, nBlockYOff,
                         nTilePos, nQC3_TileIndexStart, poQDS->nQCTfilesize);
            }
            // return CE_Failure;
            return CE_None;
        }

        if ( int nSeekRes=VSIFSeekL( poQDS->fp, nTilePos, SEEK_SET ) ) {
            CPLError( CE_Failure, CPLE_AppDefined, 
                      "QCT tile %d %d seek to %lld failed: %d",
                      nBlockXOff, nBlockYOff, nTilePos, nSeekRes);
            return CE_Failure;
        }
        /* file ptr now at start of the qct tile */
        read_QCT_Tile(poQDS->fp, nBlockXOff, nBlockYOff, pImage);
    } // else part of if (poQDS->bHaveQC3data())
    
    return CE_None;
} // QCTRasterBand::IReadBlock()

/************************************************************************/
/*                           GetNoDataValue()                           */
/************************************************************************/

double QCTRasterBand::GetNoDataValue( int *pbSuccess )
{
#if 0
  return GDALPamRasterBand::GetNoDataValue( pbSuccess );
#else
  if (pbSuccess) {
      *pbSuccess=TRUE;
  }
  return nQCTNoData;
#endif
}

/************************************************************************/
/*                        IGetDataCoverageStatus()                     */
/************************************************************************/

#if GDAL_VERSION_NUM >= 2020000
    // See https://trac.osgeo.org/gdal/wiki/rfc63_sparse_datasets_improvements
    // and https://github.com/rouault/gdal2/blob/sparse_datasets/gdal/frmts/gtiff/geotiff.cpp
int QCTRasterBand::IGetDataCoverageStatus( int nXOff, int nYOff,
                                           int nXSize, int nYSize,
                                           int nMaskFlagStop, double* pdfDataPct)
{
    CPLDebug("QCT", "IGetDataCoverageStatus(%d, %d, %d, %d, 0%o, %g)",
             nXOff, nYOff, nXSize, nYSize, nMaskFlagStop,
             pdfDataPct ? *pdfDataPct:-1);

    QCTDataset *poQDS = static_cast<QCTDataset *>(poDS);
    double dfDataPct = -1.0 ;
    int nStatus = 0;

    // unsigned int magic = poQDS->magic;
    // unsigned int version = poQDS->version;
    
    if (pdfDataPct) {
        *pdfDataPct = dfDataPct;
    }

    if (poQDS->bNeedQC3data()) {
        if(!poQDS->bHaveQC3data()) {
            if (pdfDataPct) {
                *pdfDataPct = 0.0;
            }
            CPLDebug("QCT", "IGetDataCoverageStatus(%d, %d, %d, %d, %d, %p): - QC3 data-file missing",
                     nXOff, nYOff, nXSize, nYSize, nMaskFlagStop,
                     static_cast<void*>(pdfDataPct));
            return GDAL_DATA_COVERAGE_STATUS_EMPTY;
        }
#if defined QC3_PIXEL_IS_TILE
        if (pdfDataPct) {
            *pdfDataPct = 100.0;
        }
        return GDAL_DATA_COVERAGE_STATUS_DATA;
#endif // defined QC3_PIXEL_IS_TILE
    }

    // calculate nStatus and (p)dfDataPct from 
    // nXOff, nYOff, nXSize, nYSize and nMaskFlagStop

    CPLDebug("QCT", "IGetDataCoverageStatus(%d, %d, %d, %d, 0%o, %p)",
             nXOff, nYOff, nXSize, nYSize, nMaskFlagStop, static_cast<void*>(pdfDataPct) );

    const int iXBlockStart = nXOff / nBlockXSize;
    const int iXBlockEnd = (nXOff + nXSize - 1) / nBlockXSize;
    const int iYBlockStart = nYOff / nBlockYSize;
    const int iYBlockEnd = (nYOff + nYSize - 1) / nBlockYSize;

    GIntBig nPixelsData = 0;
    for(int iY = iYBlockStart; iY <= iYBlockEnd; ++iY ) {
        for(int iX = iXBlockStart; iX <= iXBlockEnd; ++iX ) {
            // const int nBlockId = iX + iY * nBlocksPerRow;
            bool bHasData = false;
            
            long long nTilePos = poQDS->QCTGetTileIndex(iX,iY);
            
            if( nTilePos <= 0 ) {
                nStatus |= GDAL_DATA_COVERAGE_STATUS_EMPTY;
                CPLDebug("QCT", "IGetDataCoverageStatus(%d, %d, %d, %d, 0%o, %p) block %d %d empty",
                         nXOff, nYOff, nXSize, nYSize, nMaskFlagStop,
                         pdfDataPct, iX, iY);
            } else {
                bHasData = true;
#if 1
                CPLDebug("QCT", "IGetDataCoverageStatus(%d, %d, %d, %d, 0%o, %p) block %d %d exists at 0x%llx",
                         nXOff, nYOff, nXSize, nYSize, nMaskFlagStop,
                         pdfDataPct, iX, iY, nTilePos );
#endif
            }
            if( bHasData ) {
                nPixelsData += (MIN( (iX + 1) * nBlockXSize, nXOff + nXSize ) - MAX( iX * nBlockXSize, nXOff )) *
                    (MIN( (iY + 1) * nBlockYSize, nYOff + nYSize ) - MAX( iY * nBlockYSize, nYOff ));
                nStatus |= GDAL_DATA_COVERAGE_STATUS_DATA;
            }
            if( nMaskFlagStop != 0 && (nMaskFlagStop & nStatus) != 0 ) {
                CPLDebug("QCT", "IGetDataCoverageStatus(%d, %d, %d, %d, 0%o, %p) early finish for flags at block %d %d",
                         nXOff, nYOff, nXSize, nYSize, nMaskFlagStop,
                         pdfDataPct, iX, iY);
                return nStatus;
            }
        }
    }
        
    if (pdfDataPct!=nullptr) {
        *pdfDataPct =
            (100.0 * static_cast<double>(nPixelsData)) /
            (static_cast<double>(nXSize) * nYSize);

        CPLDebug("QCT", "IGetDataCoverageStatus(%d, %d, %d, %d, %d, %p) returns %d with %lf%% coverage",
                 nXOff, nYOff, nXSize, nYSize, nMaskFlagStop,
                 pdfDataPct, nStatus, *pdfDataPct);
    } else {
        CPLDebug("QCT", "IGetDataCoverageStatus(%d, %d, %d, %d, %d, nil) returns %d without setting coverage",
                 nXOff, nYOff, nXSize, nYSize, nMaskFlagStop,
                 nStatus);
    }
    return nStatus;
}
#endif // GDAL_VERSION_NUM >= 2020000

/************************************************************************/
/*                       GetColorInterpretation()                       */
/************************************************************************/

GDALColorInterp QCTRasterBand::GetColorInterpretation()
{
    return GCI_PaletteIndex;
}


/************************************************************************/
/*                           GetColorTable()                            */
/************************************************************************/

GDALColorTable *QCTRasterBand::GetColorTable()

{
    if( nBand == 1 )
        return static_cast<QCTDataset *>(poDS)->poColorTable;
    else
        return nullptr;
}

/************************************************************************/
/*                          GetOverviewCount()                          */
/************************************************************************/

int QCTRasterBand::GetOverviewCount()

{
    int nResult = static_cast<QCTDataset *>(poDS)->nOverviewCount;
#if 0
    if( nOverviewCount <= -1 ) {
        nResult = GDALRasterBand::GetOverviewCount();
        CPLDebug("QCT",
                 "GDALRasterBand::GetOverviewCount() returns %d",
                 nResult);
    }
#endif
    CPLDebug("QCT",
             "%p->QCTRasterBand::GetOverviewCount() returns %d",
             this, nResult);
    return nResult;
} // QCTRasterBand::GetOverviewCount

/************************************************************************/
/*                            GetOverview()                             */
/************************************************************************/

GDALRasterBand *QCTRasterBand::GetOverview( int iOverview )
{
    QCTDataset *poQDS = static_cast<QCTDataset *>(poDS);
    QCTRasterBand *result = nullptr;
    if( // (poQDS->papoOverviewBands!=nullptr) &&
        poQDS->nOverviewCount > 0 )
    {
        if( iOverview < 0 || iOverview >= poQDS->nOverviewCount )
            result = nullptr;
        else
            result = poQDS->papoOverviewBands[iOverview];
    }
    if (this == result) {
        // overviews should not have overviews
        CPLDebug("QCT", "overview[%d] is overview of band %p",
                 iOverview, this);
        result = nullptr;
    }
    
    CPLDebug("QCT",
             "%p->QCTRasterBand::GetOverview(%d) returns %p",
             this, iOverview, result);
    return static_cast<GDALRasterBand*>(result);
} // QCTRasterBand::GetOverview


/************************************************************************/
/* ==================================================================== */
/*                            QCTDataset                                */
/* ==================================================================== */
/************************************************************************/


/************************************************************************/
/*                            QCTDataset()                              */
/************************************************************************/

QCTDataset::QCTDataset() :
    fp(nullptr),
    poColorTable(nullptr),
    // anExtended({0}),
    papoOverviewBands{nullptr},
#ifdef CODE_ANALYSIS
    // These are only here to keep cppcheck happy
    // They may make the program think things are OK when they are not.
    magic(0),
    version(0),
#endif // CODE_ANALYSIS
    pszLongTitle(nullptr),
    pszName(nullptr),
    pszIdentifier(nullptr),
    pszEdition(nullptr),
    pszRevision(nullptr),
    pszKeywords(nullptr),
    pszCopyright(nullptr),
    pszScale(nullptr),
    pszDatum(nullptr),
    pszDepths(nullptr),
    pszHeights(nullptr),
    pszProjection(nullptr),
    pszOrigFileName(nullptr),
    nQCTFlags(0x0),
    nOrigFileSize(0),
    nReserved1(0),
    nOutlinePts(0),
    nOrigFileTime(0),
#if 1
    // adfQCTdatumShift({}), // works on g++ but not clang++
    adfQCTdatumShift{0.0, 0.0}, // works on g++789 and clang++7 but not clang++3
#endif

#if 0
    TL{INT_MIN,INT_MAX}, // lat,lon
    TR{INT_MIN,INT_MIN},
    BL{INT_MAX,INT_MAX},
    BR{INT_MAX,INT_MIN},
#endif
    adfOutlinePoints(nullptr),
    nQCTfilesize(0),
    dScalePixels(1.0),
#ifdef QCT_USE_GCP
    nGCPCount(0),
    pasGCPList(nullptr),
#endif
    poQC3_header(nullptr),
    anTileIndex(nullptr),
    nOverviewCount(0),
    acInterpolationMatrix{{0}}
{
#if 1
    adfQCTdatumShift[0] = adfQCTdatumShift[1] = 0.0;
    {
        for (int i=0;i<8;i++) {
            anExtended[i]=0;
        }
        for (int i=0;i<40;i++) {
            QCTpolynomial[i]=0.0;
        }
    }
    TL.lat=INT_MIN; TL.lon=INT_MAX;
    TR.lat=INT_MIN; TR.lon=INT_MIN;
    BL.lat=INT_MAX; BL.lon=INT_MAX;
    BR.lat=INT_MAX; BR.lon=INT_MIN;
#endif

} // QCTDataset::QCTDataset()

/************************************************************************/
/*                           ~QCTDataset()                             */
/************************************************************************/

QCTDataset::~QCTDataset()

{
    FlushCache();

#ifdef QCT_USE_GCP
    if(pasGCPList != nullptr) {
        if( nGCPCount > 0 ) {
            CPLDebug("QCT", "GDALDeinitGCPs( %d %p)", nGCPCount, pasGCPList ); 
            GDALDeinitGCPs( nGCPCount, pasGCPList );
            CPLDebug("QCT", "GDALDeinitGCPs( %d %p) completed", nGCPCount, pasGCPList );
            VSIFree( pasGCPList );
            pasGCPList = nullptr;
            nGCPCount = 0;
        } else {
            CPLDebug("QCT", "GDALDeinitGCPs( %d %p) 0 nonnullptr", nGCPCount, pasGCPList );
        }
    }
#endif

    if( fp != nullptr ) {
        VSIFCloseL( fp );
    }
    if( poColorTable != nullptr ) {
        delete ( poColorTable );
    }

    if( adfOutlinePoints != nullptr ) {
        VSIFree ( adfOutlinePoints );
        adfOutlinePoints = nullptr;
        nOutlinePts = 0;
    }
    if (anTileIndex != nullptr ) {
        VSIFree(anTileIndex);
        anTileIndex = nullptr;
    }
    if( pszLongTitle != nullptr ){
        VSIFree ( pszLongTitle );
        pszLongTitle = nullptr;
    }
    if( pszName != nullptr ){
        VSIFree ( pszName );
        pszName = nullptr;
    }
    if( pszIdentifier != nullptr ){
        VSIFree ( pszIdentifier );
        pszIdentifier = nullptr;
    }
    if( pszEdition != nullptr ){
        VSIFree ( pszEdition );
        pszEdition = nullptr;
    }
    if( pszRevision != nullptr ){
        VSIFree ( pszRevision );
        pszRevision = nullptr;
    }
    if( pszKeywords != nullptr ){
        VSIFree ( pszKeywords );
        pszKeywords = nullptr;
    }
    if( pszCopyright != nullptr ){
        VSIFree ( pszCopyright );
        pszCopyright = nullptr;
    }
    if( pszScale != nullptr ){
        VSIFree ( pszScale );
        pszScale = nullptr;
    }
    if( pszDepths != nullptr ){
        VSIFree ( pszDepths );
        pszDepths = nullptr;
    }
    if( pszHeights != nullptr ){
        VSIFree ( pszHeights );
        pszHeights = nullptr;
    }
    if( pszProjection != nullptr ){
        VSIFree ( pszProjection );
        pszProjection = nullptr;
    }
    if( pszOrigFileName != nullptr ){
        VSIFree ( pszOrigFileName );
        pszOrigFileName = nullptr;
    }    
    if( pszDatum != nullptr ){
        VSIFree ( pszDatum );
        pszDatum = nullptr;
    }
    if( poQC3_header != nullptr ){
        delete poQC3_header;
        poQC3_header = nullptr;
    }
    for (int i=0; i<nOverviewCount; i++) {
        if (papoOverviewBands[i]) {
            CPLDebug("QCT",
                     // "*not*"
                     " deleting papoOverviewBands[%d] %p",
                     i, papoOverviewBands[i]);
            delete papoOverviewBands[i];
            papoOverviewBands[i] = nullptr;
        }
    }
    
} // QCTDataset::~QCTDataset()


/************************************************************************/
/*                            GetGCPCount()                             */
/************************************************************************/
#ifdef QCT_USE_GCP
int QCTDataset::GetGCPCount()
{
    // CPLDebug("QCT", "QCTDataset::GetGCPCount() returns %d", nGCPCount);
    return nGCPCount;
} // QCTDataset::GetGCPCount
#endif // QCT_USE_GCP

/************************************************************************/
/*                          GetGCPProjection()                          */
/************************************************************************/
#ifdef QCT_USE_GCP
const char *QCTDataset::GetGCPProjection()
{
    if( nGCPCount > 0 ) {
        // CPLDebug("QCT", "QCTDataset::GetGCPProjection() returns %s",
        //         GetProjectionRef());
        return GetProjectionRef();
        
    } else {
        CPLDebug("QCT", "QCTDataset::GetGCPProjection() returns null string");
        return "";
    }
} // QCTDataset::GetGCPProjection
#endif // QCT_USE_GCP

/************************************************************************/
/*                               GetGCP()                               */
/************************************************************************/
#ifdef QCT_USE_GCP
const GDAL_GCP *QCTDataset::GetGCPs()

{
    // CPLDebug("QCT", "QCTDataset::GetGCPs() returns %p", pasGCPList);
    return pasGCPList;
}
#endif // QCT_USE_GCP

/************************************************************************/
/*                          GetGeoTransform()                           */
/************************************************************************/
// #if 1 //ndef  QCT_USE_GCP
#ifndef QCT_USE_GCP
CPLErr QCTDataset::GetGeoTransform( double * padfTransform )
{
    // Xgeo = padfTransform[0] + pixel*padfTransform[1] + line*padfTransform[2];
    // Ygeo = padfTransform[3] + pixel*padfTransform[4] + line*padfTransform[5];

    padfTransform[0] = TL.lon;
    padfTransform[1] = (TR.lon - TL.lon) / GetRasterXSize();
    padfTransform[2] = (BL.lon - TL.lon) / GetRasterYSize();
    padfTransform[3] = TL.lat;
    padfTransform[4] = (TR.lat - TL.lat) / GetRasterXSize();
    padfTransform[5] = (BL.lat - TL.lat) / GetRasterYSize();

    return CE_None;
}
#endif // ifndef  QCT_USE_GCP

/************************************************************************/
/*                          GetProjectionRef()                          */
/************************************************************************/

// Gdal <3 uses proj.4, Gdal>=3 uses proj.6, see eg:
// https://trac.osgeo.org/gdal/wiki/rfc73_proj6_wkt2_srsbarn
// https://gdal.org/development/rfc/rfc73_proj6_wkt2_srsbarn.html
#if GDAL_VERSION_MAJOR >= 3
const char *QCTDataset::_GetProjectionRef
#else
const char *QCTDataset::GetProjectionRef
#endif
(){
#if 1
    CPLDebug("QCT", "QCTDataset::GetProjectionRef() returns %s",
             pszDatum);
    return pszDatum;
#else
    return GDALDataset::GetProjectionRef();
#endif
}

/************************************************************************/
/*                              Identify()                              */
/************************************************************************/

int QCTDataset::Identify( GDALOpenInfo * poOpenInfo )

{
/* -------------------------------------------------------------------- */
/*         This has to be a file on disk ending in                      */
/*                  .qct (or perhaps .qc3)                              */
/*                with the right magic number                           */
/* -------------------------------------------------------------------- */

    CPLDebug("QCT", "QCTDataset::Identify(%p)\n", poOpenInfo);

    VSIStatBufL psRscStatBuf={};
    if ( VSIStatL( poOpenInfo->pszFilename, &psRscStatBuf ) != 0 ) {
            return FALSE;
    }

    // .qct 0x60 bytes meta data
    // .qc3 files 40+8+w*h + data bytes

    // Need to read
    // http://trac.osgeo.org/gdal/wiki/rfc11_fastidentify

    if( poOpenInfo->nHeaderBytes < 0x60 ) {
        // http://lists.osgeo.org/pipermail/gdal-dev/2013-February/035530.html
        // suggests that file extension is a bad way to detect file format
        // but if the header is not present we don't have a choice

        if( EQUAL(CPLGetExtension(poOpenInfo->pszFilename),"qct") ||
            EQUAL(CPLGetExtension(poOpenInfo->pszFilename),"qc3") )
            {
                CPLError( CE_Failure, CPLE_AppDefined, 
                          "QCT identify given %d byte header - needs 0x60 (file %s)",
                          poOpenInfo->nHeaderBytes, poOpenInfo->pszFilename );
            }
        return FALSE;
    }

    CPLString osPathname = CPLGetPath( poOpenInfo->pszFilename );
    CPLString osBasename = CPLGetBasename( poOpenInfo->pszFilename );
    CPLString osQCTFilename = CPLFormFilename( osPathname, osBasename, "qct" );
    CPLString osQC3Filename = CPLFormFilename( osPathname, osBasename, "qc3" );

    if ( VSIStatL( osQCTFilename, &psRscStatBuf ) != 0 ) {
        osQCTFilename = "";
    }
    if ( VSIStatL( osQC3Filename, &psRscStatBuf ) != 0 ) {
        osQC3Filename = "";
    }

    unsigned int magic = QCTGetUInt(poOpenInfo->pabyHeader, 0);
    unsigned int version = QCTGetUInt(poOpenInfo->pabyHeader, 4);

    if( magic == qcthree_magic ) {
        // This file has the image but not the palette or geo data.

        if( osQCTFilename[0] ) {
            // CPLDebug("QCT",
            CPLError( CE_Failure, CPLE_AppDefined,
                      "qc3 file %s not yet supported; try the .qct file %s",
                      poOpenInfo->pszFilename,
                      static_cast<const char*>(osQCTFilename) );
        } else {
            // CPLDebug("QCT",
            CPLError( CE_Failure, CPLE_AppDefined,
                      "qc3 file %s not yet supported and file %s%s.qct is missing",
                      poOpenInfo->pszFilename,
                      static_cast<const char*>(osPathname),
                      static_cast<const char*>(osBasename) );
        }
        return FALSE;
    } else if( magic == qct_magic || magic == qct_info_magic ) {
        if ( (magic == qct_magic && (version == 2 || version == 4))
             || (magic == qct_info_magic && version == 1) ) 
        {
            // This file has the image, the palette and the geo data.
            return TRUE;
        } else if ((version & qct_version_for_qc3_mask) == qct_version_for_qc3) {
            // This file has the palette and geo data
            // - the image is in another file.
            if( osQC3Filename.c_str() && osQC3Filename.c_str()[0] ) {
                CPLDebug("QCT", "qct/qc3 file pair %s %s limited support",
                         poOpenInfo->pszFilename,
                         static_cast<const char*>(osQC3Filename) );
                return TRUE;
            } else {
                // CPLDebug("QCT",
                CPLError( CE_Failure, CPLE_AppDefined,
                          "%s needs missing file %s%s.qc3",
                          poOpenInfo->pszFilename,
                          static_cast<const char*>(osPathname),
                          static_cast<const char*>(osBasename) );
                // We can still get position data from the .qct file
                // So we will say yes for now. This may be a mistake.
                return TRUE;
                // return FALSE;
            }
        } else {
            CPLDebug("QCT ", "unexpected QCT version %08x", version);
            return TRUE;
        }
    }

    return FALSE;
} // QCTDataset::Identify



/************************************************************************/
/*                              QCTGetTileIndex()                       */
/************************************************************************/

long long QCTDataset::QCTGetTileIndex(unsigned int x, unsigned int y)
{
    if ( anTileIndex == nullptr) {
        CPLDebug("QCT", "building TileIndex");
        anTileIndex = QCTBuildTileIndex();
    
        if ( anTileIndex == nullptr) {
            CPLError( CE_Failure, CPLE_AppDefined, 
                      "QCT failed to build tile index");
#if 0
            VSIFree(QCTpalette);
            delete poDS;
            if (pszQCTDatum) {
                VSIFree(pszQCTDatum);
                pszQCTDatum = nullptr;
            }
#endif
            return 0LL;
        }
    }

    unsigned int nTileIndexWidth, nTileIndexHeight;
    if (bHaveQC3data()) {
        nTileIndexWidth = poQC3_header->width3;
        nTileIndexHeight = poQC3_header->height3;
    } else {
        nTileIndexWidth = static_cast<unsigned long long>(nRasterXSize)/QCT_TILE_SIZE;
        nTileIndexHeight = static_cast<unsigned long long>(nRasterYSize)/QCT_TILE_SIZE;
    }
    if (x>=nTileIndexWidth) {
        return 0LL;
    }
    if (y>=nTileIndexHeight) {
        return 0LL;       
    }
    CPLDebug("QCT",
             // "QCTGetTileIndex(%u,%u) returns %lld",
             "QCTGetTileIndex(%u,%u) returns 0x%012llx",
             x, y, anTileIndex[x + y*nTileIndexWidth]);
    return anTileIndex[x + y*nTileIndexWidth];
} // QCTGetTileIndex(x,y)


/************************************************************************/
/*                            QCTBuildTileIndex()                       */
/************************************************************************/

long long* QCTDataset::QCTBuildTileIndex( )
{
    // The tile index is in the same file as the data;
    // for .qct/.qc3 file pairs we need the .qc3 file.
    VSILFILE *fpTI;
    int nTileIndexWidth, nTileIndexHeight;
    vsi_l_offset nTileIndexStart, nTileIndexEnd, nFileEnd;

    if (bNeedQC3data()) {
        if (bHaveQC3data()) {
            fpTI=poQC3_header->fpQC3;
            nTileIndexWidth = poQC3_header->width3;
            nTileIndexHeight = poQC3_header->height3;
            nTileIndexStart = nQC3_TileIndexStart;
            nTileIndexEnd   = nTileIndexStart + 8*nTileIndexWidth*nTileIndexHeight;
            nFileEnd        = poQC3_header->nQC3filesize;
        } else {
            CPLError( CE_Failure, CPLE_AppDefined, 
                      "QCT failed to build tile index");
            return nullptr;
        }
    } else {
        fpTI=fp;
        nTileIndexWidth = static_cast<unsigned long long>(nRasterXSize)/QCT_TILE_SIZE;
        nTileIndexHeight = static_cast<unsigned long long>(nRasterYSize)/QCT_TILE_SIZE;
        nTileIndexStart = nQCT_TileIndexStart;
        nTileIndexEnd   = nTileIndexStart + 4*nTileIndexWidth*nTileIndexHeight;
        nFileEnd        = nQCTfilesize;
    }
    
    if ( VSIFSeekL( fpTI, static_cast<size_t>(nTileIndexStart), SEEK_SET ) ) {
        CPLError( CE_Failure, CPLE_AppDefined,
                  "cannot seek to QCT tile index" );
        return nullptr;
    }
    // location seek 0x%lx failed. rasterXsize %d blockXsize %d",
    //          nTileIndexIndex, poQDS->nRasterXSize, nBlockXSize);

    long long *anNewTileIndex = static_cast<long long *>
        (VSIMalloc3(sizeof (size_t),
                    static_cast<size_t>(nTileIndexWidth),
                    static_cast<size_t>(nTileIndexHeight)) );
    if (anNewTileIndex == nullptr) {
        CPLError(CE_Failure, CPLE_OutOfMemory,
                 "Cannot allocate memory for tile index");
        return nullptr;
    }

    // Read Tile Index into memory
    long long q=0;
    long long corrupt=0;
    unsigned long long nMaxTile=0, nMaxTileIndex=0;
    for (int j=0; j<nTileIndexHeight; j++) {
        for (int i=0; i<nTileIndexWidth; i++) {
            unsigned long long nValue;
            if (bHaveQC3data()) {
                nValue = static_cast<unsigned long long>(QCTreadUInt64(fpTI));
            } else {
                nValue = static_cast<unsigned long long>(QCTreadUInt(fpTI));
                if (nValue==static_cast<unsigned long>(-1L)) {
                    nValue=MINUSONE_as_SIZET;
                }
            }
            // Ignore the index if it points
            // outside the limits of the data in the file
            // For QC3:
            // 0 means the tile is missing and may be streamed if possible
            // "-1" means the tile is not defined and does not exist.
            if (nValue==MINUSONE_as_SIZET
                || nValue < nTileIndexEnd || nValue >= nFileEnd)
            {
#if 0
                CPLDebug("QCT",
                         "anNewTileIndex[%lld] (%d %d) addr 0x%016llx not in file",
                         q, i, j, nValue);
#endif
                if (nValue!=MINUSONE_as_SIZET && nValue!=0) {
                    // Corrupt data
                    corrupt++;
#if 0
                    // Is the previous tile truncated too?
                    if (nValue >= nFileEnd && q>0) {
                        long long nPrev = anNewTileIndex[q-1];
                        if ((nPrev-nTileIndexEnd)*(q+1) > (nFileEnd-nTileIndexEnd)*q) {
                            // Prev tile might be corrupt too:
                            anNewTileIndex[q-1] = MINUSONE_as_SIZET;
                        }
                    }
#endif
                    nValue = MINUSONE_as_SIZET;
                }
            } else if (nValue > nMaxTile) {
                nMaxTile = nValue;
                nMaxTileIndex=q;
            }
#if 0            
            CPLDebug("QCT",
                     "setting anNewTileIndex[%lld] (%d %d) to %lld=x%08llx",
                     q, i, j, static_cast<signed long long>(nValue), nValue);
#endif
            anNewTileIndex[q++] = static_cast<size_t>(nValue);
        }
    }
    CPLDebug("QCT", "Last tile %lld starts at %lld=%016llx",
             nMaxTileIndex, nMaxTile, nMaxTile);
    if (corrupt>0) {
        CPLDebug("QCT", "found %lld corrupted tile pointers", corrupt);
        // Assume the last tile that is present is corrupt
        // ... ie assume it has been truncated.
        if (nMaxTile>0) {
            anNewTileIndex[nMaxTileIndex] = MINUSONE_as_SIZET;
        }
    }

    return anNewTileIndex;
} // QCTDataset::QCTBuildTileIndex()

void QCTDataset::QCTprintExtendedData(void)
{
    CPLDebug("QCT", "extended data at %p", anExtended);
    for (int i=0; i<8; i++) {
        CPLDebug("QCT", "extended data[%02x] = 0x%08x",
                 i*4, anExtended[i]);
    }
    if (anExtended[0]) {
        char* pszMapType = QCTGetString(anExtended[0]);
        if (pszMapType) {
            CPLDebug("QCT", "Map Type: %s",  pszMapType);
            VSIFree(pszMapType);
        }
    }
    // anExtended[1] Pointer to Array Datum Shift (see section 6.2.3)
    if (anExtended[2]) {
        if (char* pszDiskName = QCTGetString(anExtended[2]) ) {
            CPLDebug("QCT", "Original Disk Name: %s",  pszDiskName);
            VSIFree(pszDiskName);                    
         }
    }
    if (anExtended[3]) {
        CPLDebug("QCT", "Extended data reserved1 = 0x%08x", anExtended[3]);
    }
    if (anExtended[4]) {
        CPLDebug("QCT", "Extended data reserved2 = 0x%08x", anExtended[4]);
    }
    if (anExtended[5] && anExtended[5]+0x14<nQCTfilesize) {
        // [5] Pointer to Struct License Information
        // (Optional, see section 6.2.4)
         unsigned int nLicenseID = QCTreadInt(fp,anExtended[5]);
         unsigned int nLicReserved1 = QCTreadInt(fp, anExtended[5]+0x4);
         unsigned int nLicReserved2 = QCTreadInt(fp, anExtended[5]+0x8);
         unsigned int addLicenseDesc = QCTreadInt(fp, anExtended[5]+0xC);
         unsigned int addLicenseSerialNo = QCTreadInt(fp, anExtended[5]+0x10);
         unsigned int nLicReserved3 = QCTreadInt(fp, anExtended[5]+0x14);
         char* pszLicenseDesc;
         if(addLicenseDesc && (pszLicenseDesc = ::QCTGetString(fp, addLicenseDesc)) )
         {
             CPLDebug("QCT", "LicenceID: %08X %s",
                      nLicenseID, pszLicenseDesc);
             VSIFree(pszLicenseDesc);                    
         } else {
             CPLDebug("QCT", "LicenceID: %08X", nLicenseID);
         }
         (void)nLicenseID; // Appease cppcheck -DCODE_ANALYSIS
         if (addLicenseSerialNo && addLicenseSerialNo+32<nQCTfilesize) {
             unsigned char aby[32];
             int seekres =
                 VSIFSeekL( fp,
                            static_cast<vsi_l_offset>(addLicenseSerialNo),
                            SEEK_SET );
             if ( seekres ) {
                 CPLError( CE_Failure, CPLE_AppDefined,
                           "cannot seek to QCT Licence serial number");
                 exit(-1);
                 //return nullptr;
             }
             for (int i=0; i<32; i++) {
                 aby[i] = QCTFGetuc(fp);
             }
             (void)aby; // Appease static analyzers
             CPLDebug("QCT", "Licence Serial Number:\t%02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x",
                      aby[0],aby[1],aby[2], aby[3],  aby[4], aby[5], aby[6], aby[7],
                      aby[8],aby[9],aby[10],aby[11], aby[12],aby[13],aby[14],aby[15]
                      );
             CPLDebug("QCT", "\t\t\t\t%02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x",
                      aby[16],aby[17],aby[18],aby[19], aby[20],aby[21],aby[22],aby[23],
                      aby[24],aby[25],aby[26],aby[27], aby[28],aby[29],aby[30],aby[31]
                      );
         } else {
             CPLDebug("QCT", "Licence has no Serial Number");
         }
         CPLDebug("QCT", "Licence, reserved values: 0x%08x 0x%08x 0x%08x",
                  nLicReserved1,nLicReserved2,nLicReserved3);
         // Appease cppcheck -DCODE_ANALYSIS
         (void)nLicReserved1; (void)nLicReserved2; (void)nLicReserved3;
         
         // x18-x27 16 bytes, unknown, set to 0
         // x28-x67 64 bytes, unknown
    }
    if (anExtended[6]) {
        if (char* pszAssociatedData = ::QCTGetString(fp,anExtended[6])) {
            CPLDebug("QCT", "AssociatedData: %s", pszAssociatedData);
            VSIFree(pszAssociatedData);
        }
    }
    if (anExtended[7]) {
        // [7] Pointer to Struct Digital Map Shop
        // (Optional, see section 6.2.5)
        int nDMSsize = QCTreadInt(fp, anExtended[7]);
        if (8!=nDMSsize) {
            CPLDebug("QCT", "Expected 8 bytes for Digital Map Shop: found %d",
                     nDMSsize);
        }
        if (8<=nDMSsize) {
            int addDMS_URL = QCTreadInt(fp);
            char* pszDMS_URL;
            if (addDMS_URL && (pszDMS_URL = ::QCTGetString(fp, addDMS_URL)) ) {
                CPLDebug("QCT", "Digital Map Shop URL: %s",
                         pszDMS_URL);
                // poDS not defined :-(
                // poDS->SetMetadataItem( "QC3 URL", pszDMS_URL);
                VSIFree(pszDMS_URL);
            }
        }        
    }

} // QCTDataset::QCTprintExtendedData



/************************************************************************/
/*                   QCTsetGCPMetaData_from_coefficients()                 */
/************************************************************************/

#ifdef QCT_USE_GCP
int QCTDataset::QCTsetGCPMetaData_from_coefficients(
                       int count, const char* szMDItem, double values[], const char* szCollection)
{
    if (count<0 || count>30) {
        CPLDebug("QCT", "not setting metadata; silly number %d of coefficients given", count);
        return -1;
    }
    char szValue[1280]={};
    size_t nCurPos=0;
    for(int i = 0; i < count; i++ ) {
        if (sizeof(szValue)-nCurPos<20) {
            CPLDebug("QCT", "not enough space for %s[%d]", szMDItem, i);
            return -1;
        }

        int nRes=snprintf( szValue+nCurPos, 20, "%.16g ", values[i] );
        if (nRes<0) {
            CPLDebug("QCT", "error writing %s[%d]", szMDItem, i);
            return -1;
        }
        nCurPos +=nRes;
    }
    if (nCurPos>=sizeof(szValue)) {
        szValue[sizeof(szValue)-1] = '\0';
    } else {
        szValue[nCurPos] = '\0';
    }
    this->SetMetadataItem( szMDItem, szValue, szCollection );

    return 0;
}
#endif // QCT_USE_GCP

/************************************************************************/
/*                                Open()                                */
/************************************************************************/

GDALDataset *QCTDataset::Open( GDALOpenInfo * poOpenInfo )

{
    CPLDebug("QCT", "QCTDataset::Open(%p)\n", poOpenInfo);
    
    char *pszQCTDatum = nullptr;

    if (!Identify(poOpenInfo))
        return nullptr;

#ifdef QCT_USE_GCP
    CPLDebug("QCT", "QCT with GCP");
#else
    CPLDebug("QCT", "QCT without GCP");
#endif

#if GDAL_VERSION_MAJOR >= 2
    /* Check that the file pointer from GDALOpenInfo* is available */
    if( poOpenInfo->fpL == nullptr )
    {
        return nullptr;
    }
#endif

/* -------------------------------------------------------------------- */
/*      Confirm the requested access is supported.                      */
/* -------------------------------------------------------------------- */
    if( poOpenInfo->eAccess == GA_Update )
    {
        CPLError( CE_Failure, CPLE_NotSupported, 
                  "The QCT driver does not support update access to existing"
                  " datasets.\n" );
        return nullptr;
    }

/* -------------------------------------------------------------------- */
/*      Create a corresponding GDALDataset.                             */
/* -------------------------------------------------------------------- */
    QCTDataset *poDS;

    poDS = new QCTDataset();

#if 0 // GDAL_VERSION_MAJOR >= 2
    /* Borrow the file pointer from GDALOpenInfo* */
    poDS->fp = poOpenInfo->fpL;
    poOpenInfo->fpL = nullptr;
#else
    poDS->fp = VSIFOpenL( poOpenInfo->pszFilename, "rb" );
    if (poDS->fp == nullptr) {
        CPLDebug("QCT", "Could not open QCT file %s",
                 poOpenInfo->pszFilename);
        delete poDS;
        return nullptr;
    }
#endif
    VSIStatBufL osRscStatBuf={};
    if ( VSIStatL( poOpenInfo->pszFilename, &osRscStatBuf ) != 0 ) {
        CPLDebug("QCT", "No QCT file %s",
                 poOpenInfo->pszFilename);
        delete poDS;
        return nullptr;
    }
    poDS->nQCTfilesize=osRscStatBuf.st_size;
    CPLDebug("QCT", "QCT file %s is 0x%zx=%zu bytes",
             poOpenInfo->pszFilename, poDS->nQCTfilesize, poDS->nQCTfilesize
             );
    
/* -------------------------------------------------------------------- */
/*      Read the header.                                                */
/* -------------------------------------------------------------------- */
    VSIFReadL( poDS->abyHeader, 1, 0x5a0, poDS->fp);

    poDS->magic = QCTGetUInt(poOpenInfo->pabyHeader, 0);
    poDS->version = QCTGetUInt(poOpenInfo->pabyHeader, 4);
    if (poDS->version == 4) {
        CPLError( CE_Warning, CPLE_AppDefined, 
                  // "QCT licence management info not available for %s - might be scrambled",
                  "QCT licence management for %s not supported; image might be scrambled",
                  poOpenInfo->pszFilename);
    }

    // .qct file (not .qc3?)
    if ( (poDS->magic==qct_magic || poDS->magic==qct_info_magic)
         && (poDS->version&qct_version_for_qc3_mask) == qct_version_for_qc3 )
    {
        poDS->poQC3_header = new QC3Header;
        
        if(!poDS->poQC3_header->Open(poOpenInfo->pszFilename,
                                     false, poDS->magic, poDS->version) ) {
            CPLDebug("QCT", "QC3 file not opened");
            delete poDS->poQC3_header;
            poDS->poQC3_header=nullptr;
        }
    } else {
        poDS->poQC3_header = nullptr;
    }
    
    if (poDS->poQC3_header) {
#if defined(QC3_PIXEL_IS_TILE)
        CPLDebug("QCT", "qc3_pixel_is_tile");
        poDS->nRasterXSize = static_cast<int>(poDS->poQC3_header->width3);
        poDS->nRasterYSize = static_cast<int>(poDS->poQC3_header->height3);
#else
        CPLDebug("QCT", "qc3_pixel_is_pixel");
        // Should these be width|height * QC3_TILE_SIZE ?
        poDS->nRasterXSize = static_cast<int>(poDS->poQC3_header->width3*QC3_TILE_SIZE);
        poDS->nRasterYSize = static_cast<int>(poDS->poQC3_header->height3*QC3_TILE_SIZE);
#endif
    } else {
        poDS->nRasterXSize = QCTGetInt( &poDS->abyHeader, 8 ) * QCT_TILE_SIZE;
        poDS->nRasterYSize = QCTGetInt( &poDS->abyHeader, 0xC ) * QCT_TILE_SIZE;
    }
    if  (poDS->nRasterXSize <= 0 || poDS->nRasterYSize <= 0 ) {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Invalid dimensions : %d x %d", 
                  poDS->nRasterXSize, poDS->nRasterYSize); 
        delete poDS;
        return nullptr;
    }

    CPLDebug("QCT", "QCT file %s is 0x%zx=%zu bytes",
             poOpenInfo->pszFilename, poDS->nQCTfilesize, poDS->nQCTfilesize
             );

    poDS->pszLongTitle = poDS->QCTGetStringIndirect(0x10);
    poDS->pszName = poDS->QCTGetStringIndirect(0x14);
    poDS->pszIdentifier = poDS->QCTGetStringIndirect(0x18);
    if (nullptr==poDS->pszName) {
        // Name is often null
        CPLDebug("QCT", "qct file has no Name tag - using Identifier %s",
                 poDS->pszIdentifier);
        poDS->pszName = poDS->pszIdentifier;
    }
    poDS->pszEdition = poDS->QCTGetStringIndirect(0x1c);
    poDS->pszRevision = poDS->QCTGetStringIndirect(0x20);
    poDS->pszKeywords = poDS->QCTGetStringIndirect(0x24);
    poDS->pszCopyright = poDS->QCTGetStringIndirect(0x28);
    poDS->pszScale = poDS->QCTGetStringIndirect(0x2c);
    /* we do more with datum later */
    if (pszQCTDatum!=nullptr) {
        CPLDebug("QCT", "pszQCTDatum redefined");
        VSIFree(pszQCTDatum);
        pszQCTDatum = nullptr;
    }
    pszQCTDatum = poDS->QCTGetStringIndirect(0x30); // This is leak-prone
    poDS->pszDepths =  poDS->QCTGetStringIndirect(0x34);
    poDS->pszHeights = poDS->QCTGetStringIndirect(0x38);
    poDS->pszProjection = poDS->QCTGetStringIndirect(0x3C);
    poDS->nQCTFlags = QCTGetInt( &poDS->abyHeader, 0x40 );
    poDS->pszOrigFileName = poDS->QCTGetStringIndirect(0x44);
    poDS->nOrigFileSize = QCTGetInt( &poDS->abyHeader, 0x48 );
    poDS->nOrigFileTime = QCTGetInt( &poDS->abyHeader, 0x4C );
    poDS->nReserved1 = QCTGetInt( &poDS->abyHeader, 0x50 );
    unsigned int addrQCTextendedData = QCTGetUInt( poOpenInfo->pabyHeader, 0x54);
    poDS->nOutlinePts = QCTGetUInt( &poDS->abyHeader, 0x58 );
#if 1
    // poDS->nOutlinePts < 0 // not satisfied by a UInt
    if (poDS->nOutlinePts*16 >= poDS->nQCTfilesize ) {
        CPLError( CE_Failure, CPLE_AppDefined,
                  "Invalid number of outline points %du",
                  poDS->nOutlinePts);
        delete poDS;
        if (pszQCTDatum) {
            VSIFree(pszQCTDatum);
            pszQCTDatum = nullptr;
        }
        return nullptr;
    }
#endif
        
    unsigned int addrQCToutline =
        QCTGetUInt( &poDS->abyHeader, 0x5C);

/* -------------------------------------------------------------------- */
/*       Read the extended data                                         */
/* -------------------------------------------------------------------- */
    if ( VSIFSeekL( poDS->fp, addrQCTextendedData, SEEK_SET ) ) {
        CPLError( CE_Failure, CPLE_AppDefined,
                  "QCT extended data seek failed" );
        if (pszQCTDatum) {
            VSIFree(pszQCTDatum);
            pszQCTDatum = nullptr;
        }
        return nullptr;
    }

    for (unsigned int i=0; i<8; i++) {
        poDS->anExtended[i] = QCTreadUInt(poDS->fp);
    }


    poDS->QCTprintExtendedData();
    // Most of this block is duplicated in QCTprintExtendedData
    // since that has no poDS to set the metadata.
    unsigned int nLicIDadd = poDS->anExtended[5];
    if (nLicIDadd) {
        unsigned int nLicenseID = QCTreadInt(poDS->fp,nLicIDadd);
        char szId[9]={};
        snprintf(szId, sizeof(szId),"%08X", nLicenseID);
        szId[8]='\000';
        poDS->SetMetadataItem( "QC3_LicenseID", szId);
    }
    
    if (poDS->poQC3_header) {
        char szEncrypt3[4]={};
        snprintf(szEncrypt3, sizeof(szEncrypt3),
                 "%d",  poDS->poQC3_header->encrypt3);
        szEncrypt3[3]='\000';
        poDS->SetMetadataItem( "QC3_Encrypt_level", szEncrypt3);        
    }

    unsigned int nDMSadd = poDS->anExtended[7];
    if (nDMSadd) {
        // [7] Pointer to Struct Digital Map Shop
        // (Optional, see section 6.2.5)
        int nDMSsize = QCTreadInt(poDS->fp, nDMSadd);
        if (8!=nDMSsize) {
            CPLDebug("QCT", "Expected 8 bytes for Digital Map Shop: found %d",
                     nDMSsize);
        }
        if (8<=nDMSsize) {
            int addDMS_URL = QCTreadInt(poDS->fp);
            char* pszDMS_URL;
            if (addDMS_URL && (pszDMS_URL = poDS->QCTGetString(addDMS_URL)) ) {
                CPLDebug("QCT", "Digital Map Shop URL: %s",
                         pszDMS_URL);
                poDS->SetMetadataItem( "QC3_URL", pszDMS_URL);
                VSIFree(pszDMS_URL);
            }
        }        
    }
      
/* -------------------------------------------------------------------- */
/*        Read the outline points                                       */
/* -------------------------------------------------------------------- */
    if (poDS->nOutlinePts > 0) {
        poDS->adfOutlinePoints = 
            static_cast<double*>(VSIMalloc3(2,sizeof(double),poDS->nOutlinePts));
        if (poDS->adfOutlinePoints == nullptr) {
            CPLError(CE_Failure, CPLE_OutOfMemory,
                     "Cannot allocate space for outline points");
            delete poDS;
            if (pszQCTDatum) {
                VSIFree(pszQCTDatum);
                pszQCTDatum = nullptr;
            }
            return nullptr;
        }
        // Read outline data from file
        size_t entries_read =
            QCTreadDoubles( poDS->adfOutlinePoints, poDS->fp,
                            addrQCToutline, 2*poDS->nOutlinePts);
        if ( entries_read < 2*poDS->nOutlinePts )  {
            CPLError( CE_Failure, CPLE_AppDefined,
                      "error reading %d/2 outline points - got %ld entries",
                      2*poDS->nOutlinePts, entries_read );
            delete poDS;
            if (pszQCTDatum) {
                VSIFree(pszQCTDatum);
                pszQCTDatum = nullptr;
            }
            return nullptr;
        }
    } else {
        poDS->adfOutlinePoints = nullptr;
    }
      
      
/* -------------------------------------------------------------------- */
/*      Read the palette.                                               */
/* -------------------------------------------------------------------- */
    /* qct palette is 256 4-byte entries: blue, green, red, padding
     * entries 129-255 are all zero.
     */
    poDS->poColorTable = new GDALColorTable();
    unsigned char *QCTpalette = nullptr, *qct_pall;
    if (poOpenInfo->nHeaderBytes >= 0x5a0) {
        QCTpalette = static_cast<unsigned char*>(poDS->abyHeader) + 0x1a0;
    } else {
        QCTpalette = static_cast<unsigned char*>(VSIMalloc(1024));
        if (QCTpalette == nullptr) {
            CPLError(CE_Failure, CPLE_OutOfMemory,
                     "Cannot allocate palette buffer");
            delete poDS;
            if (pszQCTDatum) {
                VSIFree(pszQCTDatum);
                pszQCTDatum = nullptr;
            }
            return nullptr;
        }
        // Read palette data from file
        if ( VSIFSeekL( poDS->fp, 0x1a0, SEEK_SET ) ) {
            CPLError( CE_Failure, CPLE_AppDefined,
                      "cannot seek to QCT palette" );
            VSIFree(QCTpalette);
            delete poDS;
            if (pszQCTDatum) {
                VSIFree(pszQCTDatum);
                pszQCTDatum = nullptr;
            }
            return nullptr;
        }
        size_t palentries_read =
            VSIFReadL( QCTpalette, sizeof(char), 1024, poDS->fp );
        
        if ( palentries_read < 1024 )  {
            CPLError( CE_Failure, CPLE_AppDefined,
                      "error reading palette - got %zu entries", palentries_read );
            VSIFree(QCTpalette);
            delete poDS;
            if (pszQCTDatum) {
                VSIFree(pszQCTDatum);
                pszQCTDatum = nullptr;
            }
            return nullptr;
        }
    }
    qct_pall = QCTpalette;
    GDALColorEntry color={};
    for (int i=0; i<129; i++) {
        color.c3 = *qct_pall++;    // blue
        color.c2 = *qct_pall++;    // green
        color.c1 = *qct_pall++;    // red
        color.c4 = 255 ;           // set full alpha
        qct_pall++;                // padding
        
        poDS->poColorTable->SetColorEntry(i, &color);
    }
    color.c1 = color.c2 = color.c3 = color.c4 = 0;
    for (int i=129; i<256; i++) {
        // The QCT Format Spec v1.0.3 section 6.4 says
        // that entries from 129 onwards are zero
        // (I expected 128 ...)
        poDS->poColorTable->SetColorEntry(i, &color);
    }
    // Set a colour for drawing boxes and tiles that may be available online.
    color.c1 = color.c2 = color.c3 = 127; color.c4 = 255; // solid grey
    poDS->poColorTable->SetColorEntry(nQCTYesData, &color);
    color.c1 = color.c2 = color.c3 = 255; color.c4 = 127; // translucent white
    poDS->poColorTable->SetColorEntry(nQCTonlineData, &color);
        
    if (QCTpalette != nullptr &&
        QCTpalette != static_cast<unsigned char*>(poDS->abyHeader) + 0x1a0 )
    {
        // CPLDebug("QCT", "about to free QCTpalette %p", QCTpalette);
        VSIFree(QCTpalette);
        QCTpalette = nullptr;
    }

/* -------------------------------------------------------------------- */
/*      Read the Interpolation matrix.                                  */
/* -------------------------------------------------------------------- */
    /* The qct Interpolation matrix is a 128 * 128 array
     * indicating the color that results if two pixels are merged.
     * Symmetry of the matrix allows interchangeability
     * of the x and y variables.
     */
    if (size_t res=VSIFReadL(poDS->acInterpolationMatrix, 1, 128*128, poDS->fp)!= 128*128) {
        CPLError(CE_Failure, CPLE_FileIO,
                 "Cannot read InterpolationMatrix - got 0x%04zx=%zu of 16kB",
                 res, res);
        delete poDS;
        if (pszQCTDatum) {
            VSIFree(pszQCTDatum);
            pszQCTDatum = nullptr;
        }
        return nullptr;
    }
#if 0    
    for (int x=0; x<128; x++) {
        for (int y=0; y<128; y++) {
            CPLDebug("QCT",
                     "acInterpolationMatrix[%d,%d] = %d",
                     x, y,  poDS->acInterpolationMatrix[x][y]);
        }
    }
#endif

    /********************************************************************/

    if ( poDS->anTileIndex) {
        CPLDebug("QCT",
                 "QCTBuildTileIndex already exists: value %p",
                 poDS->anTileIndex
                 );
    }

    /********************************************************************/
    /*             Report some strings found in the file                */
    /********************************************************************/
    QCTreport("Long Title",poDS->pszLongTitle);
    QCTreport("Name",poDS->pszName);
    QCTreport("Identifier",poDS->pszIdentifier);
    QCTreport("Edition",poDS->pszEdition);
    QCTreport("Revision",poDS->pszRevision);
    QCTreport("Keywords",poDS->pszKeywords);
    QCTreport("Copyright",poDS->pszCopyright);
    QCTreport("Scale",poDS->pszScale);
    QCTreport("Datum",pszQCTDatum);
    QCTreport("Depths",poDS->pszDepths);
    QCTreport("Heights",poDS->pszHeights);
    QCTreport("Projection",poDS->pszProjection);
    CPLDebug("QCT", "Flags:      0x%x=0%o", poDS->nQCTFlags, poDS->nQCTFlags);
    CPLDebug("QCT", "\t\"must have original file\" flag %s",
             (poDS->nQCTFlags & 1) ? "set" : "clear");
    CPLDebug("QCT", "\t\"allow calibration\" flag %s",
             (poDS->nQCTFlags & 2) ? "set" : "clear");
    QCTreport("OrigFileName",poDS->pszOrigFileName);
    CPLDebug("QCT", "OrigFilesize: %d",poDS->nOrigFileSize);
    CPLDebug("QCT", "OrigFileTime: %ld",poDS->nOrigFileTime);
#if 0
    // Is this ok or do we need to allocate memory and use ctime_r ?
    CPLDebug("QCT", "OrigFileTime: %s", ctime(&poDS->nOrigFileTime) );
#endif
    CPLDebug("QCT", "%u Outline points",poDS->nOutlinePts);
    switch (poDS->nReserved1) {
    case 0:
        CPLDebug("QCT", "nReserved1 is 0 as expected"); break;
    case 0x4b571cea: // 1264000234 decimal, 1[89][024]0s_oxford.qct
        CPLDebug("QCT", "nReserved1: 0x%x is known but unusual.",
                 poDS->nReserved1); break;
    default:
        CPLDebug("QCT", "nReserved1: 0x%x is new to me.",
                 poDS->nReserved1); break;
    }

    /********************************************************************/
    /*                           Set datum                              */
    /********************************************************************/
    if ( pszQCTDatum ) {
        if (poDS->pszDatum) {
            CPLDebug("QCT", "poDS->pszDatum unexpectedly previously set to %p %s",
                     poDS->pszDatum, poDS->pszDatum);
            VSIFree(poDS->pszDatum);
            poDS->pszDatum=nullptr;
        }

        OGRSpatialReference oSRS;
        OGRErr WKGerr = oSRS.SetWellKnownGeogCS( pszQCTDatum );
        if (OGRERR_NONE == WKGerr) {
            oSRS.exportToWkt(&(poDS->pszDatum)); // Memory leak here ?
        } 
        /* OSRImportFromEPSG needs a number (not a string)
        else if (OGRERR_NONE == OSRImportFromEPSG( &oSRS, pszQCTDatum ) ) {
            CPLDebug("QCT", "EPSG datum %s found", pszQCTDatum);
            oSRS.exportToWkt(&(poDS->pszDatum));
        }
        */
        else {
            CPLDebug("QCT", "unknown datum %s ignored: %d",
                     pszQCTDatum, WKGerr);
            poDS->pszDatum = VSIStrdup(pszQCTDatum);
        }

        CPLDebug("QCT", "file datum %s parsed to %s",
                 pszQCTDatum, poDS->pszDatum);
    } else {
        CPLDebug("QCT", "null datum in file");
    }

    {
        unsigned int addQCTdatumShift = poDS->anExtended[4];
        size_t readret =
            QCTreadDoubles( poDS->adfQCTdatumShift, poDS->fp, addQCTdatumShift, 2);
        if ( readret <2 ) {
            CPLError(CE_Failure, CPLE_AppDefined,
                     "Cannot read datum shift - read %zu doubles\n", readret);
            delete poDS;
            if (pszQCTDatum) {
                VSIFree(pszQCTDatum);
                pszQCTDatum = nullptr;
            }
            return nullptr;
        }
        CPLDebug("QCT", "Datum shift %lg east, %lg north",
                 poDS->adfQCTdatumShift[1], poDS->adfQCTdatumShift[0]);    
    }

    /********************************************************************/
    /*                       Set coordinate model                       */
    /********************************************************************/
    /* try RPC domain
       http://www.gdal.org/gdal_datamodel.html
       http://geotiff.maptools.org/rpc_prop.html
       GDALExtractRPCInfo
           /home/src/redhat/BUILD/gdal-1.9.2-fedora/frmts/gtiff/geotiff.cpp
       GTiffDataset::ReadRPCTag()
       /home/src/redhat/BUILD/gdal-1.9.2-fedora/frmts/nitf/nitf_gcprpc.cpp

         C1 + C2.L + C3.P + C5.L.P              + C4.H
       + C8.L.L + C9.P.P                        + C6.L.H  + C7.P.H + C10.H.H
       + C12.L.L.L + C13.L.P.P  + C15.L.L.P     + C11.P.L.H + C14.L.H.H
       + C16.P.P.P                              + C17.P.H.H + C18.L.L.H + C19.P.P.H + C20.H.H.H
    */

    const size_t nPolynomialOffset = 0x60;
    if ( 40!=QCTreadDoubles( poDS->QCTpolynomial, poDS->fp, nPolynomialOffset, static_cast<size_t>(40) ) ) {
        CPLDebug("QCT", "problem reading coordinate polynomial");
        delete poDS;
        if (pszQCTDatum) {
            VSIFree(pszQCTDatum);
            pszQCTDatum = nullptr;
        }
        return nullptr;
    }    

#if defined QC3_PIXEL_IS_TILE
    if (poDS->poQC3_header) {
        poDS->dScalePixels=QC3_TILE_SIZE;
    }
#endif

#if defined QCT_USE_GCP
    CPLDebug("QCT",
             "Building GCPs for %d x %d image scaled %g",
             poDS->nRasterXSize, poDS->nRasterYSize, poDS->dScalePixels);

    // calculate corner coordinates
    poDS->xy_to_latlon(0.0,
                       0.0,
                       &poDS->TL);
    poDS->xy_to_latlon(static_cast<double>(poDS->nRasterXSize),
                       0.0,
                       &poDS->TR);
    poDS->xy_to_latlon(0.0,
                       static_cast<double>(poDS->nRasterYSize),
                       &poDS->BL);
    poDS->xy_to_latlon(static_cast<double>(poDS->nRasterXSize),
                       static_cast<double>(poDS->nRasterYSize),
                       &poDS->BR);

    CPLDebug("QCT", "nRaster.Size %d x %d",
             poDS->nRasterXSize, poDS->nRasterYSize);
    CPLDebug("QCT", "TL %.8g %.8g",
             poDS->TL.lon, poDS->TL.lat);
    CPLDebug("QCT", "TR %.8g %.8g",
             poDS->TR.lon, poDS->TR.lat);
    CPLDebug("QCT", "BL %.8g %.8g",
             poDS->BL.lon, poDS->BL.lat);
    CPLDebug("QCT", "BR %.8g %.8g",
             poDS->BR.lon, poDS->BR.lat);

    
#if 1 // Use CRS of map when available, rather than WGS84 used in file.

    // Memory Map UK maps are perhaps better treated as EPSG:27700
    // (and should try EPSG:5070/5071/5072 for USA maps too)

    int nEPSG=0;
    if (poDS->pszCopyright) {
        nEPSG = EPSG_from_Copyright(poDS->pszCopyright);
    }
    if (nEPSG) {
        CPLDebug("QCT", "using EPSG:%d",nEPSG);

        if (poDS->pszDatum) {
            CPLDebug("QCT",
                     "replacing datum %s with EPSG:...",
                     poDS->pszDatum);
            VSIFree ( poDS->pszDatum );
            poDS->pszDatum=nullptr;
        }

        if (poDS->pszProjection) {
            CPLDebug("QCT",
                     "replacing projection %s with tmerc",
                     poDS->pszProjection);
            VSIFree ( poDS->pszProjection );
        }
        poDS->pszProjection= VSIStrdup("tmerc");
                
        // need to convert TL, TR, BL and BR from WGS84 to EPSG...
        OGRSpatialReference *poSourceSRS;
#if GDAL_VERSION_MAJOR >= 3
        // proj6/7
        // poSourceSRS = new OGRSpatialReference(SRS_DN_WGS84);
        poSourceSRS = new OGRSpatialReference(SRS_WKT_WGS84_LAT_LONG);
        poSourceSRS->SetAxisMappingStrategy(OAMS_TRADITIONAL_GIS_ORDER);
#else
        // proj4 ?5
        poSourceSRS = new OGRSpatialReference(SRS_WKT_WGS84);
#endif // GDAL_VERSION_MAJOR >= 3
        if (nullptr == poSourceSRS) {
            CPLDebug("QCT", "problem with WSG84 - is proj4 installed ?");
        }

        // OGRSpatialReference oTargetSRS ={}; // appeases --bug-hunting but breaks g++ and clang++
        OGRSpatialReference* poTargetSRS = new OGRSpatialReference();
        if (nullptr == poTargetSRS) {
            CPLDebug("QCT", "failed to initialize TargetSRS - memory problem ?");
        }
        if (poSourceSRS && poTargetSRS) {
            OGRErr targetErr = poTargetSRS->importFromEPSG(nEPSG);

            if (OGRERR_NONE == targetErr) {
#if GDAL_VERSION_MAJOR >= 3
                // proj6/7
                poTargetSRS->SetAxisMappingStrategy(OAMS_TRADITIONAL_GIS_ORDER);
#endif
                // check poTargetSRS (twice) here
                poTargetSRS->exportToWkt( &poDS->pszDatum );
                
                OGRCoordinateTransformation *poCT
                    = OGRCreateCoordinateTransformation( poSourceSRS,
                                                         poTargetSRS );
                if (poCT == nullptr 
                    || !poCT->Transform( 1, &(poDS->TL.lon), &(poDS->TL.lat) )
                    || !poCT->Transform( 1, &(poDS->TR.lon), &(poDS->TR.lat) )
                    || !poCT->Transform( 1, &(poDS->BL.lon), &(poDS->BL.lat) )
                    || !poCT->Transform( 1, &(poDS->BR.lon), &(poDS->BR.lat) )
                    )  {
                    CPLDebug("QCT",
                             "problems converting to EPSG:%d", nEPSG);
                } else {
                    /* poDS->TL.lon *= -1.0;
                       poDS->BL.lon *= -1.0;
                       poDS->TR.lon *= -1.0;
                       poDS->BR.lon *= -1.0; */
                    CPLDebug("QCT", "TL -> %.8g %.8g", poDS->TL.lon, poDS->TL.lat);
                    CPLDebug("QCT", "TR -> %.8g %.8g", poDS->TR.lon, poDS->TR.lat);
                    CPLDebug("QCT", "BL -> %.8g %.8g", poDS->BL.lon, poDS->BL.lat);
                    CPLDebug("QCT", "BR -> %.8g %.8g", poDS->BR.lon, poDS->BR.lat);
                }
                delete poCT;
            } else {
                CPLDebug("QCT",
                         "problem %d with EPSG:%d - is proj4 installed ?",
                         targetErr, nEPSG);
            }
            if (poSourceSRS) { delete(poSourceSRS); poSourceSRS=nullptr; }
            if (poTargetSRS) { delete(poTargetSRS); poTargetSRS=nullptr; }
            // not here ? VSIFree(poDS->pszProjection); delete ?
        } // if (poSourceSRS && poTargetSRS)
    } // if(...EPSG_from_Copyright(poDS->pszCopyright))
#endif // Use CRS of map when available, rather than WGS84 used in file. (was EPSG:27700)

    // Build a regular 4x4 grid of GCPs
    // Since the QCT polynomials are cubic this should be sufficient ?
    poDS->pasGCPList = (GDAL_GCP *) VSIMalloc2( sizeof(GDAL_GCP), 16);
    {
        int x, nPixel=0, y, nLine=0;
        poDS->nGCPCount = 0;

        GDALInitGCPs( 16, poDS->pasGCPList );
        
        for (x=0 ; x<=3 ; x++) {
            switch (x) {
            case 0: nPixel = 0; break;
            case 1: nPixel = (poDS->nRasterXSize /* +1.0 */ )/3.0; break;
            case 2: nPixel = (poDS->nRasterXSize*2.0 /*+1.0 */ )/3.0; break;
            case 3: nPixel =  poDS->nRasterXSize; break;
            }
            for (y=0 ; y<=3 ; y++) {
                char    szId[128]={};
                
                switch (y) {
                case 0: nLine = 0; break; 
                case 1: nLine = (poDS->nRasterYSize /* +1.0 */ )/3.0; break; 
                case 2: nLine = (poDS->nRasterYSize*2.0 /* +1.0 */ )/3.0; break; 
                case 3: nLine =  poDS->nRasterYSize; break; 
                }
                
                if ( poDS->pasGCPList[poDS->nGCPCount].pszId ) {
                    VSIFree( poDS->pasGCPList[poDS->nGCPCount].pszId );
                }
                snprintf( szId, sizeof(szId), "%d", poDS->nGCPCount+1 );
                poDS->pasGCPList[poDS->nGCPCount].pszId = VSIStrdup( szId );
                if ( poDS->pasGCPList[poDS->nGCPCount].pszInfo ) {
                    VSIFree( poDS->pasGCPList[poDS->nGCPCount].pszInfo );
                }
                poDS->pasGCPList[poDS->nGCPCount].pszInfo = VSIStrdup("");
                poDS->pasGCPList[poDS->nGCPCount].dfGCPPixel = nPixel;
                poDS->pasGCPList[poDS->nGCPCount].dfGCPLine  = nLine;
                {
                    latlon tempLatLong;
                    poDS->xy_to_latlon(static_cast<double>(nPixel),
                                       static_cast<double>(nLine),
                                       &tempLatLong );
                    poDS->pasGCPList[poDS->nGCPCount].dfGCPX = tempLatLong.lon;
                    poDS->pasGCPList[poDS->nGCPCount].dfGCPY = tempLatLong.lat;
                    poDS->pasGCPList[poDS->nGCPCount].dfGCPZ = 0.0;
                }
                
                CPLDebug("QCT GCP",
                         "GCP[%d] %s, %s (%g, %g) -> (%g, %g, %g)",
                         poDS->nGCPCount,
                         poDS->pasGCPList[poDS->nGCPCount].pszId,
                         poDS->pasGCPList[poDS->nGCPCount].pszInfo,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPPixel,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPLine,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPX,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPY,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPZ
                         );
                
                poDS->nGCPCount++;
            }
        }
    } // Build a regular 4x4 grid of GCPs
    
    char szValue[64]="";
    GDALRPCInfo sRPCInfo;
    
    // Use the corner coordinates calculated above
    sRPCInfo.dfMAX_LAT = sRPCInfo.dfMIN_LAT = poDS->TL.lat;
    sRPCInfo.dfMAX_LONG = sRPCInfo.dfMIN_LONG = poDS->TL.lon;
    
    sRPCInfo.dfMIN_LAT = MIN(sRPCInfo.dfMIN_LAT, poDS->TR.lat);
    sRPCInfo.dfMAX_LAT = MAX(sRPCInfo.dfMAX_LAT, poDS->TR.lat);
    sRPCInfo.dfMIN_LONG = MIN(sRPCInfo.dfMIN_LONG, poDS->TR.lon);
    sRPCInfo.dfMAX_LONG = MAX(sRPCInfo.dfMAX_LONG, poDS->TR.lon);
    
    sRPCInfo.dfMIN_LAT = MIN(sRPCInfo.dfMIN_LAT, poDS->BL.lat);
    sRPCInfo.dfMAX_LAT = MAX(sRPCInfo.dfMAX_LAT, poDS->BL.lat);
    sRPCInfo.dfMIN_LONG = MIN(sRPCInfo.dfMIN_LONG, poDS->BL.lon);
    sRPCInfo.dfMAX_LONG = MAX(sRPCInfo.dfMAX_LONG, poDS->BL.lon);
    
    sRPCInfo.dfMIN_LAT = MIN(sRPCInfo.dfMIN_LAT, poDS->BR.lat);
    sRPCInfo.dfMAX_LAT = MAX(sRPCInfo.dfMAX_LAT, poDS->BR.lat);
    sRPCInfo.dfMIN_LONG = MIN(sRPCInfo.dfMIN_LONG, poDS->BR.lon);
    sRPCInfo.dfMAX_LONG = MAX(sRPCInfo.dfMAX_LONG, poDS->BR.lon);

    CPLDebug("QCT RPC", "min lat %13g", sRPCInfo.dfMIN_LAT);
    CPLDebug("QCT RPC", "max lat %13g", sRPCInfo.dfMAX_LAT);
    CPLDebug("QCT RPC", "min lon %13g", sRPCInfo.dfMIN_LONG);
    CPLDebug("QCT RPC", "max lon %13g", sRPCInfo.dfMAX_LONG);

    double dfDiffLat = sRPCInfo.dfMAX_LAT - sRPCInfo.dfMIN_LAT;
    sRPCInfo.dfLAT_OFF = poDS->adfQCTdatumShift[0] ; // north
    if (fabs(dfDiffLat) > 1e-8) {
        sRPCInfo.dfLAT_SCALE = 2.0 / dfDiffLat;
    } else {
        CPLDebug("QCT RPC", "bounding box has no width %13g=%13g",
                 sRPCInfo.dfMIN_LAT, sRPCInfo.dfMAX_LAT);
        sRPCInfo.dfLAT_SCALE = 0.0; // Could I set this to NaN ?
    }

    double dfDiffLon = sRPCInfo.dfMAX_LONG - sRPCInfo.dfMIN_LONG;
    sRPCInfo.dfLONG_OFF = poDS->adfQCTdatumShift[1] ; // east

    if (fabs(dfDiffLon) < 1e-8) {
        CPLDebug("QCT RPC", "bounding box has no height %13g=%13g",
                 sRPCInfo.dfMIN_LONG, sRPCInfo.dfMAX_LONG);
        sRPCInfo.dfLONG_SCALE = 0.0; // Could I set this to NaN ?
    } else {
        sRPCInfo.dfLONG_SCALE = 2.0 / dfDiffLon;
    }
    
    sRPCInfo.dfSAMP_OFF = 0.0 ;
    sRPCInfo.dfSAMP_SCALE = poDS->dScalePixels; // 1.0 ;

    sRPCInfo.dfLINE_OFF = 0.0 ;
    sRPCInfo.dfLINE_SCALE = poDS->dScalePixels; // 1.0 ;

    sRPCInfo.dfHEIGHT_OFF = 0.0 ;
    sRPCInfo.dfHEIGHT_SCALE = 1.0 ;

    sRPCInfo.adfLINE_NUM_COEFF[0] = 1.0;
    sRPCInfo.adfSAMP_NUM_COEFF[0] = 1.0;
    sRPCInfo.adfLINE_DEN_COEFF[0] = 1.0;
    sRPCInfo.adfSAMP_DEN_COEFF[0] = 1.0;
    for(int i = 1; i < 20; i++ ) {
        sRPCInfo.adfLINE_NUM_COEFF[i] = 0.0;
        sRPCInfo.adfSAMP_NUM_COEFF[i] = 0.0;
        sRPCInfo.adfLINE_DEN_COEFF[i] = 0.0;
        sRPCInfo.adfSAMP_DEN_COEFF[i] = 0.0;
    }

    /* now set some of these variables from data in the file */
    /* these are probably not correctly scaled yet */
    /* ( lat p, long l ) -> row */
    sRPCInfo.adfLINE_NUM_COEFF[0]  = poDS->QCTpolynomial[10]; // const term
    sRPCInfo.adfLINE_NUM_COEFF[1]  = poDS->QCTpolynomial[11]; //  l 
    sRPCInfo.adfLINE_NUM_COEFF[2]  = poDS->QCTpolynomial[12]; //  p
    sRPCInfo.adfLINE_NUM_COEFF[4]  = poDS->QCTpolynomial[14]; //  lp
    sRPCInfo.adfLINE_NUM_COEFF[7]  = poDS->QCTpolynomial[15]; //  ll
    sRPCInfo.adfLINE_NUM_COEFF[8]  = poDS->QCTpolynomial[13]; //  pp
    sRPCInfo.adfLINE_NUM_COEFF[11] = poDS->QCTpolynomial[19]; //  lll
    sRPCInfo.adfLINE_NUM_COEFF[12] = poDS->QCTpolynomial[18]; //  lpp
    sRPCInfo.adfLINE_NUM_COEFF[14] = poDS->QCTpolynomial[17]; //  llp
    sRPCInfo.adfLINE_NUM_COEFF[15] = poDS->QCTpolynomial[16]; //  ppp
    /* ( lat p, long l ) -> column/pixel */
    sRPCInfo.adfSAMP_NUM_COEFF[0]  = poDS->QCTpolynomial[0]; // const term
    sRPCInfo.adfSAMP_NUM_COEFF[1]  = poDS->QCTpolynomial[1]; //  l 
    sRPCInfo.adfSAMP_NUM_COEFF[2]  = poDS->QCTpolynomial[2]; //  p
    sRPCInfo.adfSAMP_NUM_COEFF[4]  = poDS->QCTpolynomial[4]; //  lp
    sRPCInfo.adfSAMP_NUM_COEFF[7]  = poDS->QCTpolynomial[5]; //  ll
    sRPCInfo.adfSAMP_NUM_COEFF[8]  = poDS->QCTpolynomial[3]; //  pp
    sRPCInfo.adfSAMP_NUM_COEFF[11] = poDS->QCTpolynomial[9]; //  lll
    sRPCInfo.adfSAMP_NUM_COEFF[12] = poDS->QCTpolynomial[8]; //  lpp
    sRPCInfo.adfSAMP_NUM_COEFF[14] = poDS->QCTpolynomial[7]; //  llp
    sRPCInfo.adfSAMP_NUM_COEFF[15] = poDS->QCTpolynomial[6]; //  ppp
 
    /* put the values where gdal will use them */
    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLINE_OFF );
    poDS->SetMetadataItem( "LINE_OFF", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLINE_SCALE );
    poDS->SetMetadataItem( "LINE_SCALE", szValue, "RPC" );
    
    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfSAMP_OFF );
    poDS->SetMetadataItem( "SAMP_OFF", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfSAMP_SCALE );
    poDS->SetMetadataItem( "SAMP_SCALE", szValue, "RPC" );
   
    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLONG_OFF );
    poDS->SetMetadataItem( "LONG_OFF", szValue, "RPC" );
       
    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLONG_SCALE );
    poDS->SetMetadataItem( "LONG_SCALE", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLAT_OFF );
    poDS->SetMetadataItem( "LAT_OFF", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLAT_SCALE );
    poDS->SetMetadataItem( "LAT_SCALE", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfHEIGHT_OFF );
    poDS->SetMetadataItem( "HEIGHT_OFF", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfHEIGHT_SCALE );
    poDS->SetMetadataItem( "HEIGHT_SCALE", szValue, "RPC" );

    poDS->QCTsetGCPMetaData_from_coefficients(20, "LINE_NUM_COEFF",
                                              sRPCInfo.adfLINE_NUM_COEFF, "RPC");
    poDS->QCTsetGCPMetaData_from_coefficients(20, "LINE_DEN_COEFF",
                                              sRPCInfo.adfLINE_DEN_COEFF, "RPC");
    poDS->QCTsetGCPMetaData_from_coefficients(20, "SAMP_NUM_COEFF",
                                              sRPCInfo.adfSAMP_NUM_COEFF, "RPC");
    poDS->QCTsetGCPMetaData_from_coefficients(20, "SAMP_DEN_COEFF",
                                              sRPCInfo.adfSAMP_DEN_COEFF, "RPC");

    // sRPCInfo.dfMIN_LONG = sRPCInfo.dfLONG_OFF - ( sRPCInfo.dfLONG_SCALE / 2.0 );
    // sRPCInfo.dfMIN_LAT  = sRPCInfo.dfLAT_OFF - ( sRPCInfo.dfLAT_SCALE / 2.0 );
    // sRPCInfo.dfMAX_LONG = sRPCInfo.dfLONG_OFF + ( sRPCInfo.dfLONG_SCALE / 2.0 );
    // sRPCInfo.dfMAX_LAT  = sRPCInfo.dfLAT_OFF + ( sRPCInfo.dfLAT_SCALE / 2.0 );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfMIN_LONG );
    poDS->SetMetadataItem( "MIN_LONG", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfMAX_LONG );
    poDS->SetMetadataItem( "MAX_LONG", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfMIN_LAT );
    poDS->SetMetadataItem( "MIN_LAT", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfMAX_LAT);
    poDS->SetMetadataItem( "MAX_LAT", szValue, "RPC" );

    // -1 indicates unknown bias error and random error
    poDS->SetMetadataItem( "ERR_BIAS", "-1.0", "RPC" );
    poDS->SetMetadataItem( "ERR_RAND", "-1.0", "RPC" );
#endif // def QCT_USE_GCP

/* -------------------------------------------------------------------- */
/*      Create TIFFTAGs from metadata information.                      */
/* -------------------------------------------------------------------- */
    poDS->SetMetadataItem( "TIFFTAG_COPYRIGHT", poDS->pszCopyright, "");
    if (poDS->pszOrigFileName && poDS->pszOrigFileName[0]) {
        poDS->SetMetadataItem( "TIFFTAG_DOCUMENTNAME",
                               poDS->pszOrigFileName, "");
    }
    if (poDS->nOrigFileTime) {
        poDS->SetMetadataItem( "TIFFTAG_DATETIME",
                               ctime(&poDS->nOrigFileTime), "");
    }

/* -------------------------------------------------------------------- */
/*      Create band information objects.                                */
/* -------------------------------------------------------------------- */
    if (poDS->bNeedQC3data()) { 
        if (poDS->bHaveQC3data()) {
            // Image data in .qc3 file
#if defined(QC3_PIXEL_IS_TILE)
            poDS->nOverviewCount=-1;
#else // QC3_PIXEL_IS_PIXEL
            CPLDebug("QCT", "qc3_pixel_is_pixel");
            
            // QC3_PIXEL_IS_PIXEL
            // this will be the default        
            poDS->nOverviewCount=10;
#endif
        } else {
            // No data in qct file, cannot access qc3 file.
            poDS->nOverviewCount=-1;
        }
    } else {
        // Image data in .qct file
        poDS->nOverviewCount=6;
        /* level   tile width
             -1       64
              0       32
              1       16
              2        8
              3        4
              4        2
              5        1
        */
    }

    QCTRasterBand *poBand =
        new QCTRasterBand( poDS, 1, -1);
    poDS->SetBand( 1, poBand);
    poBand->SetNoDataValue( nQCTNoData );

    // ... and the Overview Bands 
    if (poDS->nOverviewCount>=0) {
    /* We don't use the gdal default
     *   poDS->oOvManager.Initialize( poDS, poOpenInfo->pszFilename );
     * since rows of each tile are interleaved in a way that means
     * we don't need to read all the data, and there is an
     * interpolation matrix which explains how to
     * merge adjacent pixels into a single colour.
     */
        for (int i=0; i<poDS->nOverviewCount; i++) {
            poDS->papoOverviewBands[i]= new
                QCTRasterBand(poDS, 1, i);
            poDS->papoOverviewBands[i]->SetNoDataValue( nQCTNoData );
            CPLDebug("QCT OVRV",
                     "created overview %d of %d at %p %p level %d",
                     i, poDS->nOverviewCount,
                     poDS->papoOverviewBands,
                     poDS->papoOverviewBands[i],
                     poDS->papoOverviewBands[i]->iOverview
                     );
        }
    } else {
       CPLDebug("QCT OVRV",
                "No overviews to set"
                );
    }

/* -------------------------------------------------------------------- */
/*      Initialize any PAM information.                                 */
/* -------------------------------------------------------------------- */
    poDS->SetDescription( poOpenInfo->pszFilename );
    // poDS->TryLoadXML(); // I think I don't want .aux.xml files.

#if 1
    // Why were we throwing this info away ?
    // Enabling this code appears to stop a memory leak !
    if( pszQCTDatum != nullptr ){
        VSIFree ( pszQCTDatum );
        pszQCTDatum = nullptr;
    }
#endif

    return( poDS );
} // QCTDataset::Open

/************************************************************************/
/*                          GDALRegister_qct()                          */
/************************************************************************/

void GDALRegister_qct()
{
    if (! GDAL_CHECK_VERSION("QCT"))
        return;

    if( GDALGetDriverByName( "QCT" ) != nullptr )
        return;

    GDALDriver *poDriver = new GDALDriver();

    poDriver->SetDescription( "QCT" );

    // required in gdal version 2, not supported in gdal 1.11
#if GDAL_VERSION_MAJOR >= 2
    poDriver->SetMetadataItem( GDAL_DCAP_RASTER, "YES" );
#endif

    poDriver->SetMetadataItem( GDAL_DMD_LONGNAME, 
                               "Memory Map QuickChart map (.qct)" );
    poDriver->SetMetadataItem( GDAL_DMD_HELPTOPIC, 
                               "frmt_various.html#QCT" );
    poDriver->SetMetadataItem( GDAL_DMD_EXTENSION, "qct" );
    poDriver->SetMetadataItem( GDAL_DMD_CREATIONDATATYPES, "Byte" );
    poDriver->SetMetadataItem( GDAL_DCAP_VIRTUALIO, "YES" );
    
    // "NONRECIPROCAL" is the intent of the authors (AA + AB)
    // of the code for this driver.
    // Since they are not the authors, or owners,
    // of the qct or qc3 file formats, further research may be needed.
    poDriver->SetMetadataItem( "LICENSE_POLICY", "NONRECIPROCAL" );

    // Which of these is correct ?
    poDriver->SetMetadataItem(GDALMD_AREA_OR_POINT,
                              // GDALMD_AOP_POINT
                              GDALMD_AOP_AREA
                              );
    // GDALMD_AOP_AREA is the GDALv2 default.
    
    poDriver->pfnOpen = static_cast<GDALDataset *(*)(GDALOpenInfo *)>(QCTDataset::Open);
    poDriver->pfnIdentify = QCTDataset::Identify;

    GetGDALDriverManager()->RegisterDriver( poDriver );

#if GDAL_VERSION_MAJOR >= 2
    // defined in qctvector.cpp
    GDALRegister_qctvector();
#endif

#ifdef FRMT_qed
    // defined in qed.cpp
    GDALRegister_qed();
#endif

} // GDALRegister_qct()

char *QCTDataset::QCTGetString(size_t byteaddr) {
#if 0
    if (0>byteaddr) {
        CPLDebug("QCT","QCTGetString: 0> 0x%012zx",
             byteaddr);
        return nullptr;
    }
#endif
    if(byteaddr >= this->nQCTfilesize) {
        CPLDebug("QCT","QCTGetString: 0x%012zx >= 0x%012zx",
                 byteaddr, nQCTfilesize);
        return nullptr;
    }
    
    return ::QCTGetString(this->fp, byteaddr);
    
} // QCTDataset::QCTGetString(size_t byteaddr)

char *QCTDataset::QCTGetStringIndirect(size_t byteaddraddr)
{
    if (byteaddraddr<1) {
        CPLDebug("QCT","QCTGetStringIndirect: 0x%012zx <= 0",
             byteaddraddr);
        return nullptr;
    }
    if(byteaddraddr >= nQCTfilesize) {
        CPLDebug("QCT","QCTGetStringIndirect: 0x%012zx >= 0x%012zx",
                 byteaddraddr, nQCTfilesize);
        return nullptr;
    }

    size_t byteaddr;
    if (byteaddraddr<sizeof(abyHeader)) {
        byteaddr=QCTGetInt(abyHeader, byteaddraddr);
    } else {
        byteaddr=QCTGetInt(fp, byteaddraddr);
    }
    if (byteaddr<0x60) {
        CPLDebug("QCT","QCTGetStringIndirect: [0x%012zx]=0x%012zx < 0x60",
                 byteaddraddr, byteaddr);
        return nullptr;
    }
    CPLDebug("QCT","QCTGetStringIndirect: byteaddr [0x%012zx]=0x%012zx : nQCTfilesize 0x%012zx",
             byteaddraddr, byteaddr, nQCTfilesize);
    if(byteaddr >= this->nQCTfilesize) {
        CPLDebug("QCT","QCTGetStringIndirect: [0x%012zx]=0x%012zx >= 0x%012zx",
                 byteaddraddr, byteaddr, nQCTfilesize);
        return nullptr;
    }

    return QCTDataset::QCTGetString(byteaddr);
    
} // QCTDataset::QCTGetStringIndirect(size_t byteaddraddr)

/* -------------------------------------------------------------------------
 * Convert pixel (x,y) in image into (longitude,latitude)
 * x,y from top left
 * latitude,longitude in degrees WGS84.
 * clips out of range x,y rather than blindly converting.
 *
 * This is a wrapper for QCT_xy_to_latlon which does the real work.
 */
void
QCTDataset::xy_to_latlon(double x, double y, latlon *latlon)
{
    // Range check - clip to image limits, or allow extrapolation?
    if (x<0) {
        CPLDebug("QCT", "xy_to_latlon: %13g increased to 0", x);
        x=0;
    }
    if (y<0)  {
        CPLDebug("QCT", "xy_to_latlon: %13g increased to 0", y);
        y=0;
    }
    if (x>nRasterXSize*dScalePixels)  {
        CPLDebug("QCT", "xy_to_latlon: x %13g reduced to %d * %13g = %13g",
                 x, nRasterXSize,dScalePixels, nRasterXSize*dScalePixels);
        x=nRasterXSize*dScalePixels;
    }
    if (y>nRasterYSize*dScalePixels)  {
        CPLDebug("QCT", "xy_to_latlon: y %13g reduced to %d * %13g = %13g",
                 y, nRasterXSize,dScalePixels, nRasterXSize*dScalePixels);
        y=nRasterYSize*dScalePixels;
    }
    // Allow limited extrapolation, but clip if way out
    // if (x<-1.0*nRasterXSize*dScalePixels) x = -1.0 * nRasterXSize*dScalePixels;
    // if (y<-1.0*nRasterYSize*dScalePixels) y = -1.0 * nRasterYSize*dScalePixels;
    // if (x> 2.0*nRasterXSize*dScalePixels) x =  2.0 * nRasterXSize*dScalePixels;
    // if (y> 2.0*nRasterYSize*dScalePixels) y =  2.0 * nRasterYSize*dScalePixels;

    QCT_xy_to_latlon(
                     x*dScalePixels, y*dScalePixels,
                     latlon,
                     QCTpolynomial,
                     adfQCTdatumShift
                     );

    // Clip latitude
    if (latlon->lat >  90.0) {
        CPLDebug("QCT", "xy_to_latlon: clipping %f to north pole", latlon->lat);
        latlon->lat =  90.0 ;
    }
    if (latlon->lat < -90.0) {
        CPLDebug("QCT", "xy_to_latlon: clipping %f to south pole", latlon->lat);
        latlon->lat = -90.0 ;
    }
    // can longitude wrap infinitely, or should we clip it too ?
} // QCTDataset::xy_to_latlon

/* -------------------------------------------------------------------------
 */
static unsigned int
bits_per_pixel(unsigned int num_colours)
{
#if 1
    return static_cast<unsigned int>(log2(num_colours)+0.999);
#else
    /* Alternatively: */
    if (num_colours <=   2) return 1;
    if (num_colours <=   4) return 2;
    if (num_colours <=   8) return 3;
    if (num_colours <=  16) return 4;
    if (num_colours <=  32) return 5;
    if (num_colours <=  64) return 6;
    if (num_colours <= 128) return 7;
    return 8;
#endif
}

/* -------------------------------------------------------------------------
 */
void
QCTRasterBand::read_QCT_Tile(VSILFILE *fp,
                             int tile_xx, int tile_yy,
                             void *pImage)
{
    if (fp==nullptr || tile_xx<0 || tile_yy<0 || pImage==nullptr) {
        CPLDebug("QCT",
                  "read_QCT_Tile(%p,%d, %d, %p) pased an invalid arg",
                  fp, tile_xx, tile_yy, pImage);
        return;
    }

    unsigned char tile_data[QCT_TILE_PIXELS]={}; // this is 4K and we may not need all of it
                                                 // Use a vector ?
    unsigned char *row_ptr[QCT_TILE_SIZE]={};    // This is 64 pointers.
      signed int packing;
      signed int row;
    unsigned int pixelnum = 0;

    // Rows are interleaved in this order (reverse binary)
    static const unsigned int row_seq[] =
    {
        0, 32, 16, 48,   8, 40, 24, 56,   4, 36, 20, 52,  12, 44, 28, 60,
        2, 34, 18, 50,  10, 42, 26, 58,   6, 38, 22, 54,  14, 46, 30, 62,
        1, 33, 17, 49,   9, 41, 25, 57,   5, 37, 21, 53,  13, 45, 29, 61,
        3, 35, 19, 51,  11, 43, 27, 59,   7, 39, 23, 55,  15, 47, 31, 63
    };
#ifdef VERIFY_ROW_SEQ
    for (unsigned int n=0; n<64; n++) {
        unsigned int i = n;
        unsigned int r = 0;
        for (unsigned int j=0; j<6; j++) {
            r <<=1;
            r |= (i & 1);
            i >>=1;
        }
        if (r!=row_seq[n]) {
            CPLDebug("QCT",
                     "QCT row_seq[%u]=%u != %u",
                     n, row_seq[n], r);
        }
    }
#endif //  VERIFY_ROW_SEQ

    const unsigned int nQCT_overview_tile_size = QCT_TILE_SIZE >> (1+iOverview);
    // This is not square.
    // At this stage the unpacked overview data is
    // all of the pixels in *some* of the rows.
    const unsigned int nQCT_overview_tile_pixels =
        nQCT_overview_tile_size * QCT_TILE_SIZE;

    if (nQCT_overview_tile_size != static_cast<unsigned int>(nBlockXSize)) {
        CPLDebug("QCT",
                 "read_QCT_Tile: nQCT_overview_tile_size %u != nBlockXSize %d",
                 nQCT_overview_tile_size, nBlockXSize);
    }

    CPLDebug("QCT", "read_QCT_Tile(%p, %d, %d, %p) band %p overview %d",
             fp, tile_xx, tile_yy, pImage,
             this, iOverview);
    CPLDebug("QCT", "iOverview %d nQCT_overview_tile_size = %u nQCT_overview_tile_pixels = %u",
             iOverview, nQCT_overview_tile_size, nQCT_overview_tile_pixels);
    //CPLDebug("QCT", "about to call VSIFTellL for tile %d %d, file %p",
    //         tile_xx, tile_yy, fp);
    //vsi_l_offset fileCurrentPos = VSIFTellL(fp);
    //CPLDebug("QCT", "Tile %d, %d starts at file offset 0x%lx",
    //         tile_xx, tile_yy, fileCurrentPos);

    //CPLDebug("QCT", "Tile %d, %d: calling QCTFGetc(%p) at offset %lx",
    //         tile_xx, tile_yy, fp, fileCurrentPos);
    // Determine which method was used to pack this tile
    // packing = static_cast<unsigned char>(QCTFGetc(fp));
    packing = QCTFGetuc(fp);
    if (-128>packing||packing>255) {
        CPLDebug("QCT", "packing %d=0x%x not a byte", packing, packing);
        exit (-1);
    }
    
    CPLDebug("QCT", "Reading tile %d, %d; %s packing=0x%x",
             tile_xx, tile_yy,
             ((packing==0||packing==255)?"huffman":(packing>127?"pixel":"RLE")),
             packing
             );

    memset(tile_data, nQCTNoData, nQCT_overview_tile_pixels);


    // Calculate pointer into pImage for each row in this tile
    for (row=0; row<nBlockYSize; row++)
    {
        // Top of image
        row_ptr[row] = static_cast<unsigned char*>(pImage);
        // Interleaved rows in the above sequence
        row_ptr[row] += row_seq[row<<(1+iOverview)] * nBlockXSize;
    }

    unsigned int ii;
    
    // Uncompress each row

    if (packing == 0 || packing == 255) {
        // Huffman
        // CPLDebug("QCT", "Huffman");
        unsigned int huff_size = 256;
        unsigned char *huff = static_cast<unsigned char*>(VSIMalloc(huff_size));
        unsigned int huff_idx = 0;
        for (unsigned int n=huff_idx; n<huff_size; huff[n++] = '\0');
        int num_colours = 0;
        int num_branches = 0;
        while (num_colours <= num_branches)
        {
            huff[huff_idx] = QCTFGetuc(fp);
            // Relative jump further than 128 needs two more bytes
            if (huff[huff_idx] == 128) {
                huff[++huff_idx] = QCTFGetuc(fp);
                huff[++huff_idx] = QCTFGetuc(fp);
                num_branches++;
            }
            // Relative jump nearer is encoded directly
            else if (huff[huff_idx] > 128)
            {
                // Count number of branches so we know when tree is built
                num_branches++;
            }
            // Otherwise it's a colour index into the palette
            else
            {
                // Count number of colours so we know when tree is built
                num_colours++;
            }
            huff_idx++;
            // Make space when Huffman table runs out of space
            if (huff_idx > huff_size-3) {
                huff_size += 256;
                huff = static_cast<unsigned char*>(VSIRealloc(huff, huff_size));
                for (unsigned int n=huff_idx; n<huff_size; huff[n++] = '\0');
            }
        }
        unsigned char *huff_ptr = huff;
        
        // If only 1 colour then tile is solid colour so no data follows
        if (num_colours == 1) {
            // CPLDebug("QCT", "ONECOL: setting %d bytes at %p to %d",
            //          QCT_TILE_PIXELS, tile_data, huff[0]);

            // preparing for overviews
            for (unsigned int yy=0; yy<nQCT_overview_tile_size; yy++) {
                memset(tile_data+yy*nQCT_overview_tile_size,
                       huff[0], nQCT_overview_tile_size);
#ifdef NOISY
                CPLDebug("QCT", "ONECOL: setting %d bytes at %p (row %d) to %d",
                         nQCT_overview_tile_size,
                         tile_data+yy*nQCT_overview_tile_size,
                         yy, huff[0]
                         );
#endif // NOISY
            }
        } else {
            // Validate Huffman table by ensuring all branches are within table
            // (if not just return so tile will be not be unpacked, ie. blank)
            {
                unsigned int delta;
                for (unsigned int hi=0; hi<huff_idx; hi++)
                {
                    if (huff[hi] < 128) {
                        continue;
                    } else if (huff[hi] == 128) {
                        if (hi+2 >= huff_idx)
                            return;
                        delta = 65537 - (256 * huff[hi+2] + huff[hi+1]) + 2;
                        if (hi+delta >= huff_idx)
                            return;
                        hi += 2;
                    } else {
                        delta = 257 - huff[hi];
                        if (hi+delta >= huff_idx)
                            return;
                    }
                }
            }

            // Read tile data one bit at a time following branches in Huffman tree
            int bits_left = 8;
            int bit_value;
            ii = QCTFGetuc(fp);
            while (pixelnum < nQCT_overview_tile_pixels)
            {
                // If entry is a colour then output it
                if (*huff_ptr < 128) {
                    tile_data[pixelnum++] = *huff_ptr;
                    // Go back to top of tree for next pixel
                    huff_ptr = huff;
                    continue;
                }
                // Get the next "bit"
                bit_value = (ii & 1);
                // Prepare for the following one
                ii >>= 1;
                bits_left--;
                if (bits_left == 0) {
                    ii = QCTFGetuc(fp);
                    bits_left = 8;
                }
                // Now check value to see whether to follow branch or not
                if (bit_value == 0) {
                    // Don't jump just proceed to next entry in Huffman table
                    if (*huff_ptr == 128) huff_ptr+=2;
                    huff_ptr++;
                } else {
                    // Follow branch to different part of Huffman table
                    if (*huff_ptr > 128) {
                        // Near jump
                        huff_ptr += 257 - (*huff_ptr);
                    } else if (*huff_ptr == 128) {
                        // Far jump needs two more bytes
                        int delta = 65537 - (256 * huff_ptr[2] + huff_ptr[1]) + 2;
                        huff_ptr += delta;
                    }
                    // NB < 128 already handled above
                }
            } // while (pixelnum < nQCT_overview_tile_pixels)
        } // else (num_colours == 1)
        if (huff) {
            VSIFree(huff);
        }
    } // (packing == 0 || packing == 255)

    else if (packing == 128) {
        // An encrypted type of packing??
        CPLDebug("QCT",
                 "Encrypted packing not yet supported for tile %d %d (0x%x 0x%x)",
                 tile_xx,tile_yy, tile_xx,tile_yy
                 );
        while (pixelnum < nQCT_overview_tile_pixels) {
            tile_data[pixelnum++] = nQCTencryptedData;
        }
    }

    else if (packing > 128) // was 127
    {
        // Pixel packing
        unsigned int num_sub_colours = static_cast<unsigned int>(256 - packing);
        int shift = static_cast<int>(bits_per_pixel(num_sub_colours));
        unsigned int mask = (1 << shift) - 1;
        int num_pixels_per_word = 32 / shift;
        unsigned char palette_index[256];
        // CPLDebug("QCT",
        //          "PACKED: sub-palette size is %d (%d bits 0x%08x)\,
        //          num_sub_colours, shift, mask);
        // Read the sub-palette
        for (unsigned int i=0; i<num_sub_colours; i++)
        {
            palette_index[i] = QCTFGetuc(fp);
            // CPLDebug("QCT", "PACKED: palette %d = %d", i, palette_index[i]);
        }
        // Read the pixels in 4-byte words and unpack the bits from each
        while (pixelnum < nQCT_overview_tile_pixels)
        {
            ii = QCTreadUInt(fp);
            // CPLDebug("QCT", "PACKED: read %08x", ii);
            for ( int runs = 0; runs < num_pixels_per_word; runs++)
            {
                unsigned int colour = ii & mask;
                ii = static_cast<unsigned int>(ii) >> shift;
                // CPLDebug("QCT",
                //     "PACKED: [%d] = palette[%d] = %d (0x%08x, 0x%08x, %d)",
                //     pixelnum, colour, palette_index[colour],
                //     ii, mask, shift);
                tile_data[pixelnum++] = palette_index[colour];
            }
        }
    }
    else
    {
        // Run-length Encoding
        unsigned int num_sub_colours = 0xff & packing;
        unsigned int num_low_bits = bits_per_pixel(num_sub_colours);
        unsigned int pal_mask = (1 << num_low_bits)-1;
        unsigned int palette_index[256];
#if 0
        CPLDebug("QCT",
                 "RLE: file pos 0x%llx sub-palette size is 0x%08x (uses %d bits) mask 0x%x",
                 VSIFTellL(fp),
                 num_sub_colours, num_low_bits, pal_mask);
#endif
        for (unsigned int i=0; i<num_sub_colours; i++)
        {
             palette_index[i] = QCTFGetuc(fp);
             // CPLDebug("QCT", "RLE palette %d = %d", i, palette_index[i]);
        }

        while (pixelnum < nQCT_overview_tile_pixels)
        {
            ii = 0xff & QCTFGetc(fp);
            unsigned int colour = static_cast<unsigned char>
                (static_cast<unsigned int>(ii) & pal_mask);
            unsigned char runs = static_cast<unsigned char>
                (ii >> num_low_bits);

            if (runs+pixelnum > QCT_TILE_PIXELS) {
                // ie if we will run past the end of the *full* tile
                CPLError(CE_Failure, CPLE_AppDefined,
                         "stuck RLE tile %d %d - %d pixels of colour %d left\n",
                         tile_xx, tile_yy,
                         runs+pixelnum-QCT_TILE_PIXELS, colour);
            }
            while (runs-- > 0)
            {
                tile_data[pixelnum++] = static_cast<unsigned char>
                    (palette_index[colour]);
                if (pixelnum>=nQCT_overview_tile_pixels) { break; }
            }
        }
    } // end of else ... - the RLE case
    
    /********************* Finished unpacking into tile_data ********************/

    CPLDebug("QCT",
             "nQCT_overview_tile_size %d <?> QCT_TILE_SIZE %d <?> nBlockXSize %d",
             nQCT_overview_tile_size, QCT_TILE_SIZE, nBlockXSize);
    // Decommutate the interleaved rows
    // "interpolate" the pixels to shrink each row
    //  and copy result into the image.
    {
        for (unsigned int yy=0; yy<nQCT_overview_tile_size; yy++) {
            if (row_ptr[yy]==nullptr) {
                CPLDebug("QCT",
                         "row_ptr[%u] is null", yy);
                continue;
            }
            unsigned char* pTileRowStart = tile_data+yy*QCT_TILE_SIZE;
            int nRowWidth = QCT_TILE_SIZE;
            // shrink the image, halving for each overlay step
            while (nRowWidth > nBlockXSize) {
                nRowWidth >>= 1;
                for (int xx=0; xx<nRowWidth; xx++) {
                    unsigned char p1, p2;
                    p1 = nQCTNoData;
                    p2 = nQCTNoData;
                    if (2*xx+1 < QCT_TILE_SIZE) {
                        p1 = pTileRowStart[2*xx];
                        //if (2*xx+1 < QCT_TILE_SIZE) {
                            p2 = pTileRowStart[2*xx+1];
                            //}
                    }
                    unsigned char res = nQCTNoData;
#ifdef NOISY
                    CPLDebug("QCT",
                             "decimating row %u pixel %d/%d",
                             yy, xx, nRowWidth
                             );
                    CPLDebug("QCT", "\tp1  %hhu", p1);
                    CPLDebug("QCT", "\tp2  %hhu", p2);
                    CPLDebug("QCT", "\tres %hhu", res);
#endif
                    if (p1<128) {
                        if (p2<128)
                            res = static_cast<QCTDataset*>(poDS)->acInterpolationMatrix[p1][p2];
                        else res = p1;
                    } else {
                        if (p2<128) res = p2;
                        else if (p1==p2) { res=p1; }
                    }
                    pTileRowStart[xx] = res; 
                } // for xx
            } // for yy
            //  and copy into image.
            memcpy(row_ptr[yy], pTileRowStart,
                   nQCT_overview_tile_size);
        } // for yy - rows
        
        CPLDebug("QCT",
                 "tile %d %d: overview %d (%d^2 pixels) transferred",
                 tile_xx, tile_yy, iOverview, nQCT_overview_tile_size
                 );

    } // Decommutate, "interpolate" to shrink and copy.
    
    if (getenv("QCT_DUMP_TILE")) {
        unsigned int nPGMCountMax = static_cast<unsigned int>
            (strtol(getenv("QCT_DUMP_TILE"),nullptr,10));
        CPLString osBaseLabel = CPLString().Printf
            ("/tmp/werdna/vrc2tif/%s.%01d.x%04d.y%04d",
             // CPLGetBasename(poOpenInfo->pszFilename) doesn't quite work
             static_cast<QCTDataset *>(poDS)->pszLongTitle,
             iOverview,
             tile_xx, tile_yy
             );
        CPLDebug("QCT",
                 "About to dump pgm tile: blocksize %d by %d, raster size %d by %d, overview %d nRasterXSize >> (1+iOverview) %d",
                 nBlockXSize, nBlockYSize,
                 nRasterXSize, nRasterYSize,
                 iOverview,
                 nRasterXSize >> (1+iOverview)
                 );
#if 0 // decompressed but not decommutated or shrunk.
        dumpPGM( nBlockXSize, nBlockYSize,
                 tile_data,
                 nBlockXSize, // nRasterXSize,
                 osBaseLabel,
                 nPGMCountMax);
#else // decommutated and shrunk.
        dumpPGM( nBlockXSize, nBlockYSize,
                 static_cast<unsigned char*>(pImage),
                 nBlockXSize,
                 osBaseLabel,
                 nPGMCountMax);
#endif // decommutated and shrunk.
    } // getenv("QCT_DUMP_TILE")
} // QCTRasterBand::read_QCT_Tile



/* -------------------------------------------------------------------------
 * -------------------------------------------------------------------------
 * -------------------------------------------------------------------------
 */
void
QCTRasterBand::read_QCthree_Tile(VSILFILE *fp,
                                 int tile_xx, int tile_yy,
                                 void *pImage)
{
    QCTDataset *poQDS = static_cast<QCTDataset *>(poDS);
    // CPLDebug("QCT", "poQDS[%p] = %p", this, poQDS);
#ifdef QC3_IS_DUMMY
    // Dummy code, draws a solid colored tile
    unsigned char nTileColor =
        // nQCTonlineData
        nQCTYesData
        // static_cast<unsigned char>(iOverview*0x11)
        ;
    for(int i = 0; i < nBlockXSize*nBlockYSize; i++ ) {
        static_cast<unsigned char*>(pImage)[i] = nTileColor;
    }
           
#else //  QC3_IS_DUMMY
    unsigned char tile_data[QC3_TILE_PIXELS]={}; // this is a megabyte and we may not need all of it
                                                 // Use a vector ?
    unsigned char* row_ptr[QC3_TILE_SIZE]={};    // This is 1024 pointers.
    // unsigned int pixelnum = 0;

    // Rows are interleaved in this order (reverse binary)
    static const unsigned int row_seq[] =
    {
      // Spec says this starts 0, 512, 256, 384, 128, 640
      // but 384 doesn't match the binary 1100000000 quoted there.
        0, 512, 256, 768, 128, 640, 384, 896,  64, 576, 320, 832, 192, 704, 448, 960,
       32, 544, 288, 800, 160, 672, 416, 928,  96, 608, 352, 864, 224, 736, 480, 992,
       16, 528, 272, 784, 144, 656, 400, 912,  80, 592, 336, 848, 208, 720, 464, 976,
       48, 560, 304, 816, 176, 688, 432, 944, 112, 624, 368, 880, 240, 752, 496, 1008,

        8, 520, 264, 776, 136, 648, 392, 904,  72, 584, 328, 840, 200, 712, 456, 968,
       40, 552, 296, 808, 168, 680, 424, 936, 104, 616, 360, 872, 232, 744, 488, 1000,
       24, 536, 280, 792, 152, 664, 408, 920,  88, 600, 344, 856, 216, 728, 472, 984,
       56, 568, 312, 824, 184, 696, 440, 952, 120, 632, 376, 888, 248, 760, 504, 1016,

        4, 516, 260, 772, 132, 644, 388, 900,  68, 580, 324, 836, 196, 708, 452, 964,
       36, 548, 292, 804, 164, 676, 420, 932, 100, 612, 356, 868, 228, 740, 484, 996,
       20, 532, 276, 788, 148, 660, 404, 916,  84, 596, 340, 852, 212, 724, 468, 980,
       52, 564, 308, 820, 180, 692, 436, 948, 116, 628, 372, 884, 244, 756, 500, 1012,

       12, 524, 268, 780, 140, 652, 396, 908,  76, 588, 332, 844, 204, 716, 460, 972,
       44, 556, 300, 812, 172, 684, 428, 940, 108, 620, 364, 876, 236, 748, 492, 1004,
       28, 540, 284, 796, 156, 668, 412, 924,  92, 604, 348, 860, 220, 732, 476, 988,
       60, 572, 316, 828, 188, 700, 444, 956, 124, 636, 380, 892, 252, 764, 508, 1020,

        2, 514, 258, 770, 130, 642, 386, 898,  66, 578, 322, 834, 194, 706, 450, 962,
       34, 546, 290, 802, 162, 674, 418, 930,  98, 610, 354, 866, 226, 738, 482, 994,
       18, 530, 274, 786, 146, 658, 402, 914,  82, 594, 338, 850, 210, 722, 466, 978,
       50, 562, 306, 818, 178, 690, 434, 946, 114, 626, 370, 882, 242, 754, 498, 1010,

       10, 522, 266, 778, 138, 650, 394, 906,  74, 586, 330, 842, 202, 714, 458, 970,
       42, 554, 298, 810, 170, 682, 426, 938, 106, 618, 362, 874, 234, 746, 490, 1002,
       26, 538, 282, 794, 154, 666, 410, 922,  90, 602, 346, 858, 218, 730, 474, 986,
       58, 570, 314, 826, 186, 698, 442, 954, 122, 634, 378, 890, 250, 762, 506, 1018,

        6, 518, 262, 774, 134, 646, 390, 902,  70, 582, 326, 838, 198, 710, 454, 966,
       38, 550, 294, 806, 166, 678, 422, 934, 102, 614, 358, 870, 230, 742, 486, 998,
       22, 534, 278, 790, 150, 662, 406, 918,  86, 598, 342, 854, 214, 726, 470, 982,
       54, 566, 310, 822, 182, 694, 438, 950, 118, 630, 374, 886, 246, 758, 502, 1014,

       14, 526, 270, 782, 142, 654, 398, 910,  78, 590, 334, 846, 206, 718, 462, 974,
       46, 558, 302, 814, 174, 686, 430, 942, 110, 622, 366, 878, 238, 750, 494, 1006,
       30, 542, 286, 798, 158, 670, 414, 926,  94, 606, 350, 862, 222, 734, 478, 990,
       62, 574, 318, 830, 190, 702, 446, 958, 126, 638, 382, 894, 254, 766, 510, 1022,

        1, 513, 257, 769, 129, 641, 385, 897,  65, 577, 321, 833, 193, 705, 449, 961,
       33, 545, 289, 801, 161, 673, 417, 929,  97, 609, 353, 865, 225, 737, 481, 993,
       17, 529, 273, 785, 145, 657, 401, 913,  81, 593, 337, 849, 209, 721, 465, 977,
       49, 561, 305, 817, 177, 689, 433, 945, 113, 625, 369, 881, 241, 753, 497, 1009,

        9, 521, 265, 777, 137, 649, 393, 905,  73, 585, 329, 841, 201, 713, 457, 969,
       41, 553, 297, 809, 169, 681, 425, 937, 105, 617, 361, 873, 233, 745, 489, 1001,
       25, 537, 281, 793, 153, 665, 409, 921,  89, 601, 345, 857, 217, 729, 473, 985,
       57, 569, 313, 825, 185, 697, 441, 953, 121, 633, 377, 889, 249, 761, 505, 1017,

        5, 517, 261, 773, 133, 645, 389, 901,  69, 581, 325, 837, 197, 709, 453, 965,
       37, 549, 293, 805, 165, 677, 421, 933, 101, 613, 357, 869, 229, 741, 485, 997,
       21, 533, 277, 789, 149, 661, 405, 917,  85, 597, 341, 853, 213, 725, 469, 981,
       53, 565, 309, 821, 181, 693, 437, 949, 117, 629, 373, 885, 245, 757, 501, 1013,

       13, 525, 269, 781, 141, 653, 397, 909,  77, 589, 333, 845, 205, 717, 461, 973,
       45, 557, 301, 813, 173, 685, 429, 941, 109, 621, 365, 877, 237, 749, 493, 1005,
       29, 541, 285, 797, 157, 669, 413, 925,  93, 605, 349, 861, 221, 733, 477, 989,
       61, 573, 317, 829, 189, 701, 445, 957, 125, 637, 381, 893, 253, 765, 509, 1021,

        3, 515, 259, 771, 131, 643, 387, 899,  67, 579, 323, 835, 195, 707, 451, 963,
       35, 547, 291, 803, 163, 675, 419, 931,  99, 611, 355, 867, 227, 739, 483, 995,
       19, 531, 275, 787, 147, 659, 403, 915,  83, 595, 339, 851, 211, 723, 467, 979,
       51, 563, 307, 819, 179, 691, 435, 947, 115, 627, 371, 883, 243, 755, 499, 1011,

       11, 523, 267, 779, 139, 651, 395, 907,  75, 587, 331, 843, 203, 715, 459, 971,
       43, 555, 299, 811, 171, 683, 427, 939, 107, 619, 363, 875, 235, 747, 491, 1003,
       27, 539, 283, 795, 155, 667, 411, 923,  91, 603, 347, 859, 219, 731, 475, 987,
       59, 571, 315, 827, 187, 699, 443, 955, 123, 635, 379, 891, 251, 763, 507, 1019,

        7, 519, 263, 775, 135, 647, 391, 903,  71, 583, 327, 839, 199, 711, 455, 967,
       39, 551, 295, 807, 167, 679, 423, 935, 103, 615, 359, 871, 231, 743, 487, 999,
       23, 535, 279, 791, 151, 663, 407, 919,  87, 599, 343, 855, 215, 727, 471, 983,
       55, 567, 311, 823, 183, 695, 439, 951, 119, 631, 375, 887, 247, 759, 503, 1015,

       15, 527, 271, 783, 143, 655, 399, 911,  79, 591, 335, 847, 207, 719, 463, 975,
       47, 559, 303, 815, 175, 687, 431, 943, 111, 623, 367, 879, 239, 751, 495, 1007,
       31, 543, 287, 799, 159, 671, 415, 927,  95, 607, 351, 863, 223, 735, 479, 991,
       63, 575, 319, 831, 191, 703, 447, 959, 127, 639, 383, 895, 255, 767, 511, 1023
    };
#ifdef VERIFY_ROW_SEQ
    for (unsigned int n=0; n<1023; n++) {
        unsigned int i = n;
        unsigned int r = 0;
        for (unsigned int j=0; j<10; j++) {
            r <<=1;
            r |= (i & 1);
            i >>=1;
        }
        if (r!=row_seq[n]) {
            CPLDebug("QCT",
                     "QC3 row_seq[%u]=%u != %u",
                     n, row_seq[n], r);
        }
    }
#endif //  VERIFY_ROW_SEQ

    const unsigned int nQC3_overview_tile_size = QC3_TILE_SIZE >> (1+iOverview);
    const unsigned int nQC3_overview_tile_pixels =
        nQC3_overview_tile_size * nQC3_overview_tile_size;

    CPLDebug("QCT", "read_QCthree_Tile(%p, %d, %d, %p) band %p overview %d",
             fp, tile_xx, tile_yy, pImage,
             this, iOverview);
    CPLDebug("QCT", "iOverview %d nQC3_overview_tile_size = %u nQC3_overview_tile_pixels = %u",
             iOverview, nQC3_overview_tile_size, nQC3_overview_tile_pixels);
    
    memset(tile_data, nQCTNoData, nQC3_overview_tile_pixels);

    // Calculate pointer into pImage for each row in this tile
    for (unsigned int row=0; row<nQC3_overview_tile_size; row++)
    {
        // Top of image
        row_ptr[row] = static_cast<unsigned char*>(pImage);
        // Interleaved rows in the above sequence
        row_ptr[row] += row_seq[row<<(1+iOverview)] * nQC3_overview_tile_size;
    }

    if (iOverview>=0) {
        CPLDebug("QCT", "iOverview[%d] = %p GetOverview() returns %p",
                 iOverview,
                 poQDS->papoOverviewBands[iOverview],
                 GetOverview(iOverview)
                 );
    }
    
    
#endif // ! QC3_IS_DUMMY
} // QCTRasterBand::read_QCthree_Tile

/************************************************************************/
/*                            GetFileList()                             */
/************************************************************************/

char **QCTDataset::GetFileList()
{
    char **papszFileList = GDALPamDataset::GetFileList();

    CPLDebug("QCT", "GetDescription %s", GetDescription() );

    if (poQC3_header && poQC3_header->osQC3Filename.size() != 0 &&
        CSLFindString(papszFileList, poQC3_header->osQC3Filename) == -1)
    {
        papszFileList = CSLAddString( papszFileList, poQC3_header->osQC3Filename );
    }

    return papszFileList;
} // QCTDataSet::GetFileList

/************************************************************************/

void dumpPGM(unsigned int width,
             unsigned int height,
             unsigned char* const data,
             unsigned int rowlength,
             CPLString osBaseLabel,
             unsigned int nMaxPGM
             )
{
    // Is static the best way to count the PGMs ?
    static unsigned int nPGMcount = 0;

    CPLDebug("QCT PGM",
             "dumpPGM(%d %d %p %d %s %u) count %u",
             width, height, data, rowlength,
             osBaseLabel.c_str(), nMaxPGM,
             nPGMcount
             );
    if (osBaseLabel==nullptr) {
        CPLDebug("QCT PGM",
                 "dumpPGM: null osBaseLabel\n");
        return;
    }
    if (rowlength==0) {
        rowlength=width;
        CPLDebug("QCT PGM",
                 "dumpPGM(... %d %s) no rowlength, setting to width = %d",
                 0, osBaseLabel.c_str(), rowlength);
    }

    CPLString osPGMname =
        CPLString().Printf("%s.%05d.pgm", osBaseLabel.c_str(), nPGMcount
                           );
    if (osPGMname==nullptr) {
        CPLDebug("QCT PGM",
                 "osPGMname truncated %s %d",
                 osBaseLabel.c_str(), nPGMcount);
    }
    char const*pszPGMname = osPGMname.c_str();

    if (nPGMcount>10 && nPGMcount>nMaxPGM) {
         CPLDebug("QCT PGM",
             "... too many PGM files; skipping  %s",
             pszPGMname
             );
         nPGMcount++;
         return;
    }

    VSILFILE *fpPGM = VSIFOpenL(pszPGMname, "w");
    nPGMcount++;
    if (fpPGM == nullptr) {
        CPLDebug("QCT PGM",
                 "PGM data dump file %s failed; errno=%d %m",
                 pszPGMname, errno
                 );
        return;
    }

    const size_t nHeaderBufSize = 40;
    char acHeaderBuf[nHeaderBufSize]="";
    size_t nHeaderSize=0;

    nHeaderSize=static_cast<size_t>
        (snprintf(acHeaderBuf, nHeaderBufSize,
                  "P5\n%u %u\n255\n",
                  width, height) );
    
    // snprintf may return negative values;
    // the cast to size_t converts these to large positive
    // values, so we only need one test.
    if (nHeaderSize>=nHeaderBufSize) {
        CPLError( CE_Failure, CPLE_AppDefined,
                  "dumpPGM error generating header for %s\n",
                  pszPGMname);
        VSIFCloseL(fpPGM);
        return;
    }
    size_t nHeaderWriteResult=VSIFWriteL(acHeaderBuf, 1, nHeaderSize, fpPGM);
    if (nHeaderSize==nHeaderWriteResult) {
        const unsigned char* pRow = data;
        for (unsigned int r=0; r<height; r++) {
            size_t rowwriteresult=VSIFWriteL(pRow, 1, width, fpPGM);
#if 1 // Noisy
            if (getenv("QCT_NOISY")) {
                CPLDebug("QCT PGM",
                         "dumpPGM: wrote(%p, 1, %d, %p) row %u returned %zu",
                         pRow, width, fpPGM, r, rowwriteresult);
            }
#endif // Noisy
            if (width!=rowwriteresult) {
                int nWriteErr = errno;
                CPLError( CE_Failure, CPLE_AppDefined,
                          "dumpPGM error writing %s row %u: errno=%d %m",
                          pszPGMname, r, nWriteErr);
                break;
            }
            pRow += rowlength;
        } // for row r
    } else { // nHeaderSize!=nHeaderWriteResult
        // CPLError( CE_Failure, CPLE_AppDefined,
        CPLDebug("QCT PGM",
                 "dumpPGM error writing header for %s errno=%u %m",
                 pszPGMname, errno);
    }
    
    if (0!=VSIFCloseL(fpPGM)) {
        CPLDebug("QCT PGM",
                 "Failed to close PGM data dump file %s; errno=%d",
                 pszPGMname, errno
                 );
    } else {
        CPLDebug("QCT PGM",
                 "PGM data dumped to file %s",
                 pszPGMname
                 );
    }

    return;

} // dumpPGM

#endif /* def FRMT_qct */
