/*
 * $Id: qctvector.cpp,v 1.55 2021/01/15 12:50:11 werdna Exp werdna $
 *
 * Copyright (c) 2016-20, Andrew C Aitchison <qct@aitchison.me.uk>
 *
 * Derived from
 * http://www.gdal.org/ogr_drivertut.html
 * (now https://gdal.org/tutorials/vector_driver_tut.html)
 */

/* -*- tab-width: 4 ; indent-tabs-mode: nil ; c-basic-offset 'tab-width -*- */

/*

GDAL: OGR Driver Implementation Tutorial

   GDAL
    
      Main Page
      Related Pages
      Classes
      Files
    
OGR Driver Implementation Tutorial   

Overall Approach

In general new formats are added to OGR by implementing format
specific drivers with instantiating a GDALDriver and subclasses of
GDALDataset and OGRLayer. The GDALDriver instance is registered with
the GDALDriverManager at runtime.
Before following this tutorial to implement an OGR driver, please
review the OGR Architecture document carefully.
The tutorial will be based on implementing a simple ascii point format.

Contents

Implementing GDALDriver 

Basic Read Only Data Source 

Read Only Layer 

Implementing GDALDriver
The format specific driver class is implemented as a instance of
GDALDriver. One instance of the driver will normally be created, and
registered with the GDALDriverManager. The instantiation of the driver
is normally handled by a global C callable registration function,
similar to the following placed in the same file as the driver class.

*/

#if __cplusplus < 201103 // ie before clang++-6.0
#define override
#define nullptr 0
#endif

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdocumentation-unknown-command"
#pragma clang diagnostic ignored "-Wunused-template"
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wunknown-warning-option"
#pragma clang diagnostic ignored "-Wimplicit-int-float-conversion"
#include "gdal.h"
#pragma clang diagnostic pop
#pragma clang diagnostic ignored "-Wdocumentation-deprecated-sync"
#pragma clang diagnostic ignored "-Wmissing-prototypes"
#pragma clang diagnostic ignored "-Winconsistent-missing-destructor-override"
#pragma clang diagnostic ignored "-Wweak-vtables"
#include "ogrsf_frmts.h"
#include "ogr_spatialref.h"
#pragma clang diagnostic pop
#include "cpl_port.h"
#include "gdal_pam.h"
#include "ogr_api.h"

#include "qct.h"

#ifdef CODE_ANALYSIS

// Printing variables with CPLDebug can hide
// the fact that they are not otherwise used ...
#define CPLDebug(...)

#endif // CODE_ANALYSIS

#include "cpl_conv.h"
//#include "cpl_string.h"
//#include "cpl_error.h"
#include "ogr_p.h"
//#include "ogr_feature.h"
#include "ogrsf_frmts.h"

CPL_CVSID("$Id: qctvector.cpp,v 1.55 2021/01/15 12:50:11 werdna Exp werdna $");

CPL_C_START
#if GDAL_VERSION_MAJOR >= 2
extern "C" void   GDALRegister_qctvector(void);
#else
void CPL_DLL  RegisterOGRQCT();
#endif
CPL_C_END




/* static */ extern GDALDataset* OGRQCTDriverOpen(GDALOpenInfo* poOpenInfo);

static int          OGRQCTDriverIdentify(GDALOpenInfo* poOpenInfo);
static GDALDataset* OGRQCTDriverCreate(const char* pszName,
                                       int nBands, 
                                       int nXSize, int nYSize,
                                       GDALDataType eDT,
                                       char** papszOptions);

/************************************************************************/
/*                             OGRQCTDriver                             */
/************************************************************************/

class OGRQCTDriver : public OGRSFDriver
{
  public:
                OGRQCTDriver();
                ~OGRQCTDriver() override;

    const char*         GetName() override;
    OGRDataSource*      Open( const char *, int ) override;

    virtual int         TestCapability( const char * ) override;
};

#if GDAL_VERSION_MAJOR < 2
OGRQCTDriver::OGRQCTDriver()
{
    // bInitialized = false;
}

OGRQCTDriver::~OGRQCTDriver()
{
    CPLDebug("QCTvector", "%p->~OGRQCTDriver", this);
    // if( bInitialized && !CSLTestBoolean(
    if( !CSLTestBoolean(
            CPLGetConfigOption("IN_GDAL_GLOBAL_DESTRUCTOR", "NO")) )
    {
        // bInitialized = false;
        // odUninitialize();
    }
}
#endif

const char *OGRQCTDriver::GetName()
{
    return "QCTvector";
}

/*
Read Only Layer

The OGRQCTLayer implements layer semantics for an .qct file.
It provides access to a set of feature objects in a consistent
coordinate system with a particular set of attribute columns.
Our class definition looks like this:
*/

class OGRQCTDataSource ;

class OGRQCTLayer : public OGRLayer
{
    friend class OGRQCTDataSource;
    OGRFeatureDefn     *poFeatureDefn;

    fid_t               nFIDcount, nNextFID;

    GBool               genBBox(OGRFeature *poFeature,
                                int top, int left, int bottom, int right );
    OGRQCTDataSource   *poQCTdatasource;
    GBool               setOGRQCT_Outline_Layer();
    GBool               setOGRQCT_GCP_Layer();
    GBool               setOGRQCT_BBox_Layer();
    OGRFeature         *poFeatures[2];
public:

    explicit OGRQCTLayer( const char *pszFilename,
                          OGRQCTDataSource *poQCTdatasourceIn );
   ~OGRQCTLayer() override;

    void                ResetReading() override;
#if GDAL_VERSION_MAJOR < 2
    OGRFeature *        GetFeature( fid_t nFID );
#else
    OGRFeature *        GetFeature( fid_t nFID ) override;
#endif
    OGRFeature *        GetNextFeature() override;

    OGRFeatureDefn *    GetLayerDefn() override { return poFeatureDefn; }

    int                 TestCapability( const char * ) override { return false; }
};

/*
Basic Read Only Data Source

We will start implementing a minimal read-only datasource. No attempt
is made to optimize operations, and default implementations of many
methods inherited from GDALDataset are used.
The primary responsibility of the datasource is to manage the list of
layers. In the case of the QCT format a datasource is a single file
representing one layer so there is at most one layer
(we should, but do not yet, make the QC3 bounding box a separate layer). 

 The "name" of a datasource should generally be the name passed to the
Open() method.  The Open() method below is not overriding a base class
method, but we have it to implement the open operation delegated by
the driver class.

For this simple case we provide a stub TestCapability() that returns
false for all extended capabilities. The TestCapability() method is
pure virtual, so it does need to be implemented.

*/

class OGRQCTDataSource : public GDALDataset
{
    friend class OGRQCTLayer;

    OGRQCTLayer        *papoLayers[3];
    int                 nLayers;
    char               *pszQCTFilename;
    // CPLString           pszQC3Filename;
    QC3Header          *poQC3_header;
    // unsigned int        width3;
    // unsigned int        height3;
    double              datumShift[2];
    double              adfQCTpolynomial[40];
    int                 nOutlinePts;
  public:
                        OGRQCTDataSource();
                        ~OGRQCTDataSource() override;

    int                 Open( const char * pszFilename, int bUpdate );
    VSILFILE           *fp;
#if GDAL_VERSION_MAJOR < 2
    int                 GetLayerCount() { return nLayers; }
    OGRLayer            *GetLayer( int ) ;

    int                 TestCapability( const char * ) { return false; }
#else
    int                 GetLayerCount() override { return nLayers; }
    OGRLayer            *GetLayer( int ) override;

    int                 TestCapability( const char * ) override { return false; }
#endif
};

/*
The constructor is a simple initializer to a default state.
The Open() will take care of actually attaching it to a file.
The destructor is responsible for orderly cleanup of layers.
*/

OGRQCTDataSource::OGRQCTDataSource() :
    papoLayers{nullptr,nullptr,nullptr},
    nLayers(0),
    pszQCTFilename(nullptr),
    // pszQC3Filename(nullptr),
    poQC3_header(nullptr),
    datumShift{0.0,0.0},
    adfQCTpolynomial{0.0},
    nOutlinePts(0),
    fp(nullptr)
{
    for(int i=0;i<3;i++) {
        papoLayers[i]=nullptr;
    }
#if 0
    for(int i=0;i<40;i++) {
         adfQCTpolynomial[i]=0.0;
    }
#endif
}


OGRQCTDataSource::~OGRQCTDataSource()
{
    CPLDebug("QCTvector", "%p->~OGRQCTDataSource", this);
    // if (papoLayers!=nullptr)
    {
        for (int i=0; i < nLayers; i++) {
            if (papoLayers[i]!=nullptr) {
                delete papoLayers[i];
                papoLayers[i]=nullptr;
            }
        }
        // CPLFree( papoLayers );
    }
    if(pszQCTFilename) {
        CPLFree( pszQCTFilename );
    }
#if 0
    if (pszQC3Filename) {
        delete pszQC3Filename;
    }
#endif
    if( poQC3_header != nullptr ){
        delete poQC3_header;
        poQC3_header = nullptr;
    }
    if( fp != nullptr ) {
        VSIFCloseL( fp );
    }
}

/*
The Open() method is called by GDALOpenEx(). It should quietly return
nullptr if the passed filename is not of the format supported by the
driver. If it is the target format, then a new GDALDataset object for
the dataset should be returned.
It is common for the Open() method to be delegated to an Open() method
on the actual format's GDALDataset class.
*/

/*

The Open() method is the most important one on the datasource, though
in this particular instance it passes most of it's work off to the
OGRQCTLayer constructor if it believes the file is of the desired
format.
Note that Open() methods should try and determine that a file isn't of
the identified format as efficiently as possible, since many drivers
may be invoked with files of the wrong format before the correct
driver is reached. In this particular Open() we just test the file
extension but this is generally a poor way of identifying a file
format. If available, checking "magic header values" or something
similar is preferable.
In the case of the QCT format update is not supported,
so we always fail if bUpdate is true.
*/

int  OGRQCTDataSource::Open( const char *pszFilename, int bUpdate )
{
    if( bUpdate ) {
        CPLError( CE_Failure, CPLE_OpenFailed, 
                  "Update access not supported by the QCT driver." );
        return false;
    }

#ifdef QCT_USE_GCP
    CPLDebug("QCTvector", "QCT with GCP");
#else
    CPLDebug("QCTvector", "QCT without GCP");
#endif

    int nQCTMagic = QCTreadInt(fp, 0);
    int nQCTVersion = QCTreadInt(fp, 4);

    nRasterXSize = QCT_TILE_SIZE * QCTreadInt( fp, 8 );
    nRasterYSize = QCT_TILE_SIZE * QCTreadInt( fp, 12 );
    if  (nRasterXSize <= 0 || nRasterYSize <= 0 ) {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Invalid dimensions : %d x %d", 
                  nRasterXSize, nRasterYSize); 
        return false;
    }

    size_t nExtendedDataIndex = static_cast<size_t>
        (QCTreadInt( fp, 0x54));
    nOutlinePts = QCTreadInt( fp, 0x58 );
    
    size_t nDatumShiftIndex = static_cast<size_t>
        (QCTreadInt(fp, nExtendedDataIndex+4));
    int nDoublesRead =
        QCTreadDoubles(adfQCTpolynomial, fp,
                       static_cast<size_t>(0x60),
                       static_cast<size_t>(40) );
    if ( 40!=nDoublesRead ) {
        CPLDebug("QCTvector", "problem reading coordinate polynomial");
        return false;
    }

    if (nDatumShiftIndex>0x45A0) {
        nDoublesRead =
            QCTreadDoubles(datumShift, fp,
                           nDatumShiftIndex, static_cast<size_t>(2) );
        if ( 2!=nDoublesRead ) {
            CPLDebug("QCTvector", "problem reading datum shift");
            return false;
        }
        CPLDebug("QCTvector",
                 "datumShift: %g north, %g east at index x%zx[4]=x%zx",
                 datumShift[0], datumShift[1],
                 nExtendedDataIndex, nDatumShiftIndex
                 );
    } else if (nDatumShiftIndex!=0) {
        CPLDebug("QCTvector",
                 "datumShift index x%zx[4]=x%zx points into header. Ignoring",
                 nExtendedDataIndex, nDatumShiftIndex
                 );
    }

    pszQCTFilename = CPLStrdup( pszFilename );

    poQC3_header = new QC3Header;
    if(!poQC3_header->Open(pszQCTFilename, false, nQCTMagic, nQCTVersion) ) {
        CPLDebug("QCTvector", "QC3 file not opened");
        delete poQC3_header;
        poQC3_header=nullptr;
    }
    
// -------------------------------------------------------------------- 
//      Create the layers.
// --------------------------------------------------------------------
    // papoLayers = static_cast<OGRQCTLayer **>(CPLMalloc(3*sizeof(void*)));
    for (int nL=0; nL<3; nL++) {papoLayers[nL]=nullptr;}
    nLayers = 0;
    for (int nL=0; nL<3; nL++) {
        switch(nL) {
        case 0:
#if 1
            if (nOutlinePts >0) {
                // papoLayers[nLayers] = new OGRQCTLayer(pszFilename, this);
                papoLayers[nLayers] = new OGRQCTLayer("outline", this);
                if (papoLayers[nLayers]->setOGRQCT_Outline_Layer()) {
                    CPLDebug("QCTvector",
                             "Set outline layer %d for %s succeeded",
                             nLayers, pszFilename);
                } else {
                    CPLDebug("QCTvector",
                             "Failed to set outline layer for %s",
                             pszFilename);
                }
            } else {
                CPLDebug("QCTvector",
                         "No outline in %s",
                         pszFilename);
            }
#else
            CPLDebug("QCTvector",
                     "Setting outline layer not currently available");
#endif
            break;
        case 1:
#ifdef QCT_USE_GCP
#    if 1
            papoLayers[nLayers] = new OGRQCTLayer("GCP grid", this);
            if (papoLayers[nLayers]->setOGRQCT_GCP_Layer()) {
                CPLDebug("QCTvector", "Set GCP layer %d for %s succeeded",
                         nLayers, pszFilename);
            } else {
                CPLDebug("QCTvector",
                         "Failed to set GCP layer for %s",
                         pszFilename);
            }
#    else
            CPLDebug("QCTvector",
                     "Setting GCP layer not currently available");
#    endif
            break;
        case 2:
#endif

#if 1
            papoLayers[nLayers] = new OGRQCTLayer("BBox(es)", this);
            if (papoLayers[nLayers]->setOGRQCT_BBox_Layer()) {
                CPLDebug("QCTvector", "Set BBox layer %d for %s succeeded",
                         nLayers, pszFilename);
            } else {
                CPLDebug("QCTvector",
                         "Failed to set BBox layer for %s",
                         pszFilename);
            }
#else
            CPLDebug("QCTvector",
                     "Setting BBox(es) layer not currently available");
#endif
            break;
        default:
            break;
        }
        if(papoLayers[nLayers]) {
            nLayers++;
        }
    } // for (int nL=0; nL<3; nL++)

    return true;
} // OGRQCTDataSource::Open()


// static
GDALDataset *OGRQCTDriverOpen( GDALOpenInfo* poOpenInfo )
{
    if( !OGRQCTDriverIdentify(poOpenInfo) )
        return nullptr;

    if( poOpenInfo->eAccess == GA_Update ) {
        CPLError( CE_Failure, CPLE_NotSupported,
                  "The QCTvector driver does not support update access to existing"
                  " datasets.\n" );
        return nullptr;
    }

#ifdef QCT_USE_GCP
    CPLDebug("QCTvector", "QCTvector with GCP");
#else
    CPLDebug("QCTvector", "QCTvector without GCP");
#endif

    OGRQCTDataSource   *poDS = new OGRQCTDataSource();

#if 0 // GDAL_VERSION_MAJOR >= 2
    // Borrow the file pointer from GDALOpenInfo* ?
    // This may "warn" that we cannot open the file *for writing*
    poDS->fp = poOpenInfo->fpL;
#else
    poDS->fp = VSIFOpenL( poOpenInfo->pszFilename, "rb" );
    if (poDS->fp == nullptr)
    {
        delete poDS;
        return nullptr;
    }
#endif
    if( !poDS->Open( poOpenInfo->pszFilename, false ) ) {
        CPLDebug("QCTvector", "eAccess %08x unexpected",
                 poOpenInfo->eAccess);
        delete poDS;
        return nullptr;
    }
#if 0 //  GDAL_VERSION_MAJOR >= 2
    poDS->fp = poOpenInfo->fpL;
    poOpenInfo->fpL = nullptr;
#endif
    return poDS;

} // OGRQCTDriverOpen()

/*
The Identify() method is implemented as such :
*/
static int OGRQCTDriverIdentify( GDALOpenInfo* poOpenInfo )
{
    if (poOpenInfo->pabyHeader == nullptr ||
#if GDAL_VERSION_MAJOR >= 2
        poOpenInfo->fpL == nullptr
#else
        poOpenInfo->fp == nullptr
#endif
        )
    {
#if 0
        CPLDebug("QCTvector", "%s not yet opened",
                 poOpenInfo->pszFilename);
#endif
        return false;
    }
    unsigned int magic = QCTGetUInt(poOpenInfo->pabyHeader, 0);
    switch (magic) {
    case qct_info_magic: case qct_magic:
        CPLDebug("QCTvector", "%s has expected qct_magic",
                 poOpenInfo->pszFilename);
        return true;
    case qcthree_magic:
        CPLDebug("QCTvector", "%s: please use the the .qct file",
                 poOpenInfo->pszFilename);
        return false;
    default:
#if 0
        CPLDebug("QCTvector", "%s has magic %08x - expected %08x or %08x",
                 poOpenInfo->pszFilename, magic, qct_info_magic, qct_magic);
#endif
        return false;
    }
} // OGRQCTDriverIdentify


/*
A GetLayer() method also needs to be implemented.
Since the layer list is created in the Open()
this is just a lookup with some safety testing.
*/

OGRLayer *OGRQCTDataSource::GetLayer( int iLayer )
{
    if( iLayer < 0 || iLayer >= nLayers) {
        return nullptr;
    }

    return papoLayers[iLayer];
}

/*

The layer constructor is responsible for initialization.
The most important initialization is setting up the OGRFeatureDefn for
the layer. This defines the list of fields and their types, the
geometry type and the coordinate system for the layer.

In the QCT format we have a list of 2D points,
stored as little-endian IEEE-754 floating point double (64 bit) values.
The coordinate system is not declared in the file - it is always WGS 84.

Pay particular attention to the reference counting of the OGRFeatureDefn.
As OGRFeature's for this layer will also take a reference to this
definition it is important that we also establish a reference on
behalf of the layer itself.

*/

GBool OGRQCTLayer::setOGRQCT_Outline_Layer()
{

    nFIDcount=0;
    poFeatureDefn->SetGeomType( wkbPolygon );
    poFeatures[0] = poFeatures[1] = nullptr;

    int nOutlinePts = poQCTdatasource->nOutlinePts;
    if  (nOutlinePts < 0 ) {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Invalid number of outline points %d", 
                  nOutlinePts);
        poQCTdatasource->nOutlinePts=0;
        return false;
    } else if (nOutlinePts >1000 ) {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Too many outline points %d", 
                  nOutlinePts); 
        poQCTdatasource->nOutlinePts=0;
        return false;
    } else if (nOutlinePts == 0 ) {
        CPLDebug("QCTvector", "no outline");
        return false;
    }
    
    CPLDebug("QCTvector", "%d outline points", nOutlinePts); 

    poFeatures[0] = new OGRFeature( poFeatureDefn );
    poFeatures[0]->SetField( 0, "Outline" );
    poFeatures[0]->SetFID( 0 );

    CPLDebug("QCTvector",
             // "%s outline has %d points", pszFilename, nOutlinePts
             "outline has %d points", nOutlinePts
             );
    // For now this is a (closed) LinearRing;
    // should it be a (multi?)polygon ?

    // OGRPolygon* outline
    OGRLinearRing* outline = new OGRLinearRing ();
    outline->setCoordinateDimension(2);
    outline->setNumPoints(nOutlinePts);

    size_t addrQCToutline =
        static_cast<size_t>(QCTreadInt(poQCTdatasource->fp, 0x5C) );
    // Read outline data from file
    if ( VSIFSeekL( poQCTdatasource->fp, addrQCToutline, SEEK_SET ) ) {
        CPLError( CE_Failure, CPLE_AppDefined,
                  "cannot seek to QCT outline points" );
        return false;
    }

    double aPoint[2]={};
    for (int i=0; i<nOutlinePts; i++) {
        size_t entries_read =
            QCTreadDoubles( aPoint, poQCTdatasource->fp,
                            addrQCToutline+2*static_cast<size_t>(i)*sizeof(double), 2 );
        if ( entries_read < 2 )  {
            CPLError( CE_Failure, CPLE_AppDefined,
                      "error reading point %d of %d entries",
                      i, nOutlinePts);
            return false;
        }
        outline->setPoint( i, aPoint[1], aPoint[0] ); // Note order !

        CPLDebug("QCTvector",
                 "outline[%d] %g %g",
                 i, aPoint[1], aPoint[0] );
    }

    OGRPolygon* poly = new OGRPolygon ();
    CPLDebug("QCTvector", "poly created");

    poly->addRingDirectly(outline);
    poly->closeRings();
    CPLDebug("QCTvector","outline added");

    // see http://www.gdal.org/classOGRPolygon.html
    poFeatures[0]->SetGeometryDirectly( poly );
    nFIDcount=1;

    CPLDebug("QCTvector","Outline geometry set");
    return true;
} // OGRQCT_Outline_Layer

GBool OGRQCTLayer::setOGRQCT_GCP_Layer()
{
    nFIDcount=0;
    poFeatureDefn->SetGeomType( wkbMultiLineString );

    if (poFeatures[0]!=nullptr) {
        CPLDebug("QCTvector", "poFeatures[0] already set");
    }
    poFeatures[0] = new OGRFeature( poFeatureDefn );
    poFeatures[0]->SetField( 0, "GCP plot" );
    poFeatures[0]->SetFID( 0 );
    OGRMultiLineString* multi = new OGRMultiLineString();
    CPLDebug("QCTvector", "multi created");
    CPLDebug("QCTvector", "raster %d x %d",
             poQCTdatasource->nRasterXSize, poQCTdatasource->nRasterYSize);

    const double points_in_line=12.0; // 6.0;
    // 4 lines "across", each with 1+points_in_line points
    for (double j=0; j<=3; j++) {
        double y=(poQCTdatasource->nRasterYSize*j)/3.0;

        OGRLineString* poTempGCPcurve = new OGRLineString();
        poTempGCPcurve->setCoordinateDimension(2);

        for (double i=0; i<=points_in_line; i++) {
            double x=(poQCTdatasource->nRasterXSize*i)/points_in_line;
            
            latlon tempLatLong;
            QCT_xy_to_latlon(x,y,
                             &tempLatLong,
                             poQCTdatasource->adfQCTpolynomial,
                             poQCTdatasource->datumShift);
            poTempGCPcurve->addPoint(tempLatLong.lon,
                                     tempLatLong.lat); // Note order !

            CPLDebug("QCTvector",
                     "GCP point[%g][%g] %g %g %g %g",
                     i,j, x,y,
                     tempLatLong.lon,
                     tempLatLong.lat
                     );
        }
        
        multi->addGeometryDirectly(poTempGCPcurve);
        CPLDebug("QCTvector","poTempGCPcurve added");
        
        // poFeature->SetGeometry( poTempGCPcurve );
    } // 4 lines "across", each with 1+points_in_line points
    
    // 4 lines "up/down", each with 1+points_in_line points
    for (double i=0; i<=3; i++) {
        double x=(poQCTdatasource->nRasterXSize*i)/3.0;

        OGRLineString* poTempGCPcurve = new OGRLineString();
        poTempGCPcurve->setCoordinateDimension(2);

        for (double j=0; j<=points_in_line; j++) {
            double y=(poQCTdatasource->nRasterYSize*j)/points_in_line;
            
            latlon tempLatLong;
            QCT_xy_to_latlon(x,y,
                             &tempLatLong,
                             poQCTdatasource->adfQCTpolynomial,
                             poQCTdatasource->datumShift);
            poTempGCPcurve->addPoint(tempLatLong.lon,
                                     tempLatLong.lat); // Note order !

            CPLDebug("QCTvector",
                     "GCP point[%g][%g] %g %g %g %g",
                     i,j, x,y,
                     tempLatLong.lon,
                     tempLatLong.lat
                     );
        }
        
        multi->addGeometryDirectly(poTempGCPcurve);
        CPLDebug("QCTvector","poTempGCPcurve added");
        
        // poFeature->SetGeometry( poTempGCPcurve );
    } // 4 lines "up/down", each with 1+points_in_line points
    
    poFeatures[0]->SetGeometryDirectly( multi );
    //poFeatures[0]->SetGeometry( multi );
    CPLDebug("QCTvector","OGRQCTLayer::setOGRQCT_GCP_Layer() completed");
    
    nFIDcount=1;
    return true;
} // OGRQCTLayer::setOGRQCT_GCP_Layer()

GBool OGRQCTLayer::setOGRQCT_BBox_Layer( )
{
    CPLDebug("QCTvector","entered setOGRQCT_BBox_Layer()");
    poFeatureDefn->SetGeomType( wkbPolygon );

    // .qct BBox
    nFIDcount=0;

    poFeatures[nFIDcount] = new OGRFeature( poFeatureDefn );
    poFeatures[nFIDcount]->SetFID(nFIDcount);
    bool ret=genBBox(poFeatures[nFIDcount], 0, 0,
                     poQCTdatasource->nRasterYSize-1,
                     poQCTdatasource->nRasterXSize-1); // Note order
    if (!ret) {
        delete poFeatures[nFIDcount];
        poFeatures[nFIDcount]=nullptr;
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Bad QCT BBox"); 
        return false;
    }
    poFeatures[nFIDcount]->SetField( 0, "QCT BoundingBox" );
    nFIDcount++;
    // end of.qct BBox

    // QC3 bounding box. Top left is the same as QCT,
    // but bottom and right may extend further
    // - to fill the last 1024x1024 tile.
    if  (poQCTdatasource->poQC3_header==nullptr
         || poQCTdatasource->poQC3_header->width3 == 0
         || poQCTdatasource->poQC3_header->height3 == 0 ) {
        CPLDebug("QCTvector", "No QC3, OK");
        return true;
    }

    poFeatures[nFIDcount] = new OGRFeature( poFeatureDefn );
    ret = genBBox(poFeatures[nFIDcount], 0, 0,
                  QC3_TILE_SIZE*static_cast<int>(poQCTdatasource->poQC3_header->height3),
                  QC3_TILE_SIZE*static_cast<int>(poQCTdatasource->poQC3_header->width3)
                  );
    if (!ret) {
        delete poFeatures[nFIDcount];
        poFeatures[nFIDcount]=nullptr;
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Bad QC3 BBox"); 
        return true; // QCT box still OK.
    }
    poFeatures[nFIDcount]->SetField( 0, "QC3 BoundingBox" );
    poFeatures[nFIDcount]->SetFID(nFIDcount);
    nFIDcount++;
    CPLDebug("QCTvector", "QC3 bbox set, OK");
    
    CPLDebug("QCTvector", "OGRQCT_BBox_Layer() completed");
    return true;
} // OGRQCT_BBox_Layer

OGRQCTLayer::OGRQCTLayer( const char *pszLabel,
                          OGRQCTDataSource *poQCTdatasourceIn )
    :
    poFeatures{nullptr,nullptr}
{
    CPLDebug("QCTvector", "%p->OGRQCTLayer::OGRQCTLayer(%s,%p)",
             this, pszLabel, poQCTdatasourceIn);
    
    if (poQCTdatasourceIn == nullptr) {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "OGRQCTLayer::OGRQCTLayer(%s) given null datasource",
                 pszLabel);
        return;
    }

    if (poQCTdatasourceIn->fp == nullptr) {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "OGRQCTLayer::OGRQCTLayer(%s) given null datasource",
                 pszLabel);
        return;
    }

    poQCTdatasource = poQCTdatasourceIn;

    nNextFID = 0;
    nFIDcount= 0;
    poFeatures[0] = poFeatures[1] = nullptr;
    CPLDebug("QCTvector", "OGRQCTLayer::OGRQCTLayer(%s) entered",
             pszLabel);


    poFeatureDefn = new OGRFeatureDefn( CPLGetBasename( pszLabel ) );
    
#if GDAL_VERSION_MAJOR >= 2
#if GDAL_VERSION_MAJOR >= 3
    // proj6/7
    // OGRSpatialReference* poWGS84_SRS = new OGRSpatialReference(SRS_DN_WGS84);
    OGRSpatialReference* poWGS84_SRS = new OGRSpatialReference(SRS_WKT_WGS84_LAT_LONG);
    poWGS84_SRS->SetAxisMappingStrategy(OAMS_TRADITIONAL_GIS_ORDER);
#else
    // proj4 ?5
    OGRSpatialReference* poWGS84_SRS = new OGRSpatialReference(SRS_WKT_WGS84);
#endif // GDAL_VERSION_MAJOR >= 3

    SetDescription(poFeatureDefn->GetName());
    poFeatureDefn->GetGeomFieldDefn(0)->SetSpatialRef(poWGS84_SRS);
#else // GDAL_VERSION_MAJOR < 2
    //OGRSpatialReference* poWGS84_SRS
    //   = new OGRSpatialReference(SRS_WKT_WGS84);
    // GDAL 1.09 does not have GetGeomFieldDefn
    // and does not have SetSpatialRef here
    // poFeatureDefn->GetFieldDefn(0)->SetSpatialRef(poWGS84_SRS);
#endif
    poFeatureDefn->Reference();

    // From https://gdal.org/tutorials/vector_driver_tut.html
    // but what do these two lines do ?
    OGRFieldDefn oFieldTemplate( "Name", OFTString );
    poFeatureDefn->AddFieldDefn( &oFieldTemplate );
#if GDAL_VERSION_MAJOR >= 2
    poWGS84_SRS->Release();
#endif
} // OGRQCTLayer

/*
Note that the destructor uses Release() on the OGRFeatureDefn.
This will destroy the feature definition if the reference count drops to zero,
but if the application is still holding onto a feature from this layer,
then that feature will hold a reference to the feature definition
and it will not be destroyed here (which is good!).
*/

OGRQCTLayer::~OGRQCTLayer()
{
    CPLDebug("QCTvector", "%p->~OGRQCTLayer", this);
    for (int i=0; i<2; i++) {
        if (poFeatures[i]!=nullptr) {
            delete poFeatures[i];
            poFeatures[i]=nullptr;
        }
    }
    if(poFeatureDefn) {
        poFeatureDefn->Release();
    }
} // OGRQCTLayer::~OGRQCTLayer

GBool OGRQCTLayer::genBBox(OGRFeature *poFeature, int top, int left, int bottom, int right )
{
    // poFeature = new OGRFeature( poFeatureDefn ); // created by caller

    int nBoxWidth = right-left;
    int nBoxHeight = bottom-top;
    if  (nBoxWidth <= 0 || nBoxHeight <= 0 ) {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Invalid dimensions : %d x %d", 
                  nBoxWidth, nBoxHeight); 
        return false;
    }
 
    double lat=poQCTdatasource->adfQCTpolynomial[20];
    double lon=poQCTdatasource->adfQCTpolynomial[30];
    CPLDebug("QCTvector",
             "lat %g long %g\nnorth shift %g east shift %g\n",
             lat, lon,
             poQCTdatasource->datumShift[1],
             poQCTdatasource->datumShift[0]
             );
    (void)lat; (void)lon;
    latlon TL; //= {lat+datumShift[1],     lon+datumShift[0]};
    latlon TR; //= {lat+datumShift[1]+1.0, lon+datumShift[0]};
    latlon BL; //= {lat+datumShift[1],     lon+datumShift[0]-1.0};
    latlon BR; //= {lat+datumShift[1]+1.0, lon+datumShift[0]-1.0};
            
#if GDAL_VERSION_MAJOR >= 2
    QCT_xy_to_latlon(
                     left, top, &TL,
                     poQCTdatasource->adfQCTpolynomial, poQCTdatasource->datumShift );
    QCT_xy_to_latlon(
                     right, top, &TR,
                     poQCTdatasource->adfQCTpolynomial, poQCTdatasource->datumShift );
    QCT_xy_to_latlon(
                     left, bottom, &BL,
                     poQCTdatasource->adfQCTpolynomial, poQCTdatasource->datumShift );
    QCT_xy_to_latlon(
                     right, bottom, &BR,
                     poQCTdatasource->adfQCTpolynomial, poQCTdatasource->datumShift );
#endif
    
    // Based on OGRBNALayer::BuildFeatureFromBNARecord
    //    bnaFeatureType == BNA_POLYLINE
    OGRLinearRing* linearRing = new OGRLinearRing ();

    linearRing->setCoordinateDimension(2);
    linearRing->setNumPoints(4);

    linearRing->setPoint( 0, TL.lon, TL.lat );
    linearRing->setPoint( 1, TR.lon, TR.lat );
    linearRing->setPoint( 2, BR.lon, BR.lat );
    linearRing->setPoint( 3, BL.lon, BL.lat );

    CPLDebug("QCTvector",
             "\ntop %g %g\nbottom %g %g\nleft %g %g\nright %g %g\n",
             TL.lat, TR.lat,
             BL.lat, BR.lat,
             TL.lon, BL.lon,
             TR.lon, BR.lon
             );

    OGRPolygon* poly = new OGRPolygon ();
    
    poly->addRingDirectly(linearRing);
    poly->closeRings();

    poFeature->SetGeometryDirectly(poly);

    return true;
    
} // OGRQCTLayer::genBBox

OGRFeature *  OGRQCTLayer::GetFeature( fid_t nFID )
{
    OGRFeature *ret=nullptr;

    if(nFIDcount<0 || nFIDcount>1) {
        CPLDebug("QCTvector",
                 "%p->GetFeature(%lld) called with unexpected nFIDcount %lld",
                 this,
                 static_cast<long long>(nFID),
                 static_cast<long long>(nFIDcount) );        
    }

    if (0<=nFID && nFID<nFIDcount && nFID<2) {
        ret = poFeatures[nFID]->Clone();
    }

    CPLDebug("QCTvector",
             "%p->GetFeature(%lld of %lld) returns %p",
             this,
             static_cast<long long>(nFID),
             static_cast<long long>(nFIDcount), ret);

    return ret;
} // OGRQCTLayer::GetFeature


/*
The GetNextFeature() method is usually the work horse of OGRLayer
implementations. It is responsible for reading the next feature
according to the current spatial and attribute filters installed.
The while() loop is present to loop until we find a satisfactory feature.
The first section of code is for parsing a single line of the
QCT text file and establishing the x, y and name for the line.
*/

OGRFeature *OGRQCTLayer::GetNextFeature()
{    
    do {
        OGRFeature *poFeature = GetFeature( nNextFID++ );

        if ( poFeature == nullptr ) {
            return nullptr;
        }

        /*
          Next we check if the feature matches our current attribute
          or spatial filter if we have them. Methods on the OGRLayer
          base class support maintain filters in the OGRLayer member
          fields m_poFilterGeom (spatial filter) and m_poAttrQuery
          (attribute filter) so we can just use these values here if
          they are non-nullptr. The following test is essentially "stock"
          and done the same in all formats. Some formats also do some
          spatial filtering ahead of time using a spatial index.

          If the feature meets our criteria we return it.
          Otherwise we destroy it, and return to the top of the loop
          to fetch another to try.
        */

        if ( (m_poFilterGeom == nullptr
             || FilterGeometry( poFeature->GetGeometryRef() ) )
            && (m_poAttrQuery == nullptr
                || m_poAttrQuery->Evaluate( poFeature )) 
             )
            {
                return poFeature;
            }

        delete poFeature;
    } while (nNextFID<nFIDcount);

    nNextFID = 0;
    return nullptr;
} // OGRQCTLayer::GetNextFeature()

/*
While in the middle of reading a feature set from a layer,
or at any other time the application can call ResetReading()
which is intended to restart reading at the beginning of the feature set.
We implement this by seeking back to the beginning of the file,
and resetting our feature id counter.
*/

void OGRQCTLayer::ResetReading()
{
    VSIFSeekL( poQCTdatasource->fp, 0, SEEK_SET );
    nNextFID = 0;
}

/*
The SetDescription() sets the name of the driver. This name is
specified on the commandline when creating datasources so it is
generally good to keep it short and without any special characters or
spaces.
SetMetadataItem( GDAL_DCAP_VECTOR, "YES" ) is specified to indicate
that the driver will handle vector data.
SetMetadataItem( GDAL_DCAP_VIRTUALIO, "YES" ) is specified to indicate
that the driver can deal with files opened with the VSI*L GDAL
API. Otherwise this metadata item should not be defined.
The driver declaration generally looks something like this for a
format with read or read and update access (the Open() method) and
creation support (the Create() method).
*/

// Defined in GDAL v1.11.5  but not v1.09
#ifndef CPL_UNUSED
#define CPL_UNUSED
#endif

static GDALDataset *OGRQCTDriverCreate( const char * pszFilename,
                                          CPL_UNUSED int nBands,
                                          CPL_UNUSED int nXSize,
                                          CPL_UNUSED int nYSize,
                                          CPL_UNUSED GDALDataType eDT,
                                          CPL_UNUSED char **papszOptions )
{
    OGRQCTDataSource *poDS = new OGRQCTDataSource();

    if( !poDS->Open( pszFilename, true ) )
    {
        delete poDS;
        CPLError( CE_Failure, CPLE_AppDefined,
                  "QCTvector driver doesn't currently support file creation.\n"
                  // "Please create database with the `createdb' command."
                  );
        return nullptr;
    }

    return poDS;
}


extern "C" void GDALRegister_qctvector()
{

    CPLDebug("QCTvector", "registering QCTvector driver");

    if( GDALGetDriverByName( "QCTvector" ) == nullptr )
    {
        GDALDriver  *poDriver = new GDALDriver();

        poDriver->SetDescription( "QCTvector" );
#if GDAL_VERSION_MAJOR >= 2
        poDriver->SetMetadataItem( GDAL_DCAP_VECTOR, "YES" );
#endif
        poDriver->SetMetadataItem( GDAL_DMD_LONGNAME,
                                   "Memory Map QCT vector driver" );
        poDriver->SetMetadataItem( GDAL_DMD_EXTENSION, "qct" );
        poDriver->SetMetadataItem( GDAL_DMD_HELPTOPIC,
                                   "drv_qct.html" );

        poDriver->pfnOpen = OGRQCTDriverOpen;
        poDriver->pfnIdentify = OGRQCTDriverIdentify;
        poDriver->pfnCreate = OGRQCTDriverCreate;

        poDriver->SetMetadataItem( GDAL_DCAP_VIRTUALIO, "YES" );

        // See https://gdal.org/development/rfc/rfc34_license_policy.html
        poDriver->SetMetadataItem( "LICENSE_POLICY", "NONRECIPROCAL" );

        GetGDALDriverManager()->RegisterDriver( poDriver );
    }
} // GDALRegister_qctvector

/************************************************************************/
/*                           RegisterOGRQCT()                           */
/************************************************************************/

#if GDAL_VERSION_MAJOR < 2
void RegisterOGRQCT()
{
    if (! GDAL_CHECK_VERSION("OGR/QCT driver")) {
        CPLDebug("QCTvector", "QCT vector - wrong gdal version\n");
        return;
    }
    CPLDebug("QCTvector", "registering qctvector plugin\n");
    OGRSFDriverRegistrar::GetRegistrar()->RegisterDriver( new OGRQCTDriver );
}
#endif


/*
There! We have completed a simple read-only feature file format driver. 
*/

