/******************************************************************************
 * $Id: qed.cpp,v 1.6 2021/01/30 14:58:10 werdna Exp werdna $
 *
 * Project:  GDAL
 * Purpose:  QED (Memory Map Quick Elevation Data) Driver
 * Author:   Andrew C Aitchison
 ******************************************************************************
 * Copyright (c) 2014-21, Andrew C Aitchison <qct@aitchison.me.uk>
 ****************************************************************************/
 
/* -*- tab-width: 4 ; indent-tabs-mode: nil ; c-basic-offset 'tab-width -*- */

#ifdef FRMT_qct
#ifdef FRMT_qed

#if __cplusplus < 201103 // ie before clang++-6.0
#define override
#define nullptr 0
#endif

#include "time.h"   // to print nOrigFileTime

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Winconsistent-missing-destructor-override"
#pragma clang diagnostic ignored "-Wdocumentation-unknown-command"
#pragma clang diagnostic ignored "-Wweak-vtables"
#pragma clang diagnostic ignored "-Wunused-template"
#pragma clang diagnostic ignored "-Wunknown-warning-option"
#pragma clang diagnostic ignored "-Wimplicit-int-float-conversion"
#include "gdal_pam.h"
#pragma clang diagnostic pop
#include "ogr_spatialref.h"

#include "qct.h"

// Set *one* of these definitions (to 1)
// QC3_PIXEL_IS_PIXEL is to be assumed if none are set.
// #define QC3_PIXEL_IS_PIXEL 1
#define QC3_PIXEL_IS_TILE 1

CPL_CVSID("$Id: qed.cpp,v 1.6 2021/01/30 14:58:10 werdna Exp werdna $")

CPL_C_START

void   GDALRegister_qed(void);
CPL_C_END

#if GDAL_VERSION_NUM < 2010000
#define GDAL_IDENTIFY_UNKNOWN -1
#define GDAL_IDENTIFY_FALSE 0
#define GDAL_IDENTIFY_TRUE 1
#else
// These are defined in gdal/gdal_priv.h
#endif // GDAL_VERSION_NUM < 2010000

#ifdef CODE_ANALYSIS

// Printing variables with CPLDebug can hide
// the fact that they are not otherwise used ...
#define CPLDebug(...)

#endif // CODE_ANALYSIS

static const signed int nQEDNoData = -32768;

#if 0 // not currently used
static short QEDGetShort( short *pszField )
{
    /* this is fine for little-endian, but will need changing for bigendian */
    return *pszField;
}
#endif

int QEDFGetc(VSILFILE *fp) {
#if 0
    return VSIFGetc(fp);
#else
    char buf;
    size_t res = VSIFReadL(&buf, 1, 1, fp);
#if 0
    CPLDebug("QED",
             "QEDFGetc(%p): got x%x=%d and res x%zx",
             fp, buf, buf, res);
#endif
    if (1==res) {
        return buf;
    }
    vsi_l_offset nFPos=VSIFTellL(fp);
    CPLError( CE_Failure, CPLE_FileIO,
              "QEDFGetc: failed to read one byte (got %zu) at x%llx=%lld",
              res, nFPos, nFPos);
    return EOF;
    // Trying to force a crash
    // so that the debugger catches the program before it exits
    // int a = res/res;
    //exit(-1);
    // return CE_Failure;
#endif
} // QEDFGetc

unsigned int QEDFGetuc(VSILFILE *fp) {
    return static_cast<unsigned int>(QEDFGetc(fp));
}


int QEDGetInt( void* base, int byteOffset )
{
    unsigned char * buf = static_cast<unsigned char*>(base)+byteOffset;
#if 0
    int vv;
    vv = buf[0];
    vv |= (buf[1] << 8u);
    vv |= (buf[2] << 16u);
    vv |= (buf[3] << 24u);

    int lsb=CPL_LSBSINT32PTR(buf);
    if (vv==lsb) {
        CPLDebug("QED", "x%x is LSB-endian",vv);
    } else {
        CPLDebug("QED", "assuming x%x is MSB-endian",vv);
    }

    return(vv);
#else
    return(CPL_LSBSINT32PTR(buf));
#endif
}

unsigned int QEDGetUInt( void* base, int byteOffset )
{
#if 0
    return(static_cast<unsigned int>(QEDGetInt(base, byteOffset )));
#else
    unsigned char * buf = static_cast<unsigned char*>(base)+byteOffset;
    return(CPL_LSBUINT32PTR(buf));
#endif
}


/************************************************************************/
/* ==================================================================== */
/*                         QEDDataset                                   */
/* ==================================================================== */
/************************************************************************/

class QEDRasterBand;

class QEDDataset : public GDALPamDataset
{
    friend class QEDRasterBand;
    
    VSILFILE            *fp;
    // GDALColorTable      *poColorTable;
    GByte       abyHeader[0x5a0];
    
    unsigned int magic, version;
    double QCTpolynomial[40];
    latlon TL, TR, BL, BR;
    int qed_tile_size, qed_tile_pixels;
    size_t nQEDfilesize;
    // double dScalePixels;
    const double dScalePixels=1.0;
#ifdef QCT_USE_GCP
    int         nGCPCount;
    GDAL_GCP    *pasGCPList;
    int QCTsetGCPMetaData_from_coefficients(int count, const char* szMDItem,
                                         double values[], const char* szCollection);
#endif
    
public:
    QEDDataset();
    ~QEDDataset() override;
    
    static GDALDataset *Open( GDALOpenInfo * );
    static int Identify( GDALOpenInfo * );

#ifdef QCT_USE_GCP
    virtual int     GetGCPCount() override;
    virtual const GDAL_GCP* GetGCPs() override;
#if GDAL_VERSION_MAJOR >= 3
    virtual const char *GetGCPProjection();
#else
    virtual const char *GetGCPProjection() override;
#endif

#else // ! QCT_USE_GCP
    CPLErr GetGeoTransform( double * padfTransform ) override;
#endif // QCT_USE_GCP


#if GDAL_VERSION_MAJOR >= 3
    const OGRSpatialReference* GetSpatialRef() const override {
        return GetSpatialRefFromOldGetProjectionRef();
    }
    const char *_GetProjectionRef() override;
#else
    const char *GetProjectionRef() override;
#endif

    char       **GetFileList() override;

    char *QEDGetString(size_t byteaddr);
    //char *QEDGetStringIndirect(size_t byteaddr);
    void xy_to_latlon(double pixel_x, double pixel_y, latlon *latlon);
}; // class QEDDataset


/************************************************************************/
/* ==================================================================== */
/*                            QEDRasterBand                             */
/* ==================================================================== */
/************************************************************************/

class QEDRasterBand : public GDALPamRasterBand
{
    friend class QEDDataset;
    
    GDALColorInterp  eBandInterp;

public:

    QEDRasterBand( QEDDataset *poDSIn, int nBandIn );

    ~QEDRasterBand() override;
    
    virtual GDALColorInterp GetColorInterpretation() override;

    virtual double GetNoDataValue( int *) override;
    // virtual CPLErr SetNoDataValue( double ) override;
    
    virtual CPLErr IReadBlock( int, int, void * ) override;

    virtual void read_QED_Tile( VSILFILE *fp,
                                int tile_xx, int tile_yy,
                                void *pimage);

    virtual int    GetOverviewCount() override;

}; // class QEDRasterBand


/************************************************************************/
/*                           QEDRasterBand()                            */
/************************************************************************/

QEDRasterBand::QEDRasterBand(
                             QEDDataset *poDSIn,
                             int nBandIn
                             )
:
    eBandInterp(GCI_GrayIndex)
{
    poDS=poDSIn;
    nBand=nBandIn;

    if ( poDSIn->magic != qed_magic ) {
        CPLError(CE_Failure, CPLE_NotSupported,
                 "unexpected file magic 0x%8x", poDSIn->magic );
    }

    CPLDebug("QED",
             "nBlockXSize %d nBlockYSize %d - setting to %d",
             nBlockXSize, nBlockYSize, poDSIn->qed_tile_size);

    nBlockXSize = nBlockYSize = poDSIn->qed_tile_size;
    eDataType = GDT_Int16;
    eBandInterp = GCI_GrayIndex;

} // QEDRasterBand::QEDRasterBand

/************************************************************************/
/*                          ~QEDRasterBand()                            */
/************************************************************************/

QEDRasterBand::~QEDRasterBand()
{
}

/************************************************************************/
/*                             IReadBlock()                             */
/************************************************************************/

CPLErr QEDRasterBand::IReadBlock( int nBlockXOff, int nBlockYOff,
                                  void * pImage )
{
    QEDDataset *poQDS = static_cast<QEDDataset *>(poDS);

    if (nBlockXOff<0 || nBlockYOff<0 || pImage==nullptr) {
        CPLError( CE_Failure, CPLE_IllegalArg, "bad arg in QEDRasterBand::IReadBlock(%d,%d,%p)",
                  nBlockXOff, nBlockYOff, pImage );
            return CE_Failure;
    }
    
    if (poQDS->magic != qed_magic) {
        CPLDebug("QED", "unsuported magic x%08x in QCT file", poQDS->magic);
        return CE_Failure;
    }

    if (poQDS->version != 0) {
        CPLError( CE_Failure, CPLE_NotSupported, 
                  "QCT QEDRasterBand::IReadBlock only supports version 0 qed files" );
        return CE_Failure;
    }
    if (poQDS->qed_tile_size<1) {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "invalid QED tile size %d",
                  poQDS->qed_tile_size);
        return CE_Failure;
    }

    size_t nQEDtileIndexIndex = 0x158
        + 4 *(static_cast<unsigned int>(nBlockXOff)
              + static_cast<unsigned int>(nBlockYOff)*static_cast<unsigned long long>(poQDS->nRasterXSize)/static_cast<unsigned long long>(poQDS->qed_tile_size)
              );
    if ( VSIFSeekL( poQDS->fp, nQEDtileIndexIndex, SEEK_SET ) ) {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "QED tile location seek x%zx failed",
                  nQEDtileIndexIndex);
        return CE_Failure;
    }

    {
        unsigned int nTilePos = QCTreadUInt(poQDS->fp);
        CPLDebug("QED","QEDRasterBand::IReadBlock(%d %d %p) tile starts at %d",
                 nBlockXOff, nBlockYOff, pImage, nTilePos
                 );
        if (nTilePos>0 && nTilePos < poQDS->nQEDfilesize ) {
            if ( VSIFSeekL( poQDS->fp, nTilePos, SEEK_SET ) ) {
                CPLError( CE_Failure, CPLE_AppDefined, 
                          "QCT tile %d %d seek to %d failed",
                          nBlockXOff, nBlockYOff, nTilePos );
                return CE_Failure;
            }
            
            /* file ptr now at start of tile */
            read_QED_Tile(poQDS->fp, nBlockXOff, nBlockYOff, pImage);
        } else {
            // No data present
            for(int i = 0; i < nBlockXSize *nBlockYSize ; i++ ) {
                static_cast<signed short*>(pImage)[i] = nQEDNoData;
            }
#if 0
            // this draws a box
            for(int i = 0; i < nBlockXSize ; i++ ) {
                static_cast<signed short*>(pImage)[i] = nQEDYesData;
                static_cast<signed short*>(pImage)[nBlockXSize *nBlockYSize - i] = nQEDYesData;
            }
            for(int i = 1; i < nBlockYSize-1 ; i++ ) {
                static_cast<signed short*>(pImage)[i*nBlockXSize] = nQEDYesData;
                static_cast<signed short*>(pImage)[(i+1)*nBlockXSize-1] = nQEDYesData;
            }
#endif
        }
    }
    
    return CE_None;
} // QEDRasterBand::IReadBlock()

/************************************************************************/
/*                           GetNoDataValue()                           */
/************************************************************************/

double QEDRasterBand::GetNoDataValue( int *pbSuccess )
{
    return GDALPamRasterBand::GetNoDataValue( pbSuccess );
}


/************************************************************************/
/*                       GetColorInterpretation()                       */
/************************************************************************/

GDALColorInterp QEDRasterBand::GetColorInterpretation()
{
    return eBandInterp;
}


/************************************************************************/
/*                          GetOverviewCount()                          */
/************************************************************************/

int QEDRasterBand::GetOverviewCount()

{
    return GDALRasterBand::GetOverviewCount();
}


/************************************************************************/
/* ==================================================================== */
/*                            QEDDataset                                */
/* ==================================================================== */
/************************************************************************/


/************************************************************************/
/*                            QEDDataset()                              */
/************************************************************************/

QEDDataset::QEDDataset() :
    fp(nullptr)

#ifdef CODE_ANALYSIS
    // These are only here to keep cppcheck happy
    // They may make the program think things are OK when they are not.
    ,magic(0)
    ,version(0)
#endif // CODE_ANALYSIS

    ,qed_tile_size(0)
    ,qed_tile_pixels(0)
    ,nQEDfilesize(0)

#ifdef QCT_USE_GCP
    ,nGCPCount(0)
    ,pasGCPList(nullptr)
#endif
{

#if 1
    // adfQCTdatumShift[0] = adfQCTdatumShift[1] = 0.0;
    {
        for (int i=0;i<40;i++) {
            QCTpolynomial[i]=0.0;
        }
    }
    TL.lat=INT_MIN; TL.lon=INT_MAX;
    TR.lat=INT_MIN; TR.lon=INT_MIN;
    BL.lat=INT_MAX; BL.lon=INT_MAX;
    BR.lat=INT_MAX; BR.lon=INT_MIN;
#endif

} // QEDDataset::QEDDataset()

/************************************************************************/
/*                           ~QEDDataset()                             */
/************************************************************************/

QEDDataset::~QEDDataset()

{
    FlushCache();

#ifdef QCT_USE_GCP
    if(pasGCPList != nullptr) {
        if( nGCPCount > 0 ) {
            CPLDebug("QED", "GDALDeinitGCPs( %d %p)", nGCPCount, pasGCPList ); 
            GDALDeinitGCPs( nGCPCount, pasGCPList );
            CPLDebug("QED", "GDALDeinitGCPs( %d %p) completed", nGCPCount, pasGCPList );
            VSIFree( pasGCPList );
            pasGCPList = nullptr;
            nGCPCount = 0;
        } else {
            CPLDebug("QED", "GDALDeinitGCPs( %d %p) 0 nonnullptr", nGCPCount, pasGCPList );
        }
    }
#endif

    if( fp != nullptr ) {
        VSIFCloseL( fp );
    }
    
} // QEDDataset::~QEDDataset()


/************************************************************************/
/*                            GetGCPCount()                             */
/************************************************************************/
#ifdef QCT_USE_GCP
int QEDDataset::GetGCPCount()
{
    // CPLDebug("QED", "QEDDataset::GetGCPCount() returns %d", nGCPCount);
    return nGCPCount;
} // QEDDataset::GetGCPCount
#endif // QCT_USE_GCP

/************************************************************************/
/*                          GetGCPProjection()                          */
/************************************************************************/
#ifdef QCT_USE_GCP
const char *QEDDataset::GetGCPProjection()
{
    if( nGCPCount > 0 ) {
        // CPLDebug("QED", "QEDDataset::GetGCPProjection() returns %s",
        //         GetProjectionRef());
        return GetProjectionRef();
        
    } else {
        CPLDebug("QED", "QEDDataset::GetGCPProjection() returns null string");
        return "";
    }
} // QEDDataset::GetGCPProjection
#endif // QCT_USE_GCP

/************************************************************************/
/*                               GetGCP()                               */
/************************************************************************/
#ifdef QCT_USE_GCP
const GDAL_GCP *QEDDataset::GetGCPs()

{
    // CPLDebug("QED", "QEDDataset::GetGCPs() returns %p", pasGCPList);
    return pasGCPList;
}
#endif // QCT_USE_GCP

/************************************************************************/
/*                          GetGeoTransform()                           */
/************************************************************************/
// #if 1 //ndef  QCT_USE_GCP
#ifndef QCT_USE_GCP
CPLErr QEDDataset::GetGeoTransform( double * padfTransform )
{
    // Xgeo = padfTransform[0] + pixel*padfTransform[1] + line*padfTransform[2];
    // Ygeo = padfTransform[3] + pixel*padfTransform[4] + line*padfTransform[5];

    padfTransform[0] = TL.lon;
    padfTransform[1] = (TR.lon - TL.lon) / GetRasterXSize();
    padfTransform[2] = (BL.lon - TL.lon) / GetRasterYSize();
    padfTransform[3] = TL.lat;
    padfTransform[4] = (TR.lat - TL.lat) / GetRasterXSize();
    padfTransform[5] = (BL.lat - TL.lat) / GetRasterYSize();

    return CE_None;
}
#endif // ifndef  QCT_USE_GCP

/************************************************************************/
/*                          GetProjectionRef()                          */
/************************************************************************/

// Gdal <3 uses proj.4, Gdal>=3 uses proj.6, see eg:
// https://trac.osgeo.org/gdal/wiki/rfc73_proj6_wkt2_srsbarn
// https://gdal.org/development/rfc/rfc73_proj6_wkt2_srsbarn.html
#if GDAL_VERSION_MAJOR >= 3
const char *QEDDataset::_GetProjectionRef
#else
const char *QEDDataset::GetProjectionRef
#endif
(){
#if 1
    CPLDebug("QED", "QEDDataset::GetProjectionRef() returns WGS84");
    return "WGS84"; // pszDatum;
#else
    return GDALDataset::GetProjectionRef();
#endif
}

/************************************************************************/
/*                              Identify()                              */
/************************************************************************/

int QEDDataset::Identify( GDALOpenInfo * poOpenInfo )

{
/* -------------------------------------------------------------------- */
/*         This has to be a file on disk ending in .qed                 */
/*                with the right magic number                           */
/* -------------------------------------------------------------------- */

    CPLDebug("QED", "QEDDataset::Identify(%p)\n", poOpenInfo);

    if( poOpenInfo->nHeaderBytes < 8 ) {
        return FALSE;
    }

    unsigned int magic = QEDGetUInt(poOpenInfo->pabyHeader, 0);
    unsigned int version = QEDGetUInt(poOpenInfo->pabyHeader, 4);

    if( magic == qed_magic && EQUAL(CPLGetExtension(poOpenInfo->pszFilename),"qed")) {
      // This file contains elevation data.
      if (version == 0) {
            CPLDebug("QED", ".qed height data not yet properly supported\n");
            return TRUE;
        } else {
            CPLDebug("QCT ", "unexpected QED version %08x\n", version);
            return TRUE;
        }
    }

    return FALSE;
} // QEDDataset::Identify



/************************************************************************/
/*                   QCTsetGCPMetaData_from_coefficients()                 */
/************************************************************************/

#ifdef QCT_USE_GCP
int QEDDataset::QCTsetGCPMetaData_from_coefficients(
                       int count, const char* szMDItem, double values[], const char* szCollection)
{
    if (count<0 || count>30) {
        CPLDebug("QED", "not setting metadata; silly number %d of coefficients given", count);
        return -1;
    }
    char szValue[1280]={};
    size_t nCurPos=0;
    for(int i = 0; i < count; i++ ) {
        if (sizeof(szValue)-nCurPos<20) {
            CPLDebug("QED", "not enough space for %s[%d]", szMDItem, i);
            return -1;
        }

        int nRes=snprintf( szValue+nCurPos, 20, "%.16g ", values[i] );
        if (nRes<0) {
            CPLDebug("QED", "error writing %s[%d]", szMDItem, i);
            return -1;
        }
        nCurPos +=nRes;
    }
    if (nCurPos>=sizeof(szValue)) {
        szValue[sizeof(szValue)-1] = '\0';
    } else {
        szValue[nCurPos] = '\0';
    }
    this->SetMetadataItem( szMDItem, szValue, szCollection );

    return 0;
}
#endif // QCT_USE_GCP

/************************************************************************/
/*                                Open()                                */
/************************************************************************/

GDALDataset *QEDDataset::Open( GDALOpenInfo * poOpenInfo )

{
    CPLDebug("QED", "QEDDataset::Open(%p)\n", poOpenInfo);
    
    // char pszQCTDatum[] = "WGS84";

    if (!Identify(poOpenInfo))
        return nullptr;

#ifdef QCT_USE_GCP
    CPLDebug("QED", "QCT with GCP");
#else
    CPLDebug("QED", "QCT without GCP");
#endif

#if GDAL_VERSION_MAJOR >= 2
    /* Check that the file pointer from GDALOpenInfo* is available */
    if( poOpenInfo->fpL == nullptr )
    {
        return nullptr;
    }
#endif

/* -------------------------------------------------------------------- */
/*      Confirm the requested access is supported.                      */
/* -------------------------------------------------------------------- */
    if( poOpenInfo->eAccess == GA_Update )
    {
        CPLError( CE_Failure, CPLE_NotSupported, 
                  "The QCT driver does not support update access to existing"
                  " datasets.\n" );
        return nullptr;
    }

/* -------------------------------------------------------------------- */
/*      Create a corresponding GDALDataset.                             */
/* -------------------------------------------------------------------- */
    QEDDataset *poDS;

    poDS = new QEDDataset();

#if 0 // GDAL_VERSION_MAJOR >= 2
    /* Borrow the file pointer from GDALOpenInfo* */
    poDS->fp = poOpenInfo->fpL;
    poOpenInfo->fpL = nullptr;
#else
    poDS->fp = VSIFOpenL( poOpenInfo->pszFilename, "rb" );
    if (poDS->fp == nullptr) {
        CPLDebug("QED", "Could not open QCT file %s",
                 poOpenInfo->pszFilename);
        delete poDS;
        return nullptr;
    }
#endif
    VSIStatBufL osRscStatBuf={};
    if ( VSIStatL( poOpenInfo->pszFilename, &osRscStatBuf ) != 0 ) {
        CPLDebug("QED", "No QCT file %s",
                 poOpenInfo->pszFilename);
        delete poDS;
        return nullptr;
    }
    poDS->nQEDfilesize=osRscStatBuf.st_size;
    CPLDebug("QED", "QED file %s is x%zx=%zu bytes",
             poOpenInfo->pszFilename, poDS->nQEDfilesize, poDS->nQEDfilesize
             );
    
/* -------------------------------------------------------------------- */
/*      Read the header.                                                */
/* -------------------------------------------------------------------- */
    VSIFReadL( poDS->abyHeader, 1, 0x5a0, poDS->fp );

    poDS->magic = QEDGetUInt(poOpenInfo->pabyHeader, 0);
    poDS->version = QEDGetUInt(poOpenInfo->pabyHeader, 4);

    if ( poDS->magic == qed_magic ) {
        poDS->qed_tile_size = QEDGetInt( poDS->abyHeader, 0x8 );
        if ( poDS->qed_tile_size != 15 && poDS->qed_tile_size != 20 ) {
            CPLDebug("QED", "QED block size %d - expected 15 or 20\n",
                     poDS->qed_tile_size );
            if ( poDS->qed_tile_size >= 1l<<16) {
                CPLError( CE_Warning, CPLE_AppDefined, 
                          "QCT QED block too big" );
            }
        }
        poDS->qed_tile_pixels = poDS->qed_tile_size * poDS->qed_tile_size;
        poDS->nRasterXSize = poDS->qed_tile_size * QEDGetInt( &poDS->abyHeader, 0x10 );
        poDS->nRasterYSize = poDS->qed_tile_size * QEDGetInt( &poDS->abyHeader, 0xC );
        
        if  (poDS->nRasterXSize <= 0 || poDS->nRasterYSize <= 0 ) {
            CPLError( CE_Failure, CPLE_AppDefined, 
                      "Invalid dimensions : %d x %d", 
                      poDS->nRasterXSize, poDS->nRasterYSize); 
            delete poDS;
            return nullptr;
        }

        CPLDebug("QED", "QED bytes 0x04-0x07 %08x",
                 QEDGetInt( &poDS->abyHeader, 4 ) );
        CPLDebug("QED", "QED bytes 0x14-0x17 %08x",
                 QEDGetInt( &poDS->abyHeader, 0x14 ) );

    }

    
    /********************************************************************/
    /*                       Set coordinate model                       */
    /********************************************************************/
    /* try RPC domain
       http://www.gdal.org/gdal_datamodel.html
       http://geotiff.maptools.org/rpc_prop.html
       GDALExtractRPCInfo
           /home/src/redhat/BUILD/gdal-1.9.2-fedora/frmts/gtiff/geotiff.cpp
       GTiffDataset::ReadRPCTag()
       /home/src/redhat/BUILD/gdal-1.9.2-fedora/frmts/nitf/nitf_gcprpc.cpp

         C1 + C2.L + C3.P + C5.L.P              + C4.H
       + C8.L.L + C9.P.P                        + C6.L.H  + C7.P.H + C10.H.H
       + C12.L.L.L + C13.L.P.P  + C15.L.L.P     + C11.P.L.H + C14.L.H.H
       + C16.P.P.P                              + C17.P.H.H + C18.L.L.H + C19.P.P.H + C20.H.H.H
    */

    size_t nPolynomialOffset;
    if ( poDS->magic == qed_magic ) {
        nPolynomialOffset = 0x18;
    } else {
        nPolynomialOffset = 0x60;
    }
    if ( 40!=QCTreadDoubles( poDS->QCTpolynomial, poDS->fp, nPolynomialOffset, static_cast<size_t>(40) ) ) {
        CPLDebug("QED", "problem reading coordinate polynomial");
        return nullptr;
    }    

#if defined QCT_USE_GCP
    CPLDebug("QED",
             "Building GCPs for %d x %d image scaled %g",
             poDS->nRasterXSize, poDS->nRasterYSize,
             poDS->dScalePixels
             );

    // calculate corner coordinates
    poDS->xy_to_latlon(0.0, 0.0, &poDS->TL);
    poDS->xy_to_latlon(poDS->nRasterXSize, 0.0, &poDS->TR);
    poDS->xy_to_latlon(0.0, poDS->nRasterYSize, &poDS->BL);
    poDS->xy_to_latlon(poDS->nRasterXSize,
                       poDS->nRasterYSize, &poDS->BR);

    CPLDebug("QED", "nRaster.Size %d x %d",
             poDS->nRasterXSize, poDS->nRasterYSize);
    CPLDebug("QED", "TL %g %g",
             poDS->TL.lon, poDS->TL.lat);
    CPLDebug("QED", "TR %g %g",
             poDS->TR.lon, poDS->TR.lat);
    CPLDebug("QED", "BL %g %g",
             poDS->BL.lon, poDS->BL.lat);
    CPLDebug("QED", "BR %g %g",
             poDS->BR.lon, poDS->BR.lat);

    
    // Build a regular 4x4 grid of GCPs
    // Since the QCT polynomials are cubic this should be sufficient ?
    poDS->pasGCPList = (GDAL_GCP *) VSIMalloc2( sizeof(GDAL_GCP), 16);
    {
        int x, nPixel=0, y, nLine=0;
        poDS->nGCPCount = 0;

        GDALInitGCPs( 16, poDS->pasGCPList );
        
        for (x=0 ; x<=3 ; x++) {
            switch (x) {
            case 0: nPixel = 0; break;
            case 1: nPixel = (poDS->nRasterXSize /* +1.0 */ )/3.0; break;
            case 2: nPixel = (poDS->nRasterXSize*2.0 /*+1.0 */ )/3.0; break;
            case 3: nPixel =  poDS->nRasterXSize; break;
            }
            for (y=0 ; y<=3 ; y++) {
                char    szId[128]={};
                
                switch (y) {
                case 0: nLine = 0; break; 
                case 1: nLine = (poDS->nRasterYSize /* +1.0 */ )/3.0; break; 
                case 2: nLine = (poDS->nRasterYSize*2.0 /* +1.0 */ )/3.0; break; 
                case 3: nLine =  poDS->nRasterYSize; break; 
                }
                
                if ( poDS->pasGCPList[poDS->nGCPCount].pszId ) {
                    VSIFree( poDS->pasGCPList[poDS->nGCPCount].pszId );
                }
                snprintf( szId, sizeof(szId), "%d", poDS->nGCPCount+1 );
                poDS->pasGCPList[poDS->nGCPCount].pszId = VSIStrdup( szId );
                if ( poDS->pasGCPList[poDS->nGCPCount].pszInfo ) {
                    VSIFree( poDS->pasGCPList[poDS->nGCPCount].pszInfo );
                }
                poDS->pasGCPList[poDS->nGCPCount].pszInfo = VSIStrdup("");
                poDS->pasGCPList[poDS->nGCPCount].dfGCPPixel = nPixel;
                poDS->pasGCPList[poDS->nGCPCount].dfGCPLine  = nLine;
                {
                    latlon tempLatLong;
                    poDS->xy_to_latlon(nPixel, nLine, &tempLatLong);
                    poDS->pasGCPList[poDS->nGCPCount].dfGCPX = tempLatLong.lon;
                    poDS->pasGCPList[poDS->nGCPCount].dfGCPY = tempLatLong.lat;
                    poDS->pasGCPList[poDS->nGCPCount].dfGCPZ = 0.0;
                }
                
                CPLDebug("QED GCP",
                         "GCP[%d] %s, %s (%g, %g) -> (%g, %g, %g)",
                         poDS->nGCPCount,
                         poDS->pasGCPList[poDS->nGCPCount].pszId,
                         poDS->pasGCPList[poDS->nGCPCount].pszInfo,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPPixel,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPLine,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPX,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPY,
                         poDS->pasGCPList[poDS->nGCPCount].dfGCPZ
                         );
                
                poDS->nGCPCount++;
            }
        }
    } // Build a regular 4x4 grid of GCPs
    
    char szValue[64]="";
    GDALRPCInfo sRPCInfo;
    
    // Use the corner coordinates calculated above
    sRPCInfo.dfMAX_LAT = sRPCInfo.dfMIN_LAT = poDS->TL.lat;
    sRPCInfo.dfMAX_LONG = sRPCInfo.dfMIN_LONG = poDS->TL.lon;
    
    sRPCInfo.dfMIN_LAT = MIN(sRPCInfo.dfMIN_LAT, poDS->TR.lat);
    sRPCInfo.dfMAX_LAT = MAX(sRPCInfo.dfMAX_LAT, poDS->TR.lat);
    sRPCInfo.dfMIN_LONG = MIN(sRPCInfo.dfMIN_LONG, poDS->TR.lon);
    sRPCInfo.dfMAX_LONG = MAX(sRPCInfo.dfMAX_LONG, poDS->TR.lon);
    
    sRPCInfo.dfMIN_LAT = MIN(sRPCInfo.dfMIN_LAT, poDS->BL.lat);
    sRPCInfo.dfMAX_LAT = MAX(sRPCInfo.dfMAX_LAT, poDS->BL.lat);
    sRPCInfo.dfMIN_LONG = MIN(sRPCInfo.dfMIN_LONG, poDS->BL.lon);
    sRPCInfo.dfMAX_LONG = MAX(sRPCInfo.dfMAX_LONG, poDS->BL.lon);
    
    sRPCInfo.dfMIN_LAT = MIN(sRPCInfo.dfMIN_LAT, poDS->BR.lat);
    sRPCInfo.dfMAX_LAT = MAX(sRPCInfo.dfMAX_LAT, poDS->BR.lat);
    sRPCInfo.dfMIN_LONG = MIN(sRPCInfo.dfMIN_LONG, poDS->BR.lon);
    sRPCInfo.dfMAX_LONG = MAX(sRPCInfo.dfMAX_LONG, poDS->BR.lon);

    CPLDebug("QED RPC", "min lat %g", sRPCInfo.dfMIN_LAT);
    CPLDebug("QED RPC", "max lat %g", sRPCInfo.dfMAX_LAT);
    CPLDebug("QED RPC", "min lon %g", sRPCInfo.dfMIN_LONG);
    CPLDebug("QED RPC", "max lon %g", sRPCInfo.dfMAX_LONG);

    double dfDiffLat = sRPCInfo.dfMAX_LAT - sRPCInfo.dfMIN_LAT;
    sRPCInfo.dfLAT_OFF = 0; // poDS->adfQCTdatumShift[0] ; // north
    if (fabs(dfDiffLat) < 1e-8) {
        CPLDebug("QED RPC", "bounding box has no width %g=%g",
                 sRPCInfo.dfMIN_LAT, sRPCInfo.dfMAX_LAT);
        sRPCInfo.dfLAT_SCALE = 0.0; // Could I set this to NaN ?
    } else {
        sRPCInfo.dfLAT_SCALE = 2.0 / dfDiffLat;
    }

    double dfDiffLon = sRPCInfo.dfMAX_LONG - sRPCInfo.dfMIN_LONG;
    sRPCInfo.dfLONG_OFF = 0; // poDS->adfQCTdatumShift[1] ; // east

    if (fabs(dfDiffLon) < 1e-8) {
        CPLDebug("QED RPC", "bounding box has no height %g=%g",
                 sRPCInfo.dfMIN_LONG, sRPCInfo.dfMAX_LONG);
        sRPCInfo.dfLONG_SCALE = 0.0; // Could I set this to NaN ?
    } else {
        sRPCInfo.dfLONG_SCALE = 2.0 / dfDiffLon;
    }
    
    sRPCInfo.dfSAMP_OFF = 0.0 ;
    sRPCInfo.dfSAMP_SCALE = 1.0; // poDS->dScalePixels; // 1.0 ;

    sRPCInfo.dfLINE_OFF = 0.0 ;
    sRPCInfo.dfLINE_SCALE = 1.0; // poDS->dScalePixels; // 1.0 ;

    sRPCInfo.dfHEIGHT_OFF = 0.0 ;
    sRPCInfo.dfHEIGHT_SCALE = 1.0 ;

    sRPCInfo.adfLINE_NUM_COEFF[0] = 1.0;
    sRPCInfo.adfSAMP_NUM_COEFF[0] = 1.0;
    sRPCInfo.adfLINE_DEN_COEFF[0] = 1.0;
    sRPCInfo.adfSAMP_DEN_COEFF[0] = 1.0;
    for(int i = 1; i < 20; i++ ) {
        sRPCInfo.adfLINE_NUM_COEFF[i] = 0.0;
        sRPCInfo.adfSAMP_NUM_COEFF[i] = 0.0;
        sRPCInfo.adfLINE_DEN_COEFF[i] = 0.0;
        sRPCInfo.adfSAMP_DEN_COEFF[i] = 0.0;
    }

    /* now set some of these variables from data in the file */
    /* these are probably not correctly scaled yet */
    /* ( lat p, long l ) -> row */
    sRPCInfo.adfLINE_NUM_COEFF[0]  = poDS->QCTpolynomial[10]; // const term
    sRPCInfo.adfLINE_NUM_COEFF[1]  = poDS->QCTpolynomial[11]; //  l 
    sRPCInfo.adfLINE_NUM_COEFF[2]  = poDS->QCTpolynomial[12]; //  p
    sRPCInfo.adfLINE_NUM_COEFF[4]  = poDS->QCTpolynomial[14]; //  lp
    sRPCInfo.adfLINE_NUM_COEFF[7]  = poDS->QCTpolynomial[15]; //  ll
    sRPCInfo.adfLINE_NUM_COEFF[8]  = poDS->QCTpolynomial[13]; //  pp
    sRPCInfo.adfLINE_NUM_COEFF[11] = poDS->QCTpolynomial[19]; //  lll
    sRPCInfo.adfLINE_NUM_COEFF[12] = poDS->QCTpolynomial[18]; //  lpp
    sRPCInfo.adfLINE_NUM_COEFF[14] = poDS->QCTpolynomial[17]; //  llp
    sRPCInfo.adfLINE_NUM_COEFF[15] = poDS->QCTpolynomial[16]; //  ppp
    /* ( lat p, long l ) -> column/pixel */
    sRPCInfo.adfSAMP_NUM_COEFF[0]  = poDS->QCTpolynomial[0]; // const term
    sRPCInfo.adfSAMP_NUM_COEFF[1]  = poDS->QCTpolynomial[1]; //  l 
    sRPCInfo.adfSAMP_NUM_COEFF[2]  = poDS->QCTpolynomial[2]; //  p
    sRPCInfo.adfSAMP_NUM_COEFF[4]  = poDS->QCTpolynomial[4]; //  lp
    sRPCInfo.adfSAMP_NUM_COEFF[7]  = poDS->QCTpolynomial[5]; //  ll
    sRPCInfo.adfSAMP_NUM_COEFF[8]  = poDS->QCTpolynomial[3]; //  pp
    sRPCInfo.adfSAMP_NUM_COEFF[11] = poDS->QCTpolynomial[9]; //  lll
    sRPCInfo.adfSAMP_NUM_COEFF[12] = poDS->QCTpolynomial[8]; //  lpp
    sRPCInfo.adfSAMP_NUM_COEFF[14] = poDS->QCTpolynomial[7]; //  llp
    sRPCInfo.adfSAMP_NUM_COEFF[15] = poDS->QCTpolynomial[6]; //  ppp
 
    /* put the values where gdal will use them */
    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLINE_OFF );
    poDS->SetMetadataItem( "LINE_OFF", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLINE_SCALE );
    poDS->SetMetadataItem( "LINE_SCALE", szValue, "RPC" );
    
    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfSAMP_OFF );
    poDS->SetMetadataItem( "SAMP_OFF", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfSAMP_SCALE );              
    poDS->SetMetadataItem( "SAMP_SCALE", szValue, "RPC" );
   
    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLONG_OFF );
    poDS->SetMetadataItem( "LONG_OFF", szValue, "RPC" );
       
    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLONG_SCALE );
    poDS->SetMetadataItem( "LONG_SCALE", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLAT_OFF );
    poDS->SetMetadataItem( "LAT_OFF", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfLAT_SCALE );
    poDS->SetMetadataItem( "LAT_SCALE", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfHEIGHT_OFF );
    poDS->SetMetadataItem( "HEIGHT_OFF", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfHEIGHT_SCALE );
    poDS->SetMetadataItem( "HEIGHT_SCALE", szValue, "RPC" );

    poDS->QCTsetGCPMetaData_from_coefficients(20, "LINE_NUM_COEFF",
                                              sRPCInfo.adfLINE_NUM_COEFF, "RPC");
    poDS->QCTsetGCPMetaData_from_coefficients(20, "LINE_DEN_COEFF",
                                              sRPCInfo.adfLINE_DEN_COEFF, "RPC");
    poDS->QCTsetGCPMetaData_from_coefficients(20, "SAMP_NUM_COEFF",
                                              sRPCInfo.adfSAMP_NUM_COEFF, "RPC");
    poDS->QCTsetGCPMetaData_from_coefficients(20, "SAMP_DEN_COEFF",
                                              sRPCInfo.adfSAMP_DEN_COEFF, "RPC");

    // sRPCInfo.dfMIN_LONG = sRPCInfo.dfLONG_OFF - ( sRPCInfo.dfLONG_SCALE / 2.0 );
    // sRPCInfo.dfMIN_LAT  = sRPCInfo.dfLAT_OFF - ( sRPCInfo.dfLAT_SCALE / 2.0 );
    // sRPCInfo.dfMAX_LONG = sRPCInfo.dfLONG_OFF + ( sRPCInfo.dfLONG_SCALE / 2.0 );
    // sRPCInfo.dfMAX_LAT  = sRPCInfo.dfLAT_OFF + ( sRPCInfo.dfLAT_SCALE / 2.0 );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfMIN_LONG );
    poDS->SetMetadataItem( "MIN_LONG", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfMAX_LONG );
    poDS->SetMetadataItem( "MAX_LONG", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfMIN_LAT );
    poDS->SetMetadataItem( "MIN_LAT", szValue, "RPC" );

    snprintf( szValue, sizeof(szValue), "%.16g", sRPCInfo.dfMAX_LAT);
    poDS->SetMetadataItem( "MAX_LAT", szValue, "RPC" );

    // -1 indicates unknown bias error and random error
    poDS->SetMetadataItem( "ERR_BIAS", "-1.0", "RPC" );
    poDS->SetMetadataItem( "ERR_RAND", "-1.0", "RPC" );
#endif // def QCT_USE_GCP

/* -------------------------------------------------------------------- */
/*      Create band information objects.                                */
/* -------------------------------------------------------------------- */
    QEDRasterBand *poBand =
        new QEDRasterBand( poDS, 1);
    poDS->SetBand( 1, poBand);
    poBand->SetNoDataValue( nQEDNoData );

    
/* -------------------------------------------------------------------- */
/*      Initialize any PAM information.                                 */
/* -------------------------------------------------------------------- */
    poDS->SetDescription( poOpenInfo->pszFilename );
    // poDS->TryLoadXML();

/* -------------------------------------------------------------------- */
/*      Check for overviews.                                            */
/* -------------------------------------------------------------------- */
    /* Let gdal do this for us.
     */
    poDS->oOvManager.Initialize( poDS, poOpenInfo->pszFilename );

    return( poDS );
} // QEDDataset::Open

/************************************************************************/
/*                          GDALRegister_qed()                          */
/************************************************************************/

void GDALRegister_qed(void)
{
    if (! GDAL_CHECK_VERSION("QED"))
        return;

    if( GDALGetDriverByName( "QED" ) != nullptr )
        return;

    GDALDriver *poDriver = new GDALDriver();

    poDriver->SetDescription( "QED" );

    // required in gdal version 2, not supported in gdal 1.11
#if GDAL_VERSION_MAJOR >= 2
    poDriver->SetMetadataItem( GDAL_DCAP_RASTER, "YES" );
#endif

    poDriver->SetMetadataItem( GDAL_DMD_LONGNAME, 
                               "Memory Map Quick Elevation Data map (.qed)" );
    poDriver->SetMetadataItem( GDAL_DMD_HELPTOPIC, 
                               "frmt_various.html#QED" );
    poDriver->SetMetadataItem( GDAL_DMD_EXTENSION, "qed" );
    poDriver->SetMetadataItem( GDAL_DMD_CREATIONDATATYPES, "Byte" );
    poDriver->SetMetadataItem( GDAL_DCAP_VIRTUALIO, "YES" );
    
    poDriver->SetMetadataItem( "LICENSE_POLICY", "NONRECIPROCAL" );

    // Which of these is correct ?
    poDriver->SetMetadataItem(GDALMD_AREA_OR_POINT,
                              // GDALMD_AOP_POINT
                              GDALMD_AOP_AREA
                              );
    // GDALMD_AOP_AREA is the GDALv2 default.
    
    poDriver->pfnOpen = static_cast<GDALDataset *(*)(GDALOpenInfo *)>(QEDDataset::Open);
    poDriver->pfnIdentify = QEDDataset::Identify;

    GetGDALDriverManager()->RegisterDriver( poDriver );
}


/* -------------------------------------------------------------------------
 * Returns a string (nul-terminated) allocated from VSIMalloc.
 * The string starts at byte byteaddr of the QED file.
 * If index pointer is nul then a null pointer is returned
 * (rather than an empty string).
 */
extern char *QEDGetString( VSILFILE *fp, size_t byteaddr )
{
    size_t string_length = 0;
    size_t buf_length = 1024;
    if (byteaddr<1) {
        CPLDebug("QED", "no string at x%zx", byteaddr);
        return nullptr;
    }
    
    char *pszNewString = static_cast<char*>(VSIMalloc(buf_length));
    if (pszNewString == nullptr) {
        CPLError(CE_Failure, CPLE_OutOfMemory,
                 "Cannot allocate space for string");
        exit(-1);
        //return nullptr;
    }
    int seekres = VSIFSeekL( fp,
                             static_cast<vsi_l_offset>(byteaddr),
                             SEEK_SET );
    if ( seekres ) {
        VSIFree(pszNewString);
        CPLError( CE_Failure, CPLE_AppDefined,
                  "cannot seek to QED string" );
        exit(-1);
        //return nullptr;
    }
    CPLDebug("QED", "QEDGetString(fp, x%012zx)",byteaddr);
    while (1) {
        size_t bytesread =
            VSIFReadL( &pszNewString[string_length], 1, 1024, fp);
        CPLDebug("QED",
                 "bytesread %zu string_length %zu buf_length %zu",
                 bytesread, string_length, buf_length);
        while (string_length < buf_length &&
               string_length < bytesread)
            {
                if ( pszNewString[string_length++] == 0 ) {
                    return pszNewString;
                }
            }
        if (bytesread < 1024) {
            VSIFree(pszNewString);
            CPLError(CE_Failure, CPLE_AppDefined,
                     "end of string not found: read %zu/1024 now at %zu/%zu\n",
                     bytesread, string_length, buf_length);
            exit(-1);
            //return nullptr;     
        }
        
        buf_length += 1024;
        pszNewString = static_cast<char*>(VSIRealloc(pszNewString, buf_length));
        if (pszNewString == nullptr) {
            CPLError(CE_Failure, CPLE_OutOfMemory,
                     "Cannot reallocate %zu bytes for string",
                     buf_length);
            exit(-1);
            //return nullptr;
        }
    }

    // return pszNewString;
} // QEDGetString( VSILFILE *fp, size_t byteaddr )


char *QEDDataset::QEDGetString(size_t byteaddr) {
#if 0
    if (0>byteaddr) {
        CPLDebug("QED","QEDGetString: 0> x%012zx",
             byteaddr);
        return nullptr;
    }
#endif
    if(byteaddr >= this->nQEDfilesize) {
        CPLDebug("QED","QEDGetString: x%012zx >= x%012zx",
                 byteaddr, nQEDfilesize);
        return nullptr;
    }
    
    return ::QEDGetString(this->fp, byteaddr);
    
} // QEDDataset::QEDGetString(size_t byteaddr)


/* -------------------------------------------------------------------------
 * Convert pixel (x,y) in image into (longitude,latitude)
 * x,y from top left
 * latitude,longitude in degrees WGS84.
 * clips out of range x,y rather than blindly converting.
 *
 * This is a wrapper for QCT_xy_to_latlon which does the real work.
 */
void
QEDDataset::xy_to_latlon(double x, double y, latlon *latlon)
{
    // Range check - clip to image limits, or allow extrapolation?
    if (x<0) {
        CPLDebug("QED", "xy_to_latlon: %g increased to 0", x);
        x=0;
    }
    if (y<0)  {
        CPLDebug("QED", "xy_to_latlon: %g increased to 0", y);
        y=0;
    }
    if (x>nRasterXSize*dScalePixels)  {
        CPLDebug("QED", "xy_to_latlon: x %g reduced to %d * %g = %g",
                 x, nRasterXSize,dScalePixels, nRasterXSize*dScalePixels);
        x=nRasterXSize*dScalePixels;
    }
    if (y>nRasterYSize*dScalePixels)  {
        CPLDebug("QED", "xy_to_latlon: y %g reduced to %d * %g = %g",
                 y, nRasterXSize,dScalePixels, nRasterXSize*dScalePixels);
        y=nRasterYSize*dScalePixels;
    }
    // Allow limited extrapolation, but clip if way out
    // if (x<-1.0*nRasterXSize*dScalePixels) x = -1.0 * nRasterXSize*dScalePixels;
    // if (y<-1.0*nRasterYSize*dScalePixels) y = -1.0 * nRasterYSize*dScalePixels;
    // if (x> 2.0*nRasterXSize*dScalePixels) x =  2.0 * nRasterXSize*dScalePixels;
    // if (y> 2.0*nRasterYSize*dScalePixels) y =  2.0 * nRasterYSize*dScalePixels;

    double adfQCTdatumShift[]= {0.0,0.0};
    QCT_xy_to_latlon(
                     x*dScalePixels, y*dScalePixels,
                     latlon,
                     QCTpolynomial,
                     adfQCTdatumShift
                     );

    // Clip latitude
    if (latlon->lat >  90.0) {
        latlon->lat =  90.0 ;
    }
    if (latlon->lat < -90.0) {
        latlon->lat = -90.0 ;
    }
    // can longitude wrap infinitely, or should we clip it too ?
} // QEDDataset::xy_to_latlon

/* -------------------------------------------------------------------------
 */


/************************************************************************/
/*                            GetFileList()                             */
/************************************************************************/

char **QEDDataset::GetFileList()
{
    char **papszFileList = GDALPamDataset::GetFileList();

    CPLDebug("QED", "GetDescription %s", GetDescription() );

    return papszFileList;
} // QEDDataSet::GetFileList


void
QEDRasterBand::read_QED_Tile(VSILFILE *fp,
                             int tile_xx, int tile_yy,
                             void *pimage)
{
    signed short *tile_data = static_cast<signed short*>(pimage);
    unsigned int nBitsPerPixel;
    signed int nLowest;
    // int row, bytes_per_row;
    // int pixelnum = 0;

    //CPLDebug("QED", "about to call VSIFTellL for tile %d %d, file %p",
    //         tile_xx, tile_yy, fp);
    vsi_l_offset fileCurrentPos = VSIFTellL(fp);
    CPLDebug("QED", "Tile %d, %d starts at file offset 0x%llx",
             tile_xx, tile_yy, fileCurrentPos);
    (void)tile_xx; // Appease static analyzers
    (void)tile_yy; // Appease static analyzers
    (void)fileCurrentPos; // Appease static analyzers
    if (VSIFEofL(fp)) {
        CPLDebug("QED", "read_QED_Tile(%d %d) called at EOF file offset 0x%llx",
                 tile_xx, tile_yy, fileCurrentPos);
    }
    
    nBitsPerPixel = QCTFGetuc(fp);
    // CPLDebug("QED", "nBitsPerPixel %u", nBitsPerPixel);

    nLowest = static_cast<unsigned char>(QCTFGetuc(fp)) + 256 * static_cast<signed char>(QCTFGetc(fp));
    // if (nLowest > 0x7fff) {
    //     nLowest -= 0x10000;
    // }

    if (nBitsPerPixel==0 || nBitsPerPixel > 16) {
        if (nBitsPerPixel > 16) {
            CPLDebug("QED", "tile %d %d: bpp %d >16: ignoring data",
                     tile_xx, tile_yy, nBitsPerPixel);
        } else {
            CPLDebug("QED", "tile %d %d: 0 bpp - all %d",
                     tile_xx, tile_yy, nLowest);
        }
        for (int i=0; i< nBlockXSize * nBlockYSize; i++) {
            tile_data[i] = static_cast<short>(nLowest);
        }
    } else {
        unsigned int nMask = (1 << nBitsPerPixel) -1;
        CPLDebug("QED", "tile %d %d: %d bpp lowest %d with mask %0x",
                 tile_xx, tile_yy, nBitsPerPixel, nLowest, nMask);

        unsigned int nCurrent=0;
        unsigned int nNext=0;
          signed int nBitsLeft=0;

        for (int i=0; i< nBlockXSize * nBlockYSize; i++) {
            while (nBitsLeft < static_cast<signed int>(nBitsPerPixel)) {
                nNext = QEDFGetuc(fp);
                if (static_cast<signed int>(nNext) == EOF) {
                    break;
                }

                nCurrent |= (nNext << nBitsLeft);
                nBitsLeft += 8; // nBitsPerPixel;
            }
            if ((static_cast<unsigned int>(EOF)) == nNext) {
                CPLDebug("QED", "%s: read beyond eof",
                         "" //poOpenInfo->pszFilename
                         );
                break;
            }
            tile_data[i] = static_cast<short>(nLowest)
                + static_cast<short>(nCurrent & nMask);
            nBitsLeft -= nBitsPerPixel;
            nCurrent >>= nBitsPerPixel;
            
            //CPLDebug("QED", "tile %d %d pixel %d: %d",
            //         tile_xx, tile_yy, i, tile_data[i]);
        }
    }
} // QEDRasterBand::read_QED_Tile


#endif /* def FRMT_qed */
#endif /* def FRMT_qct */
